@@ -1,4 +1,4 @@
-// Copyright (C) 2004-2024 Artifex Software, Inc.
+// Copyright (C) 2004-2025 Artifex Software, Inc.
 //
 // This file is part of MuPDF.
 //
@@ -493,7 +493,7 @@ static void save_attachment_dialog(void)
 		{
 			fz_try(ctx)
 			{
-				pdf_obj *fs = pdf_dict_get(ctx, pdf_annot_obj(ctx, ui.selected_annot), PDF_NAME(FS));
+				pdf_obj *fs = pdf_annot_filespec(ctx, ui.selected_annot);
 				fz_buffer *buf = pdf_load_embedded_file_contents(ctx, fs);
 				fz_save_buffer(ctx, buf, attach_filename);
 				fz_drop_buffer(ctx, buf);
@@ -566,10 +566,16 @@ static void open_stamp_image_dialog(void)
 			fz_var(img);
 			fz_try(ctx)
 			{
+				fz_rect rect = pdf_annot_rect(ctx, ui.selected_annot);
 				trace_action("tmp = new Image(%q);\n", stamp_image_filename);
 				img = fz_new_image_from_file(ctx, stamp_image_filename);
 				trace_action("annot.setAppearance(tmp);\n");
 				pdf_set_annot_stamp_image(ctx, ui.selected_annot, img);
+				pdf_set_annot_rect(ctx, ui.selected_annot, fz_make_rect(
+					rect.x0, rect.y0,
+					rect.x0 + img->w * 72 / img->xres,
+					rect.y0 + img->h * 72 / img->yres)
+				);
 				trace_action("annot.setIcon(%q);\n", fz_basename(stamp_image_filename));
 				pdf_set_annot_icon_name(ctx, ui.selected_annot, fz_basename(stamp_image_filename));
 			}
@@ -768,6 +774,38 @@ static void do_annotate_date(void)
 		ui_label("Date: %s", s);
 }
 
+static void do_annotate_flags(void)
+{
+	char str[100];
+	int n = 0;
+	str[0] = 0;
+	int f = pdf_annot_flags(ctx, ui.selected_annot);
+	if (f & PDF_ANNOT_IS_INVISIBLE) ++n, fz_strlcat(str, "Invisible ", sizeof str);
+	if (f & PDF_ANNOT_IS_HIDDEN) ++n, fz_strlcat(str, "Hidden ", sizeof str);
+	if (f & PDF_ANNOT_IS_PRINT) ++n, fz_strlcat(str, "Print ", sizeof str);
+	if (f & PDF_ANNOT_IS_NO_ZOOM) ++n, fz_strlcat(str, "NoZoom ", sizeof str);
+	if (f & PDF_ANNOT_IS_NO_ROTATE) ++n, fz_strlcat(str, "NoRotate ", sizeof str);
+	if (f & PDF_ANNOT_IS_NO_VIEW) ++n, fz_strlcat(str, "NoView ", sizeof str);
+	if (f & PDF_ANNOT_IS_READ_ONLY) ++n, fz_strlcat(str, "ReadOnly ", sizeof str);
+	if (f & PDF_ANNOT_IS_LOCKED) ++n, fz_strlcat(str, "Locked ", sizeof str);
+	if (f & PDF_ANNOT_IS_TOGGLE_NO_VIEW) ++n, fz_strlcat(str, "ToggleNoView ", sizeof str);
+	if (f & PDF_ANNOT_IS_LOCKED_CONTENTS) ++n, fz_strlcat(str, "LockedContents ", sizeof str);
+	if (str[0])
+	{
+		if (n < 3)
+		{
+			ui_label("Flags: %s", str);
+		}
+		else
+		{
+			ui_label("Flags:");
+			ui.layout->padx = 10;
+			ui_label(str);
+			ui.layout->padx = 0;
+		}
+	}
+}
+
 static const char *intent_names[] = {
 	"Default", // all
 	"FreeTextCallout", // freetext
@@ -784,13 +822,20 @@ static const char *intent_names[] = {
 static enum pdf_intent do_annotate_intent(void)
 {
 	enum pdf_intent intent;
+	const char *intent_name;
 	int choice;
 
 	if (!pdf_annot_has_intent(ctx, ui.selected_annot))
 		return PDF_ANNOT_IT_DEFAULT;
 
 	intent = pdf_annot_intent(ctx, ui.selected_annot);
-	choice = label_select("Intent", "IT", intent_names[intent], intent_names, nelem(intent_names));
+
+	if (intent == PDF_ANNOT_IT_UNKNOWN)
+		intent_name = "Unknown";
+	else
+		intent_name = intent_names[intent];
+
+	choice = label_select("Intent", "IT", intent_name, intent_names, nelem(intent_names));
 	if (choice != -1)
 	{
 		trace_action("annot.setIntent(%d);\n", choice);
@@ -876,7 +921,7 @@ static const char *font_names[] = { "Cour", "Helv", "TiRo" };
 static const char *lang_names[] = { "", "ja", "ko", "zh-Hans", "zh-Hant" };
 static const char *im_redact_names[] = { "Keep images", "Remove images", "Erase pixels" };
 static const char *la_redact_names[] = { "Keep line art", "Remove covered line art", "Remove touched line art" };
-static const char *tx_redact_names[] = { "Remove text", "Keep text" };
+static const char *tx_redact_names[] = { "Remove text", "Keep text", "Remove invisible text" };
 static const char *border_styles[] = { "Solid", "Dashed", "Dotted" };
 static const char *border_intensities[] = { "None", "Small clouds", "Large clouds", "Enormous clouds" };
 
@@ -1095,6 +1140,18 @@ static void do_border(void)
 	}
 }
 
+static int image_file_filter(const char *fn)
+{
+	return !!fz_strstrcase(fn, ".jpg") || !!fz_strstrcase(fn, ".jpeg") || !!fz_strstrcase(fn, ".png");
+}
+
+static void resize_icon_annot(void)
+{
+	fz_rect rect = pdf_annot_rect(ctx, ui.selected_annot);
+	rect = fz_make_rect(rect.x0, rect.y0, rect.x0 + 16, rect.y0 + 16);
+	pdf_set_annot_rect(ctx, ui.selected_annot, rect);
+}
+
 void do_annotate_panel(void)
 {
 	static struct list annot_list;
@@ -1157,6 +1214,7 @@ void do_annotate_panel(void)
 
 		do_annotate_author();
 		do_annotate_date();
+		do_annotate_flags();
 
 		obj = pdf_dict_get(ctx, pdf_annot_obj(ctx, ui.selected_annot), PDF_NAME(Popup));
 		if (obj)
@@ -1324,6 +1382,7 @@ void do_annotate_panel(void)
 				{
 					trace_action("annot.setIcon(%q);\n", text_icons[choice]);
 					pdf_set_annot_icon_name(ctx, ui.selected_annot, text_icons[choice]);
+					resize_icon_annot();
 				}
 				break;
 			case PDF_ANNOT_FILE_ATTACHMENT:
@@ -1333,6 +1392,7 @@ void do_annotate_panel(void)
 				{
 					trace_action("annot.setIcon(%q);\n", file_attachment_icons[choice]);
 					pdf_set_annot_icon_name(ctx, ui.selected_annot, file_attachment_icons[choice]);
+					resize_icon_annot();
 				}
 				break;
 			case PDF_ANNOT_SOUND:
@@ -1342,6 +1402,7 @@ void do_annotate_panel(void)
 				{
 					trace_action("annot.setIcon(%q);\n", sound_icons[choice]);
 					pdf_set_annot_icon_name(ctx, ui.selected_annot, sound_icons[choice]);
+					resize_icon_annot();
 				}
 				break;
 			case PDF_ANNOT_STAMP:
@@ -1351,6 +1412,7 @@ void do_annotate_panel(void)
 				{
 					trace_action("annot.setIcon(%q);\n", stamp_icons[choice]);
 					pdf_set_annot_icon_name(ctx, ui.selected_annot, stamp_icons[choice]);
+					resize_icon_annot();
 				}
 				break;
 			}
@@ -1449,14 +1511,14 @@ void do_annotate_panel(void)
 			if (ui_button("Image..."))
 			{
 				fz_dirname(attname, filename, sizeof attname);
-				ui_init_open_file(attname, NULL);
+				ui_init_open_file(attname, image_file_filter);
 				ui.dialog = open_stamp_image_dialog;
 			}
 		}
 
 		if (pdf_annot_type(ctx, ui.selected_annot) == PDF_ANNOT_FILE_ATTACHMENT)
 		{
-			pdf_embedded_file_params params;
+			pdf_filespec_params params;
 			char attname[PATH_MAX];
 			pdf_obj *fs = pdf_annot_filespec(ctx, ui.selected_annot);
 			ui_spacer();
@@ -1466,7 +1528,7 @@ void do_annotate_panel(void)
 				{
 					fz_dirname(attname, filename, sizeof attname);
 					fz_strlcat(attname, "/", sizeof attname);
-					pdf_get_embedded_file_params(ctx, fs, &params);
+					pdf_get_filespec_params(ctx, fs, &params);
 					fz_strlcat(attname, params.filename, sizeof attname);
 					ui_init_save_file(attname, NULL);
 					ui.dialog = save_attachment_dialog;
