@@ -125,7 +125,8 @@ public class Viewer extends Frame implements WindowListener, ActionListener, Ite
 	protected Panel outlinePanel;
 	protected List outlineList;
 
-	protected Progressmeter meter;
+	protected OCRProgressmeter OCRmeter;
+	protected RenderProgressmeter renderMeter;
 
 	protected int number = 0;
 
@@ -578,10 +579,17 @@ public class Viewer extends Frame implements WindowListener, ActionListener, Ite
 		ctm.f -= atOrigin.y0;
 		Rect bounds = new Rect(bbox).transform(ctm);
 
-		doc.renderPage(ctm, bounds, icc, antialias, invert, tint, tintBlack, tintWhite,
+		Cookie cookie = new Cookie();
+
+		renderMeter = new RenderProgressmeter(this, "Rendering...", cookie, 250);
+		renderMeter.setLocationRelativeTo(this);
+		pageCanvas.requestFocusInWindow();
+
+		doc.renderPage(ctm, bounds, icc, antialias, invert, tint, tintBlack, tintWhite, cookie,
 			new ViewerCore.OnException() {
 				public void run(Throwable t) {
-					exception(t);
+					if (!renderMeter.cancelled)
+						exception(t);
 				}
 			}
 		);
@@ -1393,17 +1401,17 @@ public class Viewer extends Frame implements WindowListener, ActionListener, Ite
 		}
 
 		final String selectedPath = new StringBuffer(fileDialog.getDirectory()).append(File.separatorChar).append(fileDialog.getFile()).toString();
-		meter = new Progressmeter(this, "Saving...", pages);
-		meter.setLocationRelativeTo(this);
-		meter.setVisible(true);
+		OCRmeter = new OCRProgressmeter(this, "Saving...", pages);
+		OCRmeter.setLocationRelativeTo(this);
+		OCRmeter.setVisible(true);
 		pageCanvas.requestFocusInWindow();
 
 		if (options.indexOf("ocr-language=") < 0)
-			doc.save(selectedPath, options, meter, new ViewerCore.OnException() {
+			doc.save(selectedPath, options, OCRmeter, new ViewerCore.OnException() {
 				public void run(Throwable t) {
 					if (t instanceof IOException)
 						exception(t);
-					else if (t instanceof RuntimeException && !meter.cancelled)
+					else if (t instanceof RuntimeException && !OCRmeter.cancelled)
 						exception(t);
 				}
 			});
@@ -1411,9 +1419,9 @@ public class Viewer extends Frame implements WindowListener, ActionListener, Ite
 		{
 			try {
 				FileStream fs = new FileStream(selectedPath, "rw");
-				doc.save(fs, options, meter, new ViewerCore.OnException() {
+				doc.save(fs, options, OCRmeter, new ViewerCore.OnException() {
 					public void run(Throwable t) {
-						if (t instanceof RuntimeException && !meter.cancelled)
+						if (t instanceof RuntimeException && !OCRmeter.cancelled)
 							exception(t);
 					}
 				});
@@ -1424,8 +1432,8 @@ public class Viewer extends Frame implements WindowListener, ActionListener, Ite
 	}
 
 	public void onSaveComplete() {
-		if (meter != null)
-			meter.dispose();
+		if (OCRmeter != null)
+			OCRmeter.done();
 	}
 
 	class SaveOptionsDialog extends Dialog implements ActionListener, ItemListener, KeyListener {
@@ -1715,18 +1723,18 @@ public class Viewer extends Frame implements WindowListener, ActionListener, Ite
 		}
 	}
 
-	class Progressmeter extends Dialog implements DocumentWriter.OCRListener, ActionListener, KeyListener {
+	class Progressmeter extends Dialog implements ActionListener, KeyListener {
 		Label info = new Label("", Label.CENTER);
 		Button cancel = new Button("Cancel");
 		boolean cancelled = false;
-		int pages;
+		boolean done = false;
 
-		public Progressmeter(Frame parent, String title, int pages) {
-			super(parent, title, true);
+		public Progressmeter(Frame parent, String title, boolean modal, String initialText) {
+			super(parent, title, modal);
 
 			setLayout(new GridLayout(2, 1));
 
-			info.setText("Progress: Page 65535/65535: 100%");
+			info.setText(initialText);
 			add(info);
 
 			cancel.addActionListener(this);
@@ -1736,14 +1744,11 @@ public class Viewer extends Frame implements WindowListener, ActionListener, Ite
 			pack();
 			setResizable(false);
 			cancel.requestFocusInWindow();
-
-			this.pages = pages;
-			progress(-1, 0);
 		}
 
 		public void actionPerformed(ActionEvent e) {
 			if (e.getSource() == cancel)
-				cancelled = true;
+				cancel();
 		}
 
 		public void keyPressed(KeyEvent e) { }
@@ -1751,7 +1756,37 @@ public class Viewer extends Frame implements WindowListener, ActionListener, Ite
 
 		public void keyTyped(KeyEvent e) {
 			if (e.getKeyChar() == '\u001b')
-				cancelled = true;
+				cancel();
+		}
+
+		public void cancel() {
+			cancelled = true;
+		}
+
+		public void done() {
+			done = true;
+		}
+
+		public boolean progress(String text) {
+			info.setText(text);
+			return cancelled || done;
+		}
+	}
+
+	class OCRProgressmeter extends Progressmeter implements DocumentWriter.OCRListener {
+		int pages;
+
+		public OCRProgressmeter(Frame parent, String title, int pages) {
+			super(parent, title, true, "Progress: Page 65535/65535: 100%");
+			this.pages = pages;
+			progress(-1, 0);
+			setVisible(true);
+		}
+
+		public void done() {
+			super.done();
+			setVisible(false);
+			dispose();
 		}
 
 		public boolean progress(int page, int percent) {
@@ -1773,12 +1808,77 @@ public class Viewer extends Frame implements WindowListener, ActionListener, Ite
 			text.append(percent);
 			text.append("%");
 
-			info.setText(text.toString());
-
-			return cancelled;
+			return progress(text.toString());
 		}
 	}
 
+	class RenderProgressmeter extends Progressmeter {
+		Cookie cookie;
+
+		public RenderProgressmeter(Frame parent, String title, Cookie cookie, final int update) {
+			super(parent, title, false, "Progress: 100%");
+			this.cookie = cookie;
+
+			(new Thread() {
+				public void run() {
+					try {
+						int slept = 0;
+						while (!progress(slept))
+						{
+							sleep(update);
+							slept += update;
+						}
+					} catch (InterruptedException e) {
+					}
+				}
+			}).start();
+		}
+
+		public void cancel() {
+			super.cancel();
+			cookie.abort();
+		}
+
+		public boolean progress(int slept) {
+			int progress = cookie.getProgress();
+			int max = cookie.getProgressMax();
+
+			if (max <= 0 && progress < 100)
+				max = 100;
+			else if (max <= 0 && progress > 100)
+			{
+				int v = progress;
+				max = 10;
+				while (v > 10)
+				{
+					v /= 10;
+					max *= 10;
+				}
+			}
+
+			if (progress >= max)
+				done = true;
+
+			int percent = (int) ((float) progress / max * 100.0f);
+
+			StringBuilder text = new StringBuilder();
+			text.append("Progress: ");
+			text.append(percent);
+			text.append("%");
+
+			if (slept > 0)
+				setVisible(true);
+
+			if (progress(text.toString()))
+			{
+				setVisible(false);
+				dispose();
+				return true;
+			}
+
+			return false;
+		}
+	}
 
 	public static void main(String[] args) {
 		String selectedPath;
@@ -1913,6 +2013,8 @@ public class Viewer extends Frame implements WindowListener, ActionListener, Ite
 		this.linkURIs = linkURIs;
 		this.hits = hits;
 		redraw();
+		if (renderMeter != null)
+			renderMeter.done();
 	}
 	public void onSearchStart(Location startPage, Location finalPage, int direction, String needle) {
 		searchField.setEnabled(false);
