@@ -1,4 +1,4 @@
-// Copyright (C) 2004-2024 Artifex Software, Inc.
+// Copyright (C) 2004-2025 Artifex Software, Inc.
 //
 // This file is part of MuPDF.
 //
@@ -124,6 +124,20 @@ FUN(PDFDocument_finalize)(JNIEnv *env, jobject self)
 	FUN(Document_finalize)(env, self); /* Call super.finalize() */
 }
 
+JNIEXPORT void JNICALL
+FUN(PDFDocument_check)(JNIEnv *env, jobject self)
+{
+	fz_context *ctx = get_context(env);
+	pdf_document *pdf = from_PDFDocument(env, self);
+
+	if (!ctx || !pdf) return;
+
+	fz_try(ctx)
+		pdf_check_document(ctx, pdf);
+	fz_catch(ctx)
+		jni_rethrow_void(env, ctx);
+}
+
 JNIEXPORT jint JNICALL
 FUN(PDFDocument_countObjects)(JNIEnv *env, jobject self)
 {
@@ -1321,10 +1335,16 @@ FUN(PDFDocument_canUndo)(JNIEnv *env, jobject self)
 {
 	fz_context *ctx = get_context(env);
 	pdf_document *pdf = from_PDFDocument(env, self);
+	jboolean can = JNI_FALSE;
 
 	if (!ctx || !pdf) return JNI_FALSE;
 
-	return pdf_can_undo(ctx, pdf);
+	fz_try(ctx)
+		can = pdf_can_undo(ctx, pdf);
+	fz_catch(ctx)
+		jni_rethrow(env, ctx);
+
+	return can;
 }
 
 JNIEXPORT jboolean JNICALL
@@ -1332,10 +1352,16 @@ FUN(PDFDocument_canRedo)(JNIEnv *env, jobject self)
 {
 	fz_context *ctx = get_context(env);
 	pdf_document *pdf = from_PDFDocument(env, self);
+	jboolean can = JNI_FALSE;
 
 	if (!ctx || !pdf) return JNI_FALSE;
 
-	return pdf_can_redo(ctx, pdf);
+	fz_try(ctx)
+		can = pdf_can_redo(ctx, pdf);
+	fz_catch(ctx)
+		jni_rethrow(env, ctx);
+
+	return can;
 }
 
 JNIEXPORT void JNICALL
@@ -1527,23 +1553,17 @@ FUN(PDFDocument_addEmbeddedFile)(JNIEnv *env, jobject self, jstring jfilename, j
 	return to_PDFObject_safe(ctx, env, fs);
 }
 
-JNIEXPORT jstring JNICALL
-FUN(PDFDocument_getEmbeddedFileParams)(JNIEnv *env, jobject self, jobject jfs)
-{
-	return FUN(PDFDocument_getFilespecParams)(env, self, jfs);
-}
-
-JNIEXPORT jstring JNICALL
+JNIEXPORT jobject JNICALL
 FUN(PDFDocument_getFilespecParams)(JNIEnv *env, jobject self, jobject jfs)
 {
 	fz_context *ctx = get_context(env);
 	pdf_obj *fs = from_PDFObject_safe(env, jfs);
-	pdf_embedded_file_params params;
+	pdf_filespec_params params;
 	jstring jfilename = NULL;
 	jstring jmimetype = NULL;
 
 	fz_try(ctx)
-		pdf_get_embedded_file_params(ctx, fs, &params);
+		pdf_get_filespec_params(ctx, fs, &params);
 	fz_catch(ctx)
 		jni_rethrow(env, ctx);
 
@@ -1818,7 +1838,7 @@ FUN(PDFDocument_rearrangePages)(JNIEnv *env, jobject self, jobject jpages)
 	}
 
 	fz_try(ctx)
-		pdf_rearrange_pages(ctx, pdf, len, pages);
+		pdf_rearrange_pages(ctx, pdf, len, pages, PDF_CLEAN_STRUCTURE_DROP);
 	fz_always(ctx)
 		fz_free(ctx, pages);
 	fz_catch(ctx)
@@ -1832,6 +1852,8 @@ FUN(PDFDocument_countAssociatedFiles)(JNIEnv *env, jobject self)
 	pdf_document *doc = from_PDFDocument(env, self);
 	int n;
 
+	if (!ctx) return 0;
+
 	fz_try(ctx)
 		n = pdf_count_document_associated_files(ctx, doc);
 	fz_catch(ctx)
@@ -1847,6 +1869,8 @@ FUN(PDFDocument_associatedFile)(JNIEnv *env, jobject self, jint idx)
 	pdf_document *doc = from_PDFDocument(env, self);
 	pdf_obj *af;
 
+	if (!ctx) return NULL;
+
 	fz_try(ctx)
 		af = pdf_document_associated_file(ctx, doc, idx);
 	fz_catch(ctx)
@@ -1863,6 +1887,8 @@ FUN(PDFDocument_zugferdProfile)(JNIEnv *env, jobject self)
 	enum pdf_zugferd_profile p;
 	float version;
 
+	if (!ctx) return 0;
+
 	fz_try(ctx)
 		p = pdf_zugferd_profile(ctx, doc, &version);
 	fz_catch(ctx)
@@ -1878,6 +1904,8 @@ FUN(PDFDocument_zugferdVersion)(JNIEnv *env, jobject self)
 	pdf_document *doc = from_PDFDocument(env, self);
 	float version;
 
+	if (!ctx) return 0;
+
 	fz_try(ctx)
 		(void) pdf_zugferd_profile(ctx, doc, &version);
 	fz_catch(ctx)
@@ -1893,6 +1921,8 @@ FUN(PDFDocument_zugferdXML)(JNIEnv *env, jobject self)
 	pdf_document *doc = from_PDFDocument(env, self);
 	fz_buffer *buf;
 
+	if (!ctx) return NULL;
+
 	fz_try(ctx)
 		buf = pdf_zugferd_xml(ctx, doc);
 	fz_catch(ctx)
@@ -1900,3 +1930,223 @@ FUN(PDFDocument_zugferdXML)(JNIEnv *env, jobject self)
 
 	return to_Buffer_safe_own(ctx, env, buf);
 }
+
+JNIEXPORT jobject JNICALL
+FUN(PDFDocument_loadImage)(JNIEnv *env, jobject self, jobject jobj)
+{
+	fz_context *ctx = get_context(env);
+	pdf_document *doc = from_PDFDocument(env, self);
+	pdf_obj *obj = from_PDFObject(env, jobj);
+	fz_image *img;
+
+	if (!ctx) return NULL;
+
+	fz_try(ctx)
+		img = pdf_load_image(ctx, doc, obj);
+	fz_catch(ctx)
+		jni_rethrow(env, ctx);
+
+	return to_Image_safe_own(ctx, env, img);
+}
+
+JNIEXPORT jobject JNICALL
+FUN(PDFDocument_lookupDest)(JNIEnv *env, jobject self, jobject jdest)
+{
+	fz_context *ctx = get_context(env);
+	pdf_document *doc = from_PDFDocument(env, self);
+	pdf_obj *dest = from_PDFObject(env, jdest);
+	pdf_obj *obj = NULL;
+
+	if (!ctx) return NULL;
+
+	fz_try(ctx)
+		obj = pdf_lookup_dest(ctx, doc, dest);
+	fz_catch(ctx)
+		jni_rethrow(env, ctx);
+
+	return to_PDFObject_safe_own(ctx, env, obj);
+}
+
+JNIEXPORT jint JNICALL
+FUN(PDFDocument_countLayerConfigs)(JNIEnv *env, jobject self)
+{
+	fz_context *ctx = get_context(env);
+	pdf_document *doc = from_PDFDocument(env, self);
+	jint configs = 0;
+
+	fz_try(ctx)
+		configs = pdf_count_layer_configs(ctx, doc);
+	fz_catch(ctx)
+		jni_rethrow(env, ctx);
+
+	return configs;
+}
+
+JNIEXPORT jstring JNICALL
+FUN(PDFDocument_getLayerConfigName)(JNIEnv *env, jobject self, jint config)
+{
+	fz_context *ctx = get_context(env);
+	pdf_document *doc = from_PDFDocument(env, self);
+	const char *name = NULL;
+
+	fz_try(ctx)
+		name = pdf_layer_config_name(ctx, doc, config);
+	fz_catch(ctx)
+		jni_rethrow(env, ctx);
+
+	return (*env)->NewStringUTF(env, name);
+}
+
+JNIEXPORT jstring JNICALL
+FUN(PDFDocument_getLayerConfigCreator)(JNIEnv *env, jobject self, jint config)
+{
+	fz_context *ctx = get_context(env);
+	pdf_document *doc = from_PDFDocument(env, self);
+	const char *creator = NULL;
+
+	fz_try(ctx)
+		creator = pdf_layer_config_creator(ctx, doc, config);
+	fz_catch(ctx)
+		jni_rethrow(env, ctx);
+
+	return (*env)->NewStringUTF(env, creator);
+}
+
+JNIEXPORT void JNICALL
+FUN(PDFDocument_selectLayerConfig)(JNIEnv *env, jobject self, jint config)
+{
+	fz_context *ctx = get_context(env);
+	pdf_document *doc = from_PDFDocument(env, self);
+
+	fz_try(ctx)
+		pdf_select_layer_config(ctx, doc, config);
+	fz_catch(ctx)
+		jni_rethrow_void(env, ctx);
+}
+
+JNIEXPORT jint JNICALL
+FUN(PDFDocument_countLayerConfigUIs)(JNIEnv *env, jobject self)
+{
+	fz_context *ctx = get_context(env);
+	pdf_document *doc = from_PDFDocument(env, self);
+	jint configuis = 0;
+
+	fz_try(ctx)
+		configuis = pdf_count_layer_config_ui(ctx, doc);
+	fz_catch(ctx)
+		jni_rethrow(env, ctx);
+
+	return configuis;
+}
+
+JNIEXPORT jobject JNICALL
+FUN(PDFDocument_getLayerConfigUIInfo)(JNIEnv *env, jobject self, jint configui)
+{
+	fz_context *ctx = get_context(env);
+	pdf_document *doc = from_PDFDocument(env, self);
+	pdf_layer_config_ui info = { 0 };
+	jobject jinfo = NULL;
+	jobject jtext = NULL;
+
+	fz_try(ctx)
+		pdf_layer_config_ui_info(ctx, doc, configui, &info);
+	fz_catch(ctx)
+		jni_rethrow(env, ctx);
+
+	jtext = (*env)->NewStringUTF(env, info.text);
+	if (!jtext || (*env)->ExceptionCheck(env))
+		return NULL;
+
+	jinfo = (*env)->NewObject(env, cls_PDFDocument_LayerConfigUIInfo, mid_PDFDocument_LayerConfigUIInfo_init);
+	if (!jinfo || (*env)->ExceptionCheck(env))
+		return NULL;
+
+	(*env)->SetIntField(env, jinfo, fid_PDFDocument_LayerConfigUIInfo_type, info.type);
+	(*env)->SetIntField(env, jinfo, fid_PDFDocument_LayerConfigUIInfo_depth, info.depth);
+	(*env)->SetBooleanField(env, jinfo, fid_PDFDocument_LayerConfigUIInfo_selected, info.selected);
+	(*env)->SetBooleanField(env, jinfo, fid_PDFDocument_LayerConfigUIInfo_locked, info.locked);
+	(*env)->SetObjectField(env, jinfo, fid_PDFDocument_LayerConfigUIInfo_text, jtext);
+
+	return jinfo;
+}
+
+JNIEXPORT jint JNICALL
+FUN(PDFDocument_countLayers)(JNIEnv *env, jobject self)
+{
+	fz_context *ctx = get_context(env);
+	pdf_document *doc = from_PDFDocument(env, self);
+	jint layers = 0;
+
+	fz_try(ctx)
+		layers = pdf_count_layers(ctx, doc);
+	fz_catch(ctx)
+		jni_rethrow(env, ctx);
+
+	return layers;
+}
+
+JNIEXPORT jboolean JNICALL
+FUN(PDFDocument_isLayerVisible)(JNIEnv *env, jobject self, jint layer)
+{
+	fz_context *ctx = get_context(env);
+	pdf_document *doc = from_PDFDocument(env, self);
+	jboolean visible = JNI_FALSE;
+
+	fz_try(ctx)
+		visible = pdf_layer_is_enabled(ctx, doc, layer);
+	fz_catch(ctx)
+		jni_rethrow(env, ctx);
+
+	return visible;
+}
+
+JNIEXPORT void JNICALL
+FUN(PDFDocument_setLayerVisible)(JNIEnv *env, jobject self, jint layer, jboolean visible)
+{
+	fz_context *ctx = get_context(env);
+	pdf_document *doc = from_PDFDocument(env, self);
+
+	fz_try(ctx)
+		pdf_enable_layer(ctx, doc, layer, visible);
+	fz_catch(ctx)
+		jni_rethrow_void(env, ctx);
+}
+
+JNIEXPORT jstring JNICALL
+FUN(PDFDocument_getLayerName)(JNIEnv *env, jobject self, jint layer)
+{
+	fz_context *ctx = get_context(env);
+	pdf_document *doc = from_PDFDocument(env, self);
+	const char *name = NULL;
+
+	fz_try(ctx)
+		name = pdf_layer_name(ctx, doc, layer);
+	fz_catch(ctx)
+		jni_rethrow(env, ctx);
+
+	return (*env)->NewStringUTF(env, name);
+}
+
+JNIEXPORT void JNICALL
+FUN(PDFDocument_subsetFonts)(JNIEnv *env, jobject self)
+{
+	fz_context *ctx = get_context(env);
+	pdf_document *doc = from_PDFDocument(env, self);
+
+	fz_try(ctx)
+		pdf_subset_fonts(ctx, doc, 0, NULL);
+	fz_catch(ctx)
+		jni_rethrow_void(env, ctx);
+}
+
+JNIEXPORT void JNICALL
+FUN(PDFDocument_bake)(JNIEnv *env, jobject self, jboolean bake_annots, jboolean bake_widgets)
+{
+	fz_context *ctx = get_context(env);
+	pdf_document *doc = from_PDFDocument(env, self);
+
+	fz_try(ctx)
+		pdf_bake_document(ctx, doc, bake_annots, bake_widgets);
+	fz_catch(ctx)
+		jni_rethrow_void(env, ctx);
+}
