@@ -1,4 +1,4 @@
-// Copyright (C) 2004-2024 Artifex Software, Inc.
+// Copyright (C) 2004-2025 Artifex Software, Inc.
 //
 // This file is part of MuPDF.
 //
@@ -60,10 +60,14 @@ public class PDFDocument extends Document
 		super(newNative());
 	}
 
+	public native void check();
+
 	public boolean isPDF() {
 		return true;
 	}
 
+	public native void setPageTreeCache(boolean enabled);
+
 	public native PDFObject findPage(int at);
 
 	public native PDFObject getTrailer();
@@ -245,7 +249,6 @@ public class PDFDocument extends Document
 	public native int countSignatures();
 
 	public native PDFObject addEmbeddedFile(String filename, String mimetype, Buffer contents, long created, long modified, boolean addChecksum);
-	public native PDFEmbeddedFileParams getEmbeddedFileParams(PDFObject fs);
 	public native PDFFilespecParams getFilespecParams(PDFObject fs);
 	public native Buffer loadEmbeddedFileContents(PDFObject fs);
 	public native boolean verifyEmbeddedFileChecksum(PDFObject fs);
@@ -281,6 +284,31 @@ public class PDFDocument extends Document
 		}
 	}
 
+	public static final int LAYER_UI_LABEL = 0;
+	public static final int LAYER_UI_CHECKBOX = 1;
+	public static final int LAYER_UI_RADIOBOX = 2;
+
+	public static class LayerConfigUIInfo {
+		public int type;
+		public int depth;
+		public boolean selected;
+		public boolean locked;
+		public String text;
+	}
+
+	public native int countLayerConfigs();
+	public native String getLayerConfigName(int config);
+	public native String getLayerConfigCreator(int config);
+	public native void selectLayerConfig(int config);
+
+	public native int countLayerConfigUIs();
+	public native LayerConfigUIInfo getLayerConfigUIInfo(int configui);
+
+	public native int countLayers();
+	public native boolean isLayerVisible(int layer);
+	public native void setLayerVisible(int layer, boolean visible);
+	public native String getLayerName(int layer);
+
 	public native int countAssociatedFiles();
 	public native PDFObject associatedFile(int idx);
 
@@ -296,4 +324,11 @@ public class PDFDocument extends Document
 	public native int zugferdProfile();
 	public native float zugferdVersion();
 	public native Buffer zugferdXML();
+
+	public native Image loadImage(PDFObject imgobj);
+	public native PDFObject lookupDest(PDFObject destination);
+
+	public native void subsetFonts();
+
+	public native void bake(boolean bakeAnnots, boolean bakeWidgets);
 }
