@@ -1,4 +1,4 @@
-// Copyright (C) 2004-2024 Artifex Software, Inc.
+// Copyright (C) 2004-2025 Artifex Software, Inc.
 //
 // This file is part of MuPDF.
 //
@@ -39,6 +39,7 @@
 #include <sys/wait.h>
 #include <unistd.h>
 #include <signal.h>
+#include <limits.h>
 
 #define mupdf_icon_bitmap_16_width 16
 #define mupdf_icon_bitmap_16_height 16
@@ -536,6 +537,20 @@ static void winblitstatusbar(pdfapp_t *app)
 	}
 }
 
+int winisresolutionacceptable(pdfapp_t *app, fz_matrix ctm)
+{
+	fz_rect bounds;
+	fz_irect ibounds;
+	int w, h;
+
+	bounds = fz_transform_rect(app->page_bbox, ctm);
+	ibounds = fz_round_rect(bounds);
+	w = ibounds.x1 - ibounds.x0;
+	h = ibounds.y1 - ibounds.y0;
+
+	return w < (INT_MAX / 4) / h;
+}
+
 static void winblit(pdfapp_t *app)
 {
 	if (gapp.image)
@@ -876,6 +891,17 @@ static void usage(const char *argv0)
 	fprintf(stderr, "\t-S -\tfont size for EPUB layout\n");
 	fprintf(stderr, "\t-U -\tuser style sheet for EPUB layout\n");
 	fprintf(stderr, "\t-X\tdisable document styles for EPUB layout\n");
+	fprintf(stderr, "\t-c -\tICC output profile\n");
+#ifdef HAVE_CURL
+	fprintf(stderr, "\t-b -\temulate progressive loading (kbps)\n");
+#endif
+	fprintf(stderr, "\t-v\tshow version\n");
+	exit(1);
+}
+
+static void version(void)
+{
+	fprintf(stderr, "mupdf-x11 version %s\n", FZ_VERSION);
 	exit(1);
 }
 
@@ -908,7 +934,7 @@ int main(int argc, char **argv)
 
 	pdfapp_init(ctx, &gapp);
 
-	while ((c = fz_getopt(argc, argv, "Ip:r:A:C:W:H:S:U:Xb:c:")) != -1)
+	while ((c = fz_getopt(argc, argv, "Ip:r:A:C:W:H:S:U:Xb:c:v")) != -1)
 	{
 		switch (c)
 		{
@@ -928,7 +954,8 @@ int main(int argc, char **argv)
 		case 'U': gapp.layout_css = fz_optarg; break;
 		case 'X': gapp.layout_use_doc_css = 0; break;
 		case 'b': kbps = fz_atoi(fz_optarg); break;
-		default: usage(argv[0]);
+		case 'v': version(); break;
+		default: usage(argv[0]); break;
 		}
 	}
 
