@@ -1,3 +1,9 @@
+#ifdef _WIN32
+#define TESSERACT_ENDIAN_DETECT 1
+#else
+#define TESSERACT_ENDIAN_DETECT 0
+#endif
+
 #if defined(__BYTE_ORDER) && __BYTE_ORDER == __BIG_ENDIAN ||	\
 	defined(__BIG_ENDIAN__) ||				\
 	defined(__ARMEB__) ||					\
@@ -13,7 +19,8 @@
 	defined(__AARCH64EL__) ||					\
 	defined(_MIPSEL) || defined(__MIPSEL) || defined(__MIPSEL__) ||	\
 	defined(_M_IX86) || defined(_M_X64) ||				\
-	defined(_M_IS64) || defined(_M_ARM)
+	defined(_M_IS64) || defined(_M_ARM) || \
+	TESSERACT_ENDIAN_DETECT == 1
 // It's a little-endian target architecture
 #   define L_LITTLE_ENDIAN
 #else
