@@ -172,6 +172,273 @@ static fz_bidi_chartype class_from_ch_n(uint32_t ch)
 	return from_ch_ws;
 }
 
+static const unsigned char ucdn_script_from_block_table[256] = {
+	UCDN_SCRIPT_LATIN, /* U+0000 */
+	UCDN_SCRIPT_LATIN, /* U+0100 */
+	UCDN_SCRIPT_LATIN, /* U+0200 */
+	UCDN_SCRIPT_GREEK, /* U+0300 */
+	UCDN_SCRIPT_CYRILLIC, /* U+0400 */
+	UCDN_SCRIPT_ARMENIAN, /* U+0500 */
+	UCDN_SCRIPT_ARABIC, /* U+0600 */
+	UCDN_SCRIPT_SYRIAC, /* U+0700 */
+	UCDN_SCRIPT_ARABIC, /* U+0800 */
+	UCDN_SCRIPT_DEVANAGARI, /* U+0900 */
+	UCDN_SCRIPT_GUJARATI, /* U+0A00 */
+	UCDN_SCRIPT_ORIYA, /* U+0B00 */
+	UCDN_SCRIPT_TELUGU, /* U+0C00 */
+	UCDN_SCRIPT_MALAYALAM, /* U+0D00 */
+	UCDN_SCRIPT_THAI, /* U+0E00 */
+	UCDN_SCRIPT_TIBETAN, /* U+0F00 */
+	UCDN_SCRIPT_MYANMAR, /* U+1000 */
+	UCDN_SCRIPT_HANGUL, /* U+1100 */
+	UCDN_SCRIPT_ETHIOPIC, /* U+1200 */
+	UCDN_SCRIPT_ETHIOPIC, /* U+1300 */
+	UCDN_SCRIPT_CANADIAN_ABORIGINAL, /* U+1400 */
+	UCDN_SCRIPT_CANADIAN_ABORIGINAL, /* U+1500 */
+	UCDN_SCRIPT_CANADIAN_ABORIGINAL, /* U+1600 */
+	UCDN_SCRIPT_KHMER, /* U+1700 */
+	UCDN_SCRIPT_MONGOLIAN, /* U+1800 */
+	UCDN_SCRIPT_NEW_TAI_LUE, /* U+1900 */
+	UCDN_SCRIPT_TAI_THAM, /* U+1A00 */
+	UCDN_SCRIPT_BALINESE, /* U+1B00 */
+	UCDN_SCRIPT_LEPCHA, /* U+1C00 */
+	UCDN_SCRIPT_LATIN, /* U+1D00 */
+	UCDN_SCRIPT_LATIN, /* U+1E00 */
+	UCDN_SCRIPT_GREEK, /* U+1F00 */
+	UCDN_SCRIPT_COMMON, /* U+2000 */
+	UCDN_SCRIPT_LATIN, /* U+2100 */
+	UCDN_SCRIPT_COMMON, /* U+2200 */
+	UCDN_SCRIPT_COMMON, /* U+2300 */
+	UCDN_SCRIPT_COMMON, /* U+2400 */
+	UCDN_SCRIPT_COMMON, /* U+2500 */
+	UCDN_SCRIPT_COMMON, /* U+2600 */
+	UCDN_SCRIPT_COMMON, /* U+2700 */
+	UCDN_SCRIPT_BRAILLE, /* U+2800 */
+	UCDN_SCRIPT_COMMON, /* U+2900 */
+	UCDN_SCRIPT_COMMON, /* U+2A00 */
+	UCDN_SCRIPT_COMMON, /* U+2B00 */
+	UCDN_SCRIPT_COPTIC, /* U+2C00 */
+	UCDN_SCRIPT_ETHIOPIC, /* U+2D00 */
+	UCDN_SCRIPT_HAN, /* U+2E00 */
+	UCDN_SCRIPT_HAN, /* U+2F00 */
+	UCDN_SCRIPT_KATAKANA, /* U+3000 */
+	UCDN_SCRIPT_HANGUL, /* U+3100 */
+	UCDN_SCRIPT_HANGUL, /* U+3200 */
+	UCDN_SCRIPT_KATAKANA, /* U+3300 */
+	UCDN_SCRIPT_HAN, /* U+3400 */
+	UCDN_SCRIPT_HAN, /* U+3500 */
+	UCDN_SCRIPT_HAN, /* U+3600 */
+	UCDN_SCRIPT_HAN, /* U+3700 */
+	UCDN_SCRIPT_HAN, /* U+3800 */
+	UCDN_SCRIPT_HAN, /* U+3900 */
+	UCDN_SCRIPT_HAN, /* U+3A00 */
+	UCDN_SCRIPT_HAN, /* U+3B00 */
+	UCDN_SCRIPT_HAN, /* U+3C00 */
+	UCDN_SCRIPT_HAN, /* U+3D00 */
+	UCDN_SCRIPT_HAN, /* U+3E00 */
+	UCDN_SCRIPT_HAN, /* U+3F00 */
+	UCDN_SCRIPT_HAN, /* U+4000 */
+	UCDN_SCRIPT_HAN, /* U+4100 */
+	UCDN_SCRIPT_HAN, /* U+4200 */
+	UCDN_SCRIPT_HAN, /* U+4300 */
+	UCDN_SCRIPT_HAN, /* U+4400 */
+	UCDN_SCRIPT_HAN, /* U+4500 */
+	UCDN_SCRIPT_HAN, /* U+4600 */
+	UCDN_SCRIPT_HAN, /* U+4700 */
+	UCDN_SCRIPT_HAN, /* U+4800 */
+	UCDN_SCRIPT_HAN, /* U+4900 */
+	UCDN_SCRIPT_HAN, /* U+4A00 */
+	UCDN_SCRIPT_HAN, /* U+4B00 */
+	UCDN_SCRIPT_HAN, /* U+4C00 */
+	UCDN_SCRIPT_HAN, /* U+4D00 */
+	UCDN_SCRIPT_HAN, /* U+4E00 */
+	UCDN_SCRIPT_HAN, /* U+4F00 */
+	UCDN_SCRIPT_HAN, /* U+5000 */
+	UCDN_SCRIPT_HAN, /* U+5100 */
+	UCDN_SCRIPT_HAN, /* U+5200 */
+	UCDN_SCRIPT_HAN, /* U+5300 */
+	UCDN_SCRIPT_HAN, /* U+5400 */
+	UCDN_SCRIPT_HAN, /* U+5500 */
+	UCDN_SCRIPT_HAN, /* U+5600 */
+	UCDN_SCRIPT_HAN, /* U+5700 */
+	UCDN_SCRIPT_HAN, /* U+5800 */
+	UCDN_SCRIPT_HAN, /* U+5900 */
+	UCDN_SCRIPT_HAN, /* U+5A00 */
+	UCDN_SCRIPT_HAN, /* U+5B00 */
+	UCDN_SCRIPT_HAN, /* U+5C00 */
+	UCDN_SCRIPT_HAN, /* U+5D00 */
+	UCDN_SCRIPT_HAN, /* U+5E00 */
+	UCDN_SCRIPT_HAN, /* U+5F00 */
+	UCDN_SCRIPT_HAN, /* U+6000 */
+	UCDN_SCRIPT_HAN, /* U+6100 */
+	UCDN_SCRIPT_HAN, /* U+6200 */
+	UCDN_SCRIPT_HAN, /* U+6300 */
+	UCDN_SCRIPT_HAN, /* U+6400 */
+	UCDN_SCRIPT_HAN, /* U+6500 */
+	UCDN_SCRIPT_HAN, /* U+6600 */
+	UCDN_SCRIPT_HAN, /* U+6700 */
+	UCDN_SCRIPT_HAN, /* U+6800 */
+	UCDN_SCRIPT_HAN, /* U+6900 */
+	UCDN_SCRIPT_HAN, /* U+6A00 */
+	UCDN_SCRIPT_HAN, /* U+6B00 */
+	UCDN_SCRIPT_HAN, /* U+6C00 */
+	UCDN_SCRIPT_HAN, /* U+6D00 */
+	UCDN_SCRIPT_HAN, /* U+6E00 */
+	UCDN_SCRIPT_HAN, /* U+6F00 */
+	UCDN_SCRIPT_HAN, /* U+7000 */
+	UCDN_SCRIPT_HAN, /* U+7100 */
+	UCDN_SCRIPT_HAN, /* U+7200 */
+	UCDN_SCRIPT_HAN, /* U+7300 */
+	UCDN_SCRIPT_HAN, /* U+7400 */
+	UCDN_SCRIPT_HAN, /* U+7500 */
+	UCDN_SCRIPT_HAN, /* U+7600 */
+	UCDN_SCRIPT_HAN, /* U+7700 */
+	UCDN_SCRIPT_HAN, /* U+7800 */
+	UCDN_SCRIPT_HAN, /* U+7900 */
+	UCDN_SCRIPT_HAN, /* U+7A00 */
+	UCDN_SCRIPT_HAN, /* U+7B00 */
+	UCDN_SCRIPT_HAN, /* U+7C00 */
+	UCDN_SCRIPT_HAN, /* U+7D00 */
+	UCDN_SCRIPT_HAN, /* U+7E00 */
+	UCDN_SCRIPT_HAN, /* U+7F00 */
+	UCDN_SCRIPT_HAN, /* U+8000 */
+	UCDN_SCRIPT_HAN, /* U+8100 */
+	UCDN_SCRIPT_HAN, /* U+8200 */
+	UCDN_SCRIPT_HAN, /* U+8300 */
+	UCDN_SCRIPT_HAN, /* U+8400 */
+	UCDN_SCRIPT_HAN, /* U+8500 */
+	UCDN_SCRIPT_HAN, /* U+8600 */
+	UCDN_SCRIPT_HAN, /* U+8700 */
+	UCDN_SCRIPT_HAN, /* U+8800 */
+	UCDN_SCRIPT_HAN, /* U+8900 */
+	UCDN_SCRIPT_HAN, /* U+8A00 */
+	UCDN_SCRIPT_HAN, /* U+8B00 */
+	UCDN_SCRIPT_HAN, /* U+8C00 */
+	UCDN_SCRIPT_HAN, /* U+8D00 */
+	UCDN_SCRIPT_HAN, /* U+8E00 */
+	UCDN_SCRIPT_HAN, /* U+8F00 */
+	UCDN_SCRIPT_HAN, /* U+9000 */
+	UCDN_SCRIPT_HAN, /* U+9100 */
+	UCDN_SCRIPT_HAN, /* U+9200 */
+	UCDN_SCRIPT_HAN, /* U+9300 */
+	UCDN_SCRIPT_HAN, /* U+9400 */
+	UCDN_SCRIPT_HAN, /* U+9500 */
+	UCDN_SCRIPT_HAN, /* U+9600 */
+	UCDN_SCRIPT_HAN, /* U+9700 */
+	UCDN_SCRIPT_HAN, /* U+9800 */
+	UCDN_SCRIPT_HAN, /* U+9900 */
+	UCDN_SCRIPT_HAN, /* U+9A00 */
+	UCDN_SCRIPT_HAN, /* U+9B00 */
+	UCDN_SCRIPT_HAN, /* U+9C00 */
+	UCDN_SCRIPT_HAN, /* U+9D00 */
+	UCDN_SCRIPT_HAN, /* U+9E00 */
+	UCDN_SCRIPT_HAN, /* U+9F00 */
+	UCDN_SCRIPT_YI, /* U+A000 */
+	UCDN_SCRIPT_YI, /* U+A100 */
+	UCDN_SCRIPT_YI, /* U+A200 */
+	UCDN_SCRIPT_YI, /* U+A300 */
+	UCDN_SCRIPT_YI, /* U+A400 */
+	UCDN_SCRIPT_VAI, /* U+A500 */
+	UCDN_SCRIPT_CYRILLIC, /* U+A600 */
+	UCDN_SCRIPT_LATIN, /* U+A700 */
+	UCDN_SCRIPT_SAURASHTRA, /* U+A800 */
+	UCDN_SCRIPT_JAVANESE, /* U+A900 */
+	UCDN_SCRIPT_CHAM, /* U+AA00 */
+	UCDN_SCRIPT_CHEROKEE, /* U+AB00 */
+	UCDN_SCRIPT_HANGUL, /* U+AC00 */
+	UCDN_SCRIPT_HANGUL, /* U+AD00 */
+	UCDN_SCRIPT_HANGUL, /* U+AE00 */
+	UCDN_SCRIPT_HANGUL, /* U+AF00 */
+	UCDN_SCRIPT_HANGUL, /* U+B000 */
+	UCDN_SCRIPT_HANGUL, /* U+B100 */
+	UCDN_SCRIPT_HANGUL, /* U+B200 */
+	UCDN_SCRIPT_HANGUL, /* U+B300 */
+	UCDN_SCRIPT_HANGUL, /* U+B400 */
+	UCDN_SCRIPT_HANGUL, /* U+B500 */
+	UCDN_SCRIPT_HANGUL, /* U+B600 */
+	UCDN_SCRIPT_HANGUL, /* U+B700 */
+	UCDN_SCRIPT_HANGUL, /* U+B800 */
+	UCDN_SCRIPT_HANGUL, /* U+B900 */
+	UCDN_SCRIPT_HANGUL, /* U+BA00 */
+	UCDN_SCRIPT_HANGUL, /* U+BB00 */
+	UCDN_SCRIPT_HANGUL, /* U+BC00 */
+	UCDN_SCRIPT_HANGUL, /* U+BD00 */
+	UCDN_SCRIPT_HANGUL, /* U+BE00 */
+	UCDN_SCRIPT_HANGUL, /* U+BF00 */
+	UCDN_SCRIPT_HANGUL, /* U+C000 */
+	UCDN_SCRIPT_HANGUL, /* U+C100 */
+	UCDN_SCRIPT_HANGUL, /* U+C200 */
+	UCDN_SCRIPT_HANGUL, /* U+C300 */
+	UCDN_SCRIPT_HANGUL, /* U+C400 */
+	UCDN_SCRIPT_HANGUL, /* U+C500 */
+	UCDN_SCRIPT_HANGUL, /* U+C600 */
+	UCDN_SCRIPT_HANGUL, /* U+C700 */
+	UCDN_SCRIPT_HANGUL, /* U+C800 */
+	UCDN_SCRIPT_HANGUL, /* U+C900 */
+	UCDN_SCRIPT_HANGUL, /* U+CA00 */
+	UCDN_SCRIPT_HANGUL, /* U+CB00 */
+	UCDN_SCRIPT_HANGUL, /* U+CC00 */
+	UCDN_SCRIPT_HANGUL, /* U+CD00 */
+	UCDN_SCRIPT_HANGUL, /* U+CE00 */
+	UCDN_SCRIPT_HANGUL, /* U+CF00 */
+	UCDN_SCRIPT_HANGUL, /* U+D000 */
+	UCDN_SCRIPT_HANGUL, /* U+D100 */
+	UCDN_SCRIPT_HANGUL, /* U+D200 */
+	UCDN_SCRIPT_HANGUL, /* U+D300 */
+	UCDN_SCRIPT_HANGUL, /* U+D400 */
+	UCDN_SCRIPT_HANGUL, /* U+D500 */
+	UCDN_SCRIPT_HANGUL, /* U+D600 */
+	UCDN_SCRIPT_HANGUL, /* U+D700 */
+	UCDN_SCRIPT_COMMON, /* U+D800 */
+	UCDN_SCRIPT_COMMON, /* U+D900 */
+	UCDN_SCRIPT_COMMON, /* U+DA00 */
+	UCDN_SCRIPT_COMMON, /* U+DB00 */
+	UCDN_SCRIPT_COMMON, /* U+DC00 */
+	UCDN_SCRIPT_COMMON, /* U+DD00 */
+	UCDN_SCRIPT_COMMON, /* U+DE00 */
+	UCDN_SCRIPT_COMMON, /* U+DF00 */
+	UCDN_SCRIPT_COMMON, /* U+E000 */
+	UCDN_SCRIPT_COMMON, /* U+E100 */
+	UCDN_SCRIPT_COMMON, /* U+E200 */
+	UCDN_SCRIPT_COMMON, /* U+E300 */
+	UCDN_SCRIPT_COMMON, /* U+E400 */
+	UCDN_SCRIPT_COMMON, /* U+E500 */
+	UCDN_SCRIPT_COMMON, /* U+E600 */
+	UCDN_SCRIPT_COMMON, /* U+E700 */
+	UCDN_SCRIPT_COMMON, /* U+E800 */
+	UCDN_SCRIPT_COMMON, /* U+E900 */
+	UCDN_SCRIPT_COMMON, /* U+EA00 */
+	UCDN_SCRIPT_COMMON, /* U+EB00 */
+	UCDN_SCRIPT_COMMON, /* U+EC00 */
+	UCDN_SCRIPT_COMMON, /* U+ED00 */
+	UCDN_SCRIPT_COMMON, /* U+EE00 */
+	UCDN_SCRIPT_COMMON, /* U+EF00 */
+	UCDN_SCRIPT_COMMON, /* U+F000 */
+	UCDN_SCRIPT_COMMON, /* U+F100 */
+	UCDN_SCRIPT_COMMON, /* U+F200 */
+	UCDN_SCRIPT_COMMON, /* U+F300 */
+	UCDN_SCRIPT_COMMON, /* U+F400 */
+	UCDN_SCRIPT_COMMON, /* U+F500 */
+	UCDN_SCRIPT_COMMON, /* U+F600 */
+	UCDN_SCRIPT_COMMON, /* U+F700 */
+	UCDN_SCRIPT_COMMON, /* U+F800 */
+	UCDN_SCRIPT_HAN, /* U+F900 */
+	UCDN_SCRIPT_HAN, /* U+FA00 */
+	UCDN_SCRIPT_ARABIC, /* U+FB00 */
+	UCDN_SCRIPT_ARABIC, /* U+FC00 */
+	UCDN_SCRIPT_ARABIC, /* U+FD00 */
+	UCDN_SCRIPT_ARABIC, /* U+FE00 */
+	UCDN_SCRIPT_KATAKANA, /* U+FF00 */
+};
+
+static int
+guess_script_from_block(int c)
+{
+	if (c < 0x10000)
+		return ucdn_script_from_block_table[c >> 8];
+	return UCDN_SCRIPT_COMMON;
+}
+
 /* Split fragments into single scripts (or punctuation + single script) */
 static void
 split_at_script(const uint32_t *fragment,
@@ -180,6 +447,7 @@ split_at_script(const uint32_t *fragment,
 		void *arg,
 		fz_bidi_fragment_fn *callback)
 {
+	int script_guess = UCDN_SCRIPT_COMMON;
 	int script = UCDN_SCRIPT_COMMON;
 	size_t script_start, i;
 
@@ -187,15 +455,18 @@ split_at_script(const uint32_t *fragment,
 	for (i = 0; i < fragment_len; i++)
 	{
 		int s = ucdn_get_script(fragment[i]);
-		if (s == UCDN_SCRIPT_COMMON || s == UCDN_SCRIPT_INHERITED)
+		if (s == UCDN_SCRIPT_COMMON || s == UCDN_SCRIPT_INHERITED || s == UCDN_SCRIPT_UNKNOWN)
 		{
 			/* Punctuation etc. This is fine. */
+			/* Guess script using the unicode block if we've not determined it yet. */
+			if (script_guess == UCDN_SCRIPT_COMMON)
+				script_guess = guess_script_from_block(fragment[i]);
 		}
 		else if (s == script)
 		{
 			/* Same script. Still fine. */
 		}
-		else if (script == UCDN_SCRIPT_COMMON || script == UCDN_SCRIPT_INHERITED)
+		else if (script == UCDN_SCRIPT_COMMON || script == UCDN_SCRIPT_INHERITED || script == UCDN_SCRIPT_UNKNOWN)
 		{
 			/* First non punctuation thing. Set the script. */
 			script = s;
@@ -203,13 +474,18 @@ split_at_script(const uint32_t *fragment,
 		else
 		{
 			/* Change of script. Break the fragment. */
+			assert(script != UCDN_SCRIPT_COMMON);
 			(*callback)(&fragment[script_start], i - script_start, level, script, arg);
 			script_start = i;
+			script_guess = UCDN_SCRIPT_COMMON;
 			script = s;
 		}
 	}
+
 	if (script_start != fragment_len)
 	{
+		if (script == UCDN_SCRIPT_COMMON)
+			script = script_guess;
 		(*callback)(&fragment[script_start], fragment_len - script_start, level, script, arg);
 	}
 }
@@ -542,7 +818,7 @@ void fz_bidi_fragment_text(fz_context *ctx,
 	if (text == NULL || callback == NULL || textlen == 0)
 		return;
 
-	DBUGH(("fz_bidi_fragment_text('%S', len = %d)\n", text, textlen));
+	DBUGH((ctx, "fz_bidi_fragment_text('%S', len = %d)\n", text, textlen));
 
 	levels = create_levels(ctx, text, textlen, baseDir, FALSE, flags);
 
