@@ -1,4 +1,4 @@
-// Copyright (C) 2004-2024 Artifex Software, Inc.
+// Copyright (C) 2004-2025 Artifex Software, Inc.
 //
 // This file is part of MuPDF.
 //
@@ -22,6 +22,8 @@
 
 #include "mupdf/fitz.h"
 
+#include "context-imp.h"
+
 #include <string.h>
 #ifndef _WIN32
 #include <unistd.h> /* For unlink */
@@ -241,6 +243,8 @@ do_recognize_document_stream_and_dir_content(fz_context *ctx, fz_stream **stream
 	if (dc->count == 0)
 		fz_throw(ctx, FZ_ERROR_ARGUMENT, "No document handlers registered");
 
+	if (magic == NULL)
+		magic = "";
 	ext = strrchr(magic, '.');
 	if (ext)
 		ext = ext + 1;
@@ -274,80 +278,78 @@ do_recognize_document_stream_and_dir_content(fz_context *ctx, fz_stream **stream
 
 	fz_try(ctx)
 	{
-		if ((stream && stream->seek != NULL) || (stream == NULL && dir != NULL))
-		{
-			for (i = 0; i < dc->count; i++)
-			{
-				void *state = NULL;
-				fz_document_recognize_state_free_fn *free_state = NULL;
-				int score = 0;
+		int can_recognize_stream = ((stream && stream->seek != NULL) || (stream == NULL && dir != NULL));
 
-				if (dc->handler[i]->recognize_content)
+		for (i = 0; i < dc->count; i++)
+		{
+			void *state = NULL;
+			fz_document_recognize_state_free_fn *free_state = NULL;
+			int score = 0;
+			int magic_score = 0;
+			const char **entry;
+
+			/* Get a score from recognizing the stream */
+			if (dc->handler[i]->recognize_content && can_recognize_stream)
+			{
+				if (stream)
+					fz_seek(ctx, stream, 0, SEEK_SET);
+				fz_try(ctx)
 				{
-					if (stream)
-						fz_seek(ctx, stream, 0, SEEK_SET);
-					fz_try(ctx)
-					{
-						score = dc->handler[i]->recognize_content(ctx, dc->handler[i], stream, dir, &state, &free_state);
-					}
-					fz_catch(ctx)
-					{
-						/* in case of zip errors when recognizing EPUB/XPS/DOCX files */
-						fz_rethrow_unless(ctx, FZ_ERROR_FORMAT);
-						(void)fz_convert_error(ctx, NULL); /* ugly hack to silence the error message */
-						score = 0;
-					}
+					score = dc->handler[i]->recognize_content(ctx, dc->handler[i], stream, dir, &state, &free_state);
 				}
-				if (best_score < score)
+				fz_catch(ctx)
 				{
-					best_score = score;
-					best_i = i;
-					if (best_free_state)
-						best_free_state(ctx, best_state);
-					best_free_state = free_state;
-					best_state = state;
+					/* in case of zip errors when recognizing EPUB/XPS/DOCX files */
+					fz_rethrow_unless(ctx, FZ_ERROR_FORMAT);
+					(void)fz_convert_error(ctx, NULL); /* ugly hack to silence the error message */
+					score = 0;
 				}
-				else if (free_state)
-					free_state(ctx, state);
 			}
-			if (stream)
-				fz_seek(ctx, stream, 0, SEEK_SET);
-		}
 
-		if (best_score < 100)
-		{
-			for (i = 0; i < dc->count; i++)
+			/* Now get a score from recognizing the magic */
+			if (dc->handler[i]->recognize)
+				magic_score = dc->handler[i]->recognize(ctx, dc->handler[i], magic);
+
+			for (entry = &dc->handler[i]->mimetypes[0]; *entry; entry++)
+				if (!fz_strcasecmp(magic, *entry))
+				{
+					magic_score = 100;
+					break;
+				}
+
+			if (ext)
 			{
-				int score = 0;
-				const char **entry;
-
-				if (dc->handler[i]->recognize)
-					score = dc->handler[i]->recognize(ctx, dc->handler[i], magic);
-
-				for (entry = &dc->handler[i]->mimetypes[0]; *entry; entry++)
-					if (!fz_strcasecmp(magic, *entry) && score < 100)
+				for (entry = &dc->handler[i]->extensions[0]; *entry; entry++)
+					if (!fz_strcasecmp(ext, *entry))
 					{
-						score = 100;
+						magic_score = 100;
 						break;
 					}
-
-				if (ext)
-				{
-					for (entry = &dc->handler[i]->extensions[0]; *entry; entry++)
-						if (!fz_strcasecmp(ext, *entry) && score < 100)
-						{
-							score = 100;
-							break;
-						}
-				}
-
-				if (best_score < score)
-				{
-					best_score = score;
-					best_i = i;
-				}
 			}
+
+			/* If we recognized the format (at least partially), and the magic_score matches, then that's
+			 * definitely the one we want to use. Use 100 + score here, to allow for having multiple
+			 * handlers that support a given magic, where one agent is better than the other. */
+			if (score > 0 && magic_score > 0)
+				score = 100 + score;
+			/* Otherwise, if we didn't recognize the format, we'll weakly believe in the magic, but
+			 * we won't let it override anything that actually will cope. */
+			else if (magic_score > 0)
+				score = 1;
+			if (best_score < score)
+			{
+				best_score = score;
+				best_i = i;
+				if (best_free_state)
+					best_free_state(ctx, best_state);
+				best_free_state = free_state;
+				best_state = state;
+			}
+			else if (free_state)
+				free_state(ctx, state);
 		}
+		if (stream)
+			fz_seek(ctx, stream, 0, SEEK_SET);
 	}
 	fz_catch(ctx)
 	{
@@ -608,6 +610,10 @@ fz_new_document_of_size(fz_context *ctx, int size)
 {
 	fz_document *doc = fz_calloc(ctx, 1, size);
 	doc->refs = 1;
+	doc->id = fz_new_document_id(ctx);
+
+	fz_log_activity(ctx, FZ_ACTIVITY_NEW_DOC, NULL);
+
 	return doc;
 }
 
@@ -873,7 +879,7 @@ int fz_page_number_from_location(fz_context *ctx, fz_document *doc, fz_location
 }
 
 int
-fz_lookup_metadata(fz_context *ctx, fz_document *doc, const char *key, char *buf, int size)
+fz_lookup_metadata(fz_context *ctx, fz_document *doc, const char *key, char *buf, size_t size)
 {
 	if (buf && size > 0)
 		buf[0] = 0;
@@ -956,6 +962,7 @@ fz_load_chapter_page(fz_context *ctx, fz_document *doc, int chapter, int number)
 				doc->open->prev = &page->next;
 			doc->open = page;
 			page->prev = &doc->open;
+			page->in_doc = 1;
 		}
 		return page;
 	}
@@ -1097,6 +1104,12 @@ fz_drop_page(fz_context *ctx, fz_page *page)
 		page->chapter = -1;
 		page->number = -1;
 
+		// If page has never been added to the list of open pages in a document,
+		// it will not get be reaped upon document freeing; instead free the page
+		// immediately.
+		if (!page->in_doc)
+			fz_free(ctx, page);
+
 		fz_drop_document(ctx, doc);
 	}
 }
