@@ -1,4 +1,4 @@
-// Copyright (C) 2004-2021 Artifex Software, Inc.
+// Copyright (C) 2004-2025 Artifex Software, Inc.
 //
 // This file is part of MuPDF.
 //
@@ -17,8 +17,8 @@
 //
 // Alternative licensing terms are available from the licensor.
 // For commercial licensing, see <https://www.artifex.com/> or contact
-// Artifex Software, Inc., 1305 Grant Avenue - Suite 200, Novato,
-// CA 94945, U.S.A., +1(415)492-9861, for further information.
+// Artifex Software, Inc., 39 Mesa Street, Suite 108A, San Francisco,
+// CA 94129, USA, for further information.
 
 #include "mupdf/fitz.h"
 #include "draw-imp.h"
@@ -249,26 +249,14 @@ fz_insert_gel_rect(fz_context *ctx, fz_rasterizer *ras, float fx0, float fy0, fl
 	const int hscale = fz_rasterizer_aa_hscale(ras);
 	const int vscale = fz_rasterizer_aa_vscale(ras);
 
-	if (fx0 <= fx1)
-	{
-		fx0 = floorf(fx0 * hscale);
-		fx1 = ceilf(fx1 * hscale);
-	}
-	else
-	{
-		fx0 = ceilf(fx0 * hscale);
-		fx1 = floorf(fx1 * hscale);
-	}
-	if (fy0 <= fy1)
-	{
-		fy0 = floorf(fy0 * vscale);
-		fy1 = ceilf(fy1 * vscale);
-	}
-	else
-	{
-		fy0 = ceilf(fy0 * vscale);
-		fy1 = floorf(fy1 * vscale);
-	}
+	fx0 = floorf(fx0 * hscale);
+	fx1 = floorf(fx1 * hscale);
+	if (fx1 == fx0)
+		fx1++;
+	fy0 = floorf(fy0 * vscale);
+	fy1 = floorf(fy1 * vscale);
+	if (fy1 == fy0)
+		fy1++;
 
 	fx0 = fz_clamp(fx0, ras->clip.x0, ras->clip.x1);
 	fx1 = fz_clamp(fx1, ras->clip.x0, ras->clip.x1);
@@ -601,8 +589,8 @@ fz_scan_convert_aa(fz_context *ctx, fz_gel *gel, int eofill, const fz_irect *cli
 		fz_free(ctx, gel->deltas);
 		gel->alphas = NULL;
 		gel->deltas = NULL;
-		alphas = gel->alphas = Memento_label(fz_malloc_array(ctx, bcap, unsigned char), "gel_alphas");
-		deltas = gel->deltas = Memento_label(fz_malloc_array(ctx, bcap, int), "gel_deltas");
+		gel->alphas = Memento_label(fz_malloc_array(ctx, bcap, unsigned char), "gel_alphas");
+		gel->deltas = Memento_label(fz_malloc_array(ctx, bcap, int), "gel_deltas");
 	}
 	alphas = gel->alphas;
 	deltas = gel->deltas;
@@ -876,7 +864,6 @@ fz_convert_gel(fz_context *ctx, fz_rasterizer *rast, int eofill, const fz_irect
 			fn = (void *)fz_get_span_color_painter(dst->n, dst->alpha, color, eop);
 		else
 			fn = (void *)fz_get_span_painter(dst->alpha, 1, 0, 255, eop);
-		assert(fn);
 		if (fn == NULL)
 			return;
 		fz_scan_convert_aa(ctx, gel, eofill, clip, dst, color, fn, eop);
