@@ -1,4 +1,4 @@
-// Copyright (C) 2004-2024 Artifex Software, Inc.
+// Copyright (C) 2004-2025 Artifex Software, Inc.
 //
 // This file is part of MuPDF.
 //
@@ -131,8 +131,9 @@ bezier(fz_context *ctx, fz_rasterizer *rast, fz_matrix ctm, float flatness,
 	bezier(ctx, rast, ctm, flatness, xabcd, yabcd, xbcd, ybcd, xcd, ycd, xd, yd, depth + 1);
 }
 
+/* AIX uses 'quad' as a definition, so we can't use that here. */
 static void
-quad(fz_context *ctx, fz_rasterizer *rast, fz_matrix ctm, float flatness,
+quadratic(fz_context *ctx, fz_rasterizer *rast, fz_matrix ctm, float flatness,
 	float xa, float ya,
 	float xb, float yb,
 	float xc, float yc, int depth)
@@ -166,8 +167,8 @@ quad(fz_context *ctx, fz_rasterizer *rast, fz_matrix ctm, float flatness,
 
 	xabc *= 0.25f; yabc *= 0.25f;
 
-	quad(ctx, rast, ctm, flatness, xa, ya, xab, yab, xabc, yabc, depth + 1);
-	quad(ctx, rast, ctm, flatness, xabc, yabc, xbc, ybc, xc, yc, depth + 1);
+	quadratic(ctx, rast, ctm, flatness, xa, ya, xab, yab, xabc, yabc, depth + 1);
+	quadratic(ctx, rast, ctm, flatness, xabc, yabc, xbc, ybc, xc, yc, depth + 1);
 }
 
 typedef struct
@@ -219,7 +220,7 @@ flatten_quadto(fz_context *ctx, void *arg_, float x1, float y1, float x2, float
 {
 	flatten_arg *arg = (flatten_arg *)arg_;
 
-	quad(ctx, arg->rast, arg->ctm, arg->flatness, arg->c.x, arg->c.y, x1, y1, x2, y2, 0);
+	quadratic(ctx, arg->rast, arg->ctm, arg->flatness, arg->c.x, arg->c.y, x1, y1, x2, y2, 0);
 	arg->c.x = x2;
 	arg->c.y = y2;
 }
@@ -907,8 +908,10 @@ fz_stroke_closepath(fz_context *ctx, sctx *s)
 		 * penultimate segment and the end segment. This is what we want. */
 		fz_add_line_join(ctx, s, s->seg[0].x, s->seg[0].y, s->beg[0].x, s->beg[0].y, s->beg[1].x, s->beg[1].y, 0);
 	}
-	else if (s->not_just_moves && s->cap == FZ_LINECAP_ROUND)
+	else if (s->cap == FZ_LINECAP_ROUND)
+	{
 		fz_add_line_dot(ctx, s, s->beg[0].x, s->beg[0].y);
+	}
 
 	s->seg[0] = s->beg[0];
 	s->sn = 0;
@@ -1122,7 +1125,7 @@ fz_dash_lineto(fz_context *ctx, struct sctx *s, float bx, float by, int from_bez
 	float total, used, ratio, tail;
 	float ax, ay;
 	float mx, my;
-	float old_bx, old_by;
+	float old_bx = 0, old_by = 0;
 	int n;
 	int dash_cap = s->stroke->dash_cap;
 
