@@ -1,4 +1,4 @@
-// Copyright (C) 2004-2021 Artifex Software, Inc.
+// Copyright (C) 2004-2025 Artifex Software, Inc.
 //
 // This file is part of MuPDF.
 //
@@ -43,6 +43,8 @@ const char *fz_pdfocr_write_options_usage =
 	"\tstrip-height=N: Strip height (default 0=fullpage)\n"
 	"\tocr-language=<lang>: OCR language (default=eng)\n"
 	"\tocr-datadir=<datadir>: OCR data path (default=rely on TESSDATA_PREFIX)\n"
+	"\tskew=none,auto,<angle>: Whether to skew correct (default=none).\n"
+	"\tskew-border=increase,maintain,decrease: Size change for border pixels (default=increase).\n"
 	"\n";
 
 static const char funky_font[] =
@@ -183,6 +185,27 @@ fz_parse_pdfocr_options(fz_context *ctx, fz_pdfocr_options *opts, const char *ar
 	{
 		fz_copy_option(ctx, val, opts->datadir, nelem(opts->datadir));
 	}
+	if (fz_has_option(ctx, args, "skew", &val))
+	{
+		if (fz_option_eq(val, "auto"))
+			opts->skew_correct = 1;
+		else
+		{
+			opts->skew_correct = 2;
+			opts->skew_angle = fz_atof(val);
+		}
+	}
+	if (fz_has_option(ctx, args, "skew-border", &val))
+	{
+		if (fz_option_eq(val, "increase"))
+			opts->skew_border = 0;
+		else if (fz_option_eq(val, "maintain"))
+			opts->skew_border = 1;
+		else if (fz_option_eq(val, "decrease"))
+			opts->skew_border = 2;
+		else
+			fz_throw(ctx, FZ_ERROR_ARGUMENT, "Unsupported skew-border option");
+	}
 
 	return opts;
 #endif
@@ -219,6 +242,10 @@ typedef struct pdfocr_band_writer_s
 	fz_band_writer super;
 	fz_pdfocr_options options;
 
+	/* The actual output size */
+	int deskewed_w;
+	int deskewed_h;
+
 	int obj_num;
 	int xref_max;
 	int64_t *xref;
@@ -229,6 +256,8 @@ typedef struct pdfocr_band_writer_s
 	unsigned char *compbuf;
 	size_t complen;
 
+	fz_pixmap *skew_bitmap;
+
 	void *tessapi;
 	fz_pixmap *ocrbitmap;
 
@@ -255,6 +284,41 @@ new_obj(fz_context *ctx, pdfocr_band_writer *writer)
 	return writer->obj_num++;
 }
 
+static void
+post_skew_write_header(fz_context *ctx, pdfocr_band_writer *writer, int w, int h)
+{
+	fz_output *out = writer->super.out;
+	int xres = writer->super.xres;
+	int yres = writer->super.yres;
+	int sh = writer->options.strip_height;
+	int n = writer->super.n;
+	int strips;
+	int i;
+
+	if (sh == 0)
+		sh = h;
+	assert(sh != 0 && "pdfocr_write_header() should not be given zero height input.");
+	strips = (h + sh-1)/sh;
+
+	writer->deskewed_w = w;
+	writer->deskewed_h = h;
+
+	writer->stripbuf = Memento_label(fz_malloc(ctx, (size_t)w * sh * n), "pdfocr_stripbuf");
+	writer->complen = fz_deflate_bound(ctx, (size_t)w * sh * n);
+	writer->compbuf = Memento_label(fz_malloc(ctx, writer->complen), "pdfocr_compbuf");
+
+	/* Always round the width of ocrbitmap up to a multiple of 4. */
+	writer->ocrbitmap = fz_new_pixmap(ctx, NULL, (w+3)&~3, h, NULL, 0);
+	fz_set_pixmap_resolution(ctx, writer->ocrbitmap, xres, yres);
+
+	/* Send the Page Object */
+	fz_write_printf(ctx, out, "%d 0 obj\n<</Type/Page/Parent 2 0 R/Resources<</XObject<<", new_obj(ctx, writer));
+	for (i = 0; i < strips; i++)
+		fz_write_printf(ctx, out, "/I%d %d 0 R", i, writer->obj_num + i);
+	fz_write_printf(ctx, out, ">>/Font<</F0 3 0 R>>>>/MediaBox[0 0 %g %g]/Contents %d 0 R>>\nendobj\n",
+		w * 72.0f / xres, h * 72.0f / yres, writer->obj_num + strips);
+}
+
 static void
 pdfocr_write_header(fz_context *ctx, fz_band_writer *writer_, fz_colorspace *cs)
 {
@@ -265,16 +329,11 @@ pdfocr_write_header(fz_context *ctx, fz_band_writer *writer_, fz_colorspace *cs)
 	int n = writer->super.n;
 	int s = writer->super.s;
 	int a = writer->super.alpha;
-	int xres = writer->super.xres;
-	int yres = writer->super.yres;
 	int sh = writer->options.strip_height;
-	int strips;
-	int i;
 
 	if (sh == 0)
 		sh = h;
 	assert(sh != 0 && "pdfocr_write_header() should not be given zero height input.");
-	strips = (h + sh-1)/sh;
 
 	if (a != 0)
 		fz_throw(ctx, FZ_ERROR_ARGUMENT, "PDFOCR cannot write alpha channel");
@@ -289,12 +348,6 @@ pdfocr_write_header(fz_context *ctx, fz_band_writer *writer_, fz_colorspace *cs)
 	writer->compbuf = NULL;
 	fz_drop_pixmap(ctx, writer->ocrbitmap);
 	writer->ocrbitmap = NULL;
-	writer->stripbuf = Memento_label(fz_malloc(ctx, (size_t)w * sh * n), "pdfocr_stripbuf");
-	writer->complen = fz_deflate_bound(ctx, (size_t)w * sh * n);
-	writer->compbuf = Memento_label(fz_malloc(ctx, writer->complen), "pdfocr_compbuf");
-	/* Always round the width of ocrbitmap up to a multiple of 4. */
-	writer->ocrbitmap = fz_new_pixmap(ctx, NULL, (w+3)&~3, h, NULL, 0);
-	fz_set_pixmap_resolution(ctx, writer->ocrbitmap, xres, yres);
 
 	/* Send the file header on the first page */
 	if (writer->pages == 0)
@@ -332,12 +385,10 @@ pdfocr_write_header(fz_context *ctx, fz_band_writer *writer_, fz_colorspace *cs)
 	writer->page_obj[writer->pages] = writer->obj_num;
 	writer->pages++;
 
-	/* Send the Page Object */
-	fz_write_printf(ctx, out, "%d 0 obj\n<</Type/Page/Parent 2 0 R/Resources<</XObject<<", new_obj(ctx, writer));
-	for (i = 0; i < strips; i++)
-		fz_write_printf(ctx, out, "/I%d %d 0 R", i, writer->obj_num + i);
-	fz_write_printf(ctx, out, ">>/Font<</F0 3 0 R>>>>/MediaBox[0 0 %g %g]/Contents %d 0 R>>\nendobj\n",
-		w * 72.0f / xres, h * 72.0f / yres, writer->obj_num + strips);
+	if (writer->options.skew_correct)
+		writer->skew_bitmap = fz_new_pixmap(ctx, n == 3 ? fz_device_rgb(ctx) : fz_device_gray(ctx), w, h, NULL, 0);
+	else
+		post_skew_write_header(ctx, writer, w, h);
 }
 
 static void
@@ -345,7 +396,7 @@ flush_strip(fz_context *ctx, pdfocr_band_writer *writer, int fill)
 {
 	unsigned char *data = writer->stripbuf;
 	fz_output *out = writer->super.out;
-	int w = writer->super.w;
+	int w = writer->deskewed_w;
 	int n = writer->super.n;
 	size_t len = (size_t)w*n*fill;
 
@@ -365,19 +416,15 @@ flush_strip(fz_context *ctx, pdfocr_band_writer *writer, int fill)
 }
 
 static void
-pdfocr_write_band(fz_context *ctx, fz_band_writer *writer_, int stride, int band_start, int band_height, const unsigned char *sp)
+post_skew_write_band(fz_context *ctx, pdfocr_band_writer *writer, int stride, int band_start, int band_height, const unsigned char *sp)
 {
-	pdfocr_band_writer *writer = (pdfocr_band_writer *)writer_;
-	fz_output *out = writer->super.out;
-	int w = writer->super.w;
-	int h = writer->super.h;
+	int w = writer->deskewed_w;
+	int h = writer->deskewed_h;
 	int n = writer->super.n;
+	int x, y;
 	int sh = writer->options.strip_height;
 	int line;
-	unsigned char *d = writer->ocrbitmap->samples;
-
-	if (!out)
-		return;
+	unsigned char *d;
 
 	if (sh == 0)
 		sh = h;
@@ -386,20 +433,19 @@ pdfocr_write_band(fz_context *ctx, fz_band_writer *writer_, int stride, int band
 	{
 		int dstline = (band_start+line) % sh;
 		memcpy(writer->stripbuf + (size_t)w*n*dstline,
-			   sp + (size_t)line * w * n,
-			   (size_t)w * n);
+			sp + (size_t)line * w * n,
+			(size_t)w * n);
 		if (dstline+1 == sh)
 			flush_strip(ctx, writer, dstline+1);
 	}
-
 	if (band_start + band_height == h && h % sh != 0)
 		flush_strip(ctx, writer, h % sh);
 
 	/* Copy strip to ocrbitmap, converting if required. */
+	d = writer->ocrbitmap->samples;
 	d += band_start*w;
 	if (n == 1)
 	{
-		int y;
 		for (y = band_height; y > 0; y--)
 		{
 			memcpy(d, sp, w);
@@ -410,7 +456,6 @@ pdfocr_write_band(fz_context *ctx, fz_band_writer *writer_, int stride, int band
 	}
 	else
 	{
-		int x, y;
 		for (y = band_height; y > 0; y--)
 		{
 			for (x = w; x > 0; x--)
@@ -424,6 +469,28 @@ pdfocr_write_band(fz_context *ctx, fz_band_writer *writer_, int stride, int band
 	}
 }
 
+static void
+pdfocr_write_band(fz_context *ctx, fz_band_writer *writer_, int stride, int band_start, int band_height, const unsigned char *sp)
+{
+	pdfocr_band_writer *writer = (pdfocr_band_writer *)writer_;
+	fz_output *out = writer->super.out;
+	int w = writer->super.w;
+	int n = writer->super.n;
+	unsigned char *d;
+
+	if (!out)
+		return;
+
+	if (writer->skew_bitmap)
+	{
+		d = writer->skew_bitmap->samples;
+		d += band_start*w*n;
+		memcpy(d, sp, w*n*band_height);
+	}
+	else
+		post_skew_write_band(ctx, writer, stride, band_start, band_height, sp);
+}
+
 enum
 {
 	WORD_CONTAINS_L2R = 1,
@@ -438,7 +505,7 @@ typedef struct word_t
 	float bbox[4];
 	int dirn;
 	int len;
-	int chars[1];
+	int chars[FZ_FLEXIBLE_ARRAY];
 } word_t;
 
 typedef struct
@@ -617,7 +684,7 @@ queue_word(fz_context *ctx, char_callback_data_t *cb)
 	if (cb->word_len == 0)
 		return;
 
-	word = fz_malloc(ctx, sizeof(*word) + (cb->word_len-1)*sizeof(int));
+	word = fz_malloc_flexible(ctx, word_t, chars, cb->word_len);
 	word->next = NULL;
 	word->len = cb->word_len;
 	memcpy(word->bbox, cb->word_bbox, 4*sizeof(float));
@@ -761,23 +828,47 @@ pdfocr_progress(fz_context *ctx, void *arg, int prog)
 	return writer->progress(ctx, writer->progress_arg, writer->pages - 1, prog);
 }
 
+static void
+do_skew_correct(fz_context *ctx, pdfocr_band_writer *writer)
+{
+	fz_pixmap *deskewed;
+
+	if (writer->options.skew_correct == 1)
+		writer->options.skew_angle = fz_detect_skew(ctx, writer->skew_bitmap);
+
+	deskewed = fz_deskew_pixmap(ctx, writer->skew_bitmap, writer->options.skew_angle, writer->options.skew_border);
+
+	fz_try(ctx)
+	{
+		post_skew_write_header(ctx, writer, deskewed->w, deskewed->h);
+		post_skew_write_band(ctx, writer, deskewed->stride, 0, deskewed->h, deskewed->samples);
+	}
+	fz_always(ctx)
+		fz_drop_pixmap(ctx, deskewed);
+	fz_catch(ctx)
+		fz_rethrow(ctx);
+}
+
 static void
 pdfocr_write_trailer(fz_context *ctx, fz_band_writer *writer_)
 {
 	pdfocr_band_writer *writer = (pdfocr_band_writer *)writer_;
 	fz_output *out = writer->super.out;
-	int w = writer->super.w;
-	int h = writer->super.h;
 	int xres = writer->super.xres;
 	int yres = writer->super.yres;
 	int sh = writer->options.strip_height;
 	int strips;
-	int i;
+	int w, h, i;
 	size_t len;
 	unsigned char *data;
 	fz_buffer *buf = NULL;
 	char_callback_data_t cb = { NULL };
 
+	if (writer->options.skew_correct)
+		do_skew_correct(ctx, writer);
+
+	w = writer->deskewed_w;
+	h = writer->deskewed_h;
 	if (sh == 0)
 		sh = h;
 	strips = (h + sh-1)/sh;
@@ -873,6 +964,8 @@ static void
 pdfocr_drop_band_writer(fz_context *ctx, fz_band_writer *writer_)
 {
 	pdfocr_band_writer *writer = (pdfocr_band_writer *)writer_;
+
+	fz_free(ctx, writer->options.options);
 	fz_free(ctx, writer->stripbuf);
 	fz_free(ctx, writer->compbuf);
 	fz_free(ctx, writer->page_obj);
@@ -896,7 +989,10 @@ fz_band_writer *fz_new_pdfocr_band_writer(fz_context *ctx, fz_output *out, const
 	writer->super.drop = pdfocr_drop_band_writer;
 
 	if (options)
+	{
 		writer->options = *options;
+		writer->options.options = NULL;
+	}
 	else
 		memset(&writer->options, 0, sizeof(writer->options));
 
@@ -914,12 +1010,14 @@ fz_band_writer *fz_new_pdfocr_band_writer(fz_context *ctx, fz_output *out, const
 
 	fz_try(ctx)
 	{
-		writer->tessapi = ocr_init(ctx, writer->options.language, writer->options.datadir);
+		if (options && options->options)
+			writer->options.options = fz_strdup(ctx, options->options);
+		writer->tessapi = ocr_init(ctx, writer->options.language, writer->options.datadir, writer->options.options);
 	}
 	fz_catch(ctx)
 	{
 		fz_drop_band_writer(ctx, &writer->super);
-		fz_throw(ctx, FZ_ERROR_LIBRARY, "OCR initialisation failed");
+		fz_rethrow(ctx);
 	}
 
 	return &writer->super;
