@@ -1,4 +1,4 @@
-// Copyright (C) 2004-2021 Artifex Software, Inc.
+// Copyright (C) 2004-2024 Artifex Software, Inc.
 //
 // This file is part of MuPDF.
 //
@@ -24,19 +24,13 @@
 
 #include <limits.h>
 
-#ifndef PATH_MAX
-#define PATH_MAX 4096
-#endif
-
 typedef struct
 {
 	fz_document_writer super;
 	char *path;
 	int count;
 	fz_output *out;
-	int text_format;
-	int reuse_images;
-	int id;
+	fz_svg_device_options opts;
 } fz_svg_writer;
 
 const char *fz_svg_write_options_usage =
@@ -44,6 +38,7 @@ const char *fz_svg_write_options_usage =
 	"\ttext=text: Emit text as <text> elements (inaccurate fonts).\n"
 	"\ttext=path: Emit text as <path> elements (accurate fonts).\n"
 	"\tno-reuse-images: Do not reuse images using <symbol> definitions.\n"
+	"\tresolution: Resolution to use when rasterizing elements.\n"
 	"\n"
 	;
 
@@ -58,9 +53,18 @@ svg_begin_page(fz_context *ctx, fz_document_writer *wri_, fz_rect mediabox)
 
 	wri->count += 1;
 
-	fz_format_output_path(ctx, path, sizeof path, wri->path, wri->count);
-	wri->out = fz_new_output_with_path(ctx, path, 0);
-	return fz_new_svg_device_with_id(ctx, wri->out, w, h, wri->text_format, wri->reuse_images, &wri->id);
+	if (wri->path)
+	{
+		fz_format_output_path(ctx, path, sizeof path, wri->path, wri->count);
+		wri->out = fz_new_output_with_path(ctx, path, 0);
+	}
+	else
+	{
+		if (!wri->out)
+			fz_throw(ctx, FZ_ERROR_ARGUMENT, "cannot write multiple pages to a single SVG output");
+	}
+
+	return fz_new_svg_device_with_options(ctx, wri->out, w, h, &wri->opts);
 }
 
 static void
@@ -94,24 +98,10 @@ svg_drop_writer(fz_context *ctx, fz_document_writer *wri_)
 fz_document_writer *
 fz_new_svg_writer(fz_context *ctx, const char *path, const char *args)
 {
-	const char *val;
 	fz_svg_writer *wri = fz_new_derived_document_writer(ctx, fz_svg_writer, svg_begin_page, svg_end_page, NULL, svg_drop_writer);
-
-	wri->text_format = FZ_SVG_TEXT_AS_PATH;
-	wri->reuse_images = 1;
-
 	fz_try(ctx)
 	{
-		if (fz_has_option(ctx, args, "text", &val))
-		{
-			if (fz_option_eq(val, "text"))
-				wri->text_format = FZ_SVG_TEXT_AS_TEXT;
-			else if (fz_option_eq(val, "path"))
-				wri->text_format = FZ_SVG_TEXT_AS_PATH;
-		}
-		if (fz_has_option(ctx, args, "no-reuse-images", &val))
-			if (fz_option_eq(val, "yes"))
-				wri->reuse_images = 0;
+		fz_parse_svg_device_options(ctx, &wri->opts, args);
 		wri->path = fz_strdup(ctx, path ? path : "out-%04d.svg");
 	}
 	fz_catch(ctx)
@@ -119,6 +109,22 @@ fz_new_svg_writer(fz_context *ctx, const char *path, const char *args)
 		fz_free(ctx, wri);
 		fz_rethrow(ctx);
 	}
+	return (fz_document_writer*)wri;
+}
 
+fz_document_writer *
+fz_new_svg_writer_with_output(fz_context *ctx, fz_output *out, const char *args)
+{
+	fz_svg_writer *wri = fz_new_derived_document_writer(ctx, fz_svg_writer, svg_begin_page, svg_end_page, NULL, svg_drop_writer);
+	fz_try(ctx)
+	{
+		fz_parse_svg_device_options(ctx, &wri->opts, args);
+		wri->out = out;
+	}
+	fz_catch(ctx)
+	{
+		fz_free(ctx, wri);
+		fz_rethrow(ctx);
+	}
 	return (fz_document_writer*)wri;
 }
