@@ -989,7 +989,7 @@ fz_process_shade(fz_context *ctx, fz_shade *shade, fz_matrix ctm, fz_rect scisso
 	painter.prepare = prepare;
 	painter.process = process;
 	painter.process_arg = process_arg;
-	painter.ncomp = (shade->use_function > 0 ? 1 : fz_colorspace_n(ctx, shade->colorspace));
+	painter.ncomp = (shade->function_stride > 0 ? 1 : fz_colorspace_n(ctx, shade->colorspace));
 
 	if (shade->type == FZ_FUNCTION_BASED)
 		fz_process_shade_type1(ctx, shade, ctm, &painter);
@@ -1006,7 +1006,7 @@ fz_process_shade(fz_context *ctx, fz_shade *shade, fz_matrix ctm, fz_rect scisso
 	else if (shade->type == FZ_MESH_TYPE7)
 		fz_process_shade_type7(ctx, shade, ctm, &painter);
 	else
-		fz_throw(ctx, FZ_ERROR_GENERIC, "Unexpected mesh type %d\n", shade->type);
+		fz_throw(ctx, FZ_ERROR_ARGUMENT, "Unexpected mesh type %d\n", shade->type);
 }
 
 static fz_rect
@@ -1094,7 +1094,7 @@ fz_bound_mesh(fz_context *ctx, fz_shade *shade)
 		shade->type == FZ_MESH_TYPE7)
 		return fz_bound_mesh_type4567(ctx, shade);
 	else
-		fz_throw(ctx, FZ_ERROR_GENERIC, "Unexpected mesh type %d\n", shade->type);
+		fz_throw(ctx, FZ_ERROR_ARGUMENT, "Unexpected mesh type %d\n", shade->type);
 }
 
 fz_shade *
@@ -1112,6 +1112,7 @@ fz_drop_shade_imp(fz_context *ctx, fz_storable *shade_)
 	if (shade->type == FZ_FUNCTION_BASED)
 		fz_free(ctx, shade->u.f.fn_vals);
 	fz_drop_compressed_buffer(ctx, shade->buffer);
+	fz_free(ctx, shade->function);
 	fz_free(ctx, shade);
 }
 
