@@ -254,8 +254,22 @@ fz_trace_clip_stroke_path(fz_context *ctx, fz_device *dev_, const fz_path *path,
 {
 	fz_trace_device *dev = (fz_trace_device*)dev_;
 	fz_output *out = dev->out;
+	int i;
+
 	fz_trace_indent(ctx, out, dev->depth);
 	fz_write_printf(ctx, out, "<clip_stroke_path");
+	fz_write_printf(ctx, out, " linewidth=\"%g\"", stroke->linewidth);
+	fz_write_printf(ctx, out, " miterlimit=\"%g\"", stroke->miterlimit);
+	fz_write_printf(ctx, out, " linecap=\"%d,%d,%d\"", stroke->start_cap, stroke->dash_cap, stroke->end_cap);
+	fz_write_printf(ctx, out, " linejoin=\"%d\"", stroke->linejoin);
+	if (stroke->dash_len)
+	{
+		fz_write_printf(ctx, out, " dash_phase=\"%g\" dash=\"", stroke->dash_phase);
+		for (i = 0; i < stroke->dash_len; i++)
+			fz_write_printf(ctx, out, "%s%g", i > 0 ? " " : "", stroke->dash_list[i]);
+		fz_write_printf(ctx, out, "\"");
+	}
+
 	fz_trace_matrix(ctx, out, ctm);
 	fz_write_printf(ctx, out, ">\n");
 	fz_trace_path(ctx, dev, path);
@@ -517,12 +531,12 @@ fz_trace_end_group(fz_context *ctx, fz_device *dev_)
 }
 
 static int
-fz_trace_begin_tile(fz_context *ctx, fz_device *dev_, fz_rect area, fz_rect view, float xstep, float ystep, fz_matrix ctm, int id)
+fz_trace_begin_tile(fz_context *ctx, fz_device *dev_, fz_rect area, fz_rect view, float xstep, float ystep, fz_matrix ctm, int id, int doc_id)
 {
 	fz_trace_device *dev = (fz_trace_device*)dev_;
 	fz_output *out = dev->out;
 	fz_trace_indent(ctx, out, dev->depth);
-	fz_write_printf(ctx, out, "<tile id=\"%d\"", id);
+	fz_write_printf(ctx, out, "<tile id=\"%d\" doc_id=\"%d\"", id, doc_id);
 	fz_write_printf(ctx, out, " area=\"%g %g %g %g\"", area.x0, area.y0, area.x1, area.y1);
 	fz_write_printf(ctx, out, " view=\"%g %g %g %g\"", view.x0, view.y0, view.x1, view.y1);
 	fz_write_printf(ctx, out, " xstep=\"%g\" ystep=\"%g\"", xstep, ystep);
@@ -570,7 +584,9 @@ fz_trace_begin_structure(fz_context *ctx, fz_device *dev_, fz_structure standard
 	const char *str = fz_structure_to_string(standard);
 	fz_trace_indent(ctx, out, dev->depth);
 	fz_write_printf(ctx, out, "<structure standard=\"%s\"", str);
-	if (raw && strcmp(str, raw))
+	if (raw == NULL)
+		raw = "";
+	if (strcmp(str, raw))
 		fz_write_printf(ctx, out, " raw=\"%s\"", raw);
 	if (idx != 0)
 		fz_write_printf(ctx, out, " idx=\"%d\"", idx);
