@@ -47,27 +47,6 @@ static const char *border_style_kw[] = {
 	"solid",
 };
 
-static const char *color_kw[] = {
-	"aqua",
-	"black",
-	"blue",
-	"fuchsia",
-	"gray",
-	"green",
-	"lime",
-	"maroon",
-	"navy",
-	"olive",
-	"orange",
-	"purple",
-	"red",
-	"silver",
-	"teal",
-	"transparent",
-	"white",
-	"yellow",
-};
-
 static const char *list_style_type_kw[] = {
 	"armenian",
 	"circle",
@@ -176,6 +155,201 @@ is_italic_from_font_style(const char *style)
 	return !strcmp(style, "italic") || !strcmp(style, "oblique");
 }
 
+/*
+ * CSS/SVG named color list
+ * https://drafts.csswg.org/css-color-3/#svg-color
+ * https://developer.mozilla.org/en-US/docs/Web/CSS/named-color
+ */
+
+static const struct {
+	const char *name;
+	fz_css_color color;
+} named_color_list[] = {
+	{"aliceblue", {240,248,255,255}},
+	{"antiquewhite", {250,235,215,255}},
+	{"aqua", {0,255,255,255}},
+	{"aquamarine", {127,255,212,255}},
+	{"azure", {240,255,255,255}},
+	{"beige", {245,245,220,255}},
+	{"bisque", {255,228,196,255}},
+	{"black", {0,0,0,255}},
+	{"blanchedalmond", {255,235,205,255}},
+	{"blue", {0,0,255,255}},
+	{"blueviolet", {138,43,226,255}},
+	{"brown", {165,42,42,255}},
+	{"burlywood", {222,184,135,255}},
+	{"cadetblue", {95,158,160,255}},
+	{"chartreuse", {127,255,0,255}},
+	{"chocolate", {210,105,30,255}},
+	{"coral", {255,127,80,255}},
+	{"cornflowerblue", {100,149,237,255}},
+	{"cornsilk", {255,248,220,255}},
+	{"crimson", {220,20,60,255}},
+	{"cyan", {0,255,255,255}},
+	{"darkblue", {0,0,139,255}},
+	{"darkcyan", {0,139,139,255}},
+	{"darkgoldenrod", {184,134,11,255}},
+	{"darkgray", {169,169,169,255}},
+	{"darkgreen", {0,100,0,255}},
+	{"darkgrey", {169,169,169,255}},
+	{"darkkhaki", {189,183,107,255}},
+	{"darkmagenta", {139,0,139,255}},
+	{"darkolivegreen", {85,107,47,255}},
+	{"darkorange", {255,140,0,255}},
+	{"darkorchid", {153,50,204,255}},
+	{"darkred", {139,0,0,255}},
+	{"darksalmon", {233,150,122,255}},
+	{"darkseagreen", {143,188,143,255}},
+	{"darkslateblue", {72,61,139,255}},
+	{"darkslategray", {47,79,79,255}},
+	{"darkslategrey", {47,79,79,255}},
+	{"darkturquoise", {0,206,209,255}},
+	{"darkviolet", {148,0,211,255}},
+	{"deeppink", {255,20,147,255}},
+	{"deepskyblue", {0,191,255,255}},
+	{"dimgray", {105,105,105,255}},
+	{"dimgrey", {105,105,105,255}},
+	{"dodgerblue", {30,144,255,255}},
+	{"firebrick", {178,34,34,255}},
+	{"floralwhite", {255,250,240,255}},
+	{"forestgreen", {34,139,34,255}},
+	{"fuchsia", {255,0,255,255}},
+	{"gainsboro", {220,220,220,255}},
+	{"ghostwhite", {248,248,255,255}},
+	{"gold", {255,215,0,255}},
+	{"goldenrod", {218,165,32,255}},
+	{"gray", {128,128,128,255}},
+	{"green", {0,128,0,255}},
+	{"greenyellow", {173,255,47,255}},
+	{"grey", {128,128,128,255}},
+	{"honeydew", {240,255,240,255}},
+	{"hotpink", {255,105,180,255}},
+	{"indianred", {205,92,92,255}},
+	{"indigo", {75,0,130,255}},
+	{"ivory", {255,255,240,255}},
+	{"khaki", {240,230,140,255}},
+	{"lavender", {230,230,250,255}},
+	{"lavenderblush", {255,240,245,255}},
+	{"lawngreen", {124,252,0,255}},
+	{"lemonchiffon", {255,250,205,255}},
+	{"lightblue", {173,216,230,255}},
+	{"lightcoral", {240,128,128,255}},
+	{"lightcyan", {224,255,255,255}},
+	{"lightgoldenrodyellow", {250,250,210,255}},
+	{"lightgray", {211,211,211,255}},
+	{"lightgreen", {144,238,144,255}},
+	{"lightgrey", {211,211,211,255}},
+	{"lightpink", {255,182,193,255}},
+	{"lightsalmon", {255,160,122,255}},
+	{"lightseagreen", {32,178,170,255}},
+	{"lightskyblue", {135,206,250,255}},
+	{"lightslategray", {119,136,153,255}},
+	{"lightslategrey", {119,136,153,255}},
+	{"lightsteelblue", {176,196,222,255}},
+	{"lightyellow", {255,255,224,255}},
+	{"lime", {0,255,0,255}},
+	{"limegreen", {50,205,50,255}},
+	{"linen", {250,240,230,255}},
+	{"magenta", {255,0,255,255}},
+	{"maroon", {128,0,0,255}},
+	{"mediumaquamarine", {102,205,170,255}},
+	{"mediumblue", {0,0,205,255}},
+	{"mediumorchid", {186,85,211,255}},
+	{"mediumpurple", {147,112,219,255}},
+	{"mediumseagreen", {60,179,113,255}},
+	{"mediumslateblue", {123,104,238,255}},
+	{"mediumspringgreen", {0,250,154,255}},
+	{"mediumturquoise", {72,209,204,255}},
+	{"mediumvioletred", {199,21,133,255}},
+	{"midnightblue", {25,25,112,255}},
+	{"mintcream", {245,255,250,255}},
+	{"mistyrose", {255,228,225,255}},
+	{"moccasin", {255,228,181,255}},
+	{"navajowhite", {255,222,173,255}},
+	{"navy", {0,0,128,255}},
+	{"oldlace", {253,245,230,255}},
+	{"olive", {128,128,0,255}},
+	{"olivedrab", {107,142,35,255}},
+	{"orange", {255,165,0,255}},
+	{"orangered", {255,69,0,255}},
+	{"orchid", {218,112,214,255}},
+	{"palegoldenrod", {238,232,170,255}},
+	{"palegreen", {152,251,152,255}},
+	{"paleturquoise", {175,238,238,255}},
+	{"palevioletred", {219,112,147,255}},
+	{"papayawhip", {255,239,213,255}},
+	{"peachpuff", {255,218,185,255}},
+	{"peru", {205,133,63,255}},
+	{"pink", {255,192,203,255}},
+	{"plum", {221,160,221,255}},
+	{"powderblue", {176,224,230,255}},
+	{"purple", {128,0,128,255}},
+	{"rebeccapurple", {102,51,153,255}},
+	{"red", {255,0,0,255}},
+	{"rosybrown", {188,143,143,255}},
+	{"royalblue", {65,105,225,255}},
+	{"saddlebrown", {139,69,19,255}},
+	{"salmon", {250,128,114,255}},
+	{"sandybrown", {244,164,96,255}},
+	{"seagreen", {46,139,87,255}},
+	{"seashell", {255,245,238,255}},
+	{"sienna", {160,82,45,255}},
+	{"silver", {192,192,192,255}},
+	{"skyblue", {135,206,235,255}},
+	{"slateblue", {106,90,205,255}},
+	{"slategray", {112,128,144,255}},
+	{"slategrey", {112,128,144,255}},
+	{"snow", {255,250,250,255}},
+	{"springgreen", {0,255,127,255}},
+	{"steelblue", {70,130,180,255}},
+	{"tan", {210,180,140,255}},
+	{"teal", {0,128,128,255}},
+	{"thistle", {216,191,216,255}},
+	{"tomato", {255,99,71,255}},
+	{"transparent", {0,0,0,0}},
+	{"turquoise", {64,224,208,255}},
+	{"violet", {238,130,238,255}},
+	{"wheat", {245,222,179,255}},
+	{"white", {255,255,255,255}},
+	{"whitesmoke", {245,245,245,255}},
+	{"yellow", {255,255,0,255}},
+	{"yellowgreen", {154,205,50,255}},
+};
+
+static inline int fz_tolower_ascii(int c)
+{
+	if ((unsigned int)c - 'A' < 26)
+		return c | 32;
+	return c;
+}
+
+static int fz_strcasecmp_ascii(const char *a, const char *b)
+{
+	const unsigned char *l = (void *)a, *r = (void *)b;
+	while (*l && *r && (*l == *r || fz_tolower_ascii(*l) == fz_tolower_ascii(*r)))
+		l++, r++;
+	return fz_tolower_ascii(*l) - fz_tolower_ascii(*r);
+}
+
+static const fz_css_color *
+lookup_named_color(const char *name)
+{
+	int l = 0;
+	int r = nelem(named_color_list) - 1;
+	while (l <= r)
+	{
+		int m = (l + r) >> 1;
+		int c = fz_strcasecmp_ascii(name, named_color_list[m].name);
+		if (c < 0)
+			r = m - 1;
+		else if (c > 0)
+			l = m + 1;
+		else
+			return &named_color_list[m].color;
+	}
+	return NULL;
+}
+
 /*
  * Compute specificity
  */
@@ -308,6 +482,169 @@ match_att_has_condition(fz_xml *node, const char *att, const char *needle)
 	return 0;
 }
 
+static int
+match_nth(int count, int a, int b)
+{
+	if (a == 0)
+	{
+		return count == b;
+	}
+	else
+	{
+		/* (count - b) / a is positive or (count - b) is 0 */
+		int delta = count - b;
+		if (((delta > 0) == (a > 0)) || delta == 0)
+			/* (count - b) / a is integer */
+			return (delta % a == 0);
+		return 0;
+	}
+}
+
+static int
+count_siblings_before(fz_xml *target)
+{
+	fz_xml *node;
+	int count = 0;
+	for (node = fz_xml_down(fz_xml_up(target)); node && node != target; node = fz_xml_next(node))
+		if (fz_xml_tag(node) != NULL)
+			++count;
+	return count;
+}
+
+static int
+count_siblings_after(fz_xml *node)
+{
+	int count = 0;
+	for (; node; node = fz_xml_next(node))
+		if (fz_xml_tag(node) != NULL)
+			++count;
+	return count;
+}
+
+static int
+count_siblings_of_type_before(fz_xml *target)
+{
+	const char *tag = fz_xml_tag(target);
+	fz_xml *node;
+	int count = 0;
+	for (node = fz_xml_down(fz_xml_up(target)); node && node != target; node = fz_xml_next(node))
+		if (fz_xml_is_tag(node, tag))
+			++count;
+	return count;
+}
+
+static int
+count_siblings_of_type_after(fz_xml *node)
+{
+	const char *tag = fz_xml_tag(node);
+	int count = 0;
+	for (; node; node = fz_xml_next(node))
+		if (fz_xml_is_tag(node, tag))
+			++count;
+	return count;
+}
+
+static int
+match_nth_child(fz_xml *target, int a, int b)
+{
+	return match_nth(count_siblings_before(target) + 1, a, b);
+}
+
+static int
+match_nth_last_child(fz_xml *target, int a, int b)
+{
+	return match_nth(count_siblings_after(target) + 1, a, b);
+}
+
+static int
+match_nth_of_type(fz_xml *target, int a, int b)
+{
+	return match_nth(count_siblings_of_type_before(target) + 1, a, b);
+}
+
+static int
+match_nth_last_of_type(fz_xml *target, int a, int b)
+{
+	return match_nth(count_siblings_of_type_after(target) + 1, a, b);
+}
+
+static int
+match_an_plus_b_microsyntax(fz_xml *node, const char *val, int (*callback)(fz_xml *node, int a, int b))
+{
+	// see https://www.w3.org/TR/css-syntax-3/#anb-microsyntax
+	// - A is 0 for nth-child(B) -- same as nth-child(0n+B)
+	// - A is 1 for nth-child(n+B) -- same as nth-child(1n+B)
+	// - A is -1 for nth-child(-n+B) -- same as nth-child(-1n+B)
+
+	int a, b, n;
+	int m = strlen(val);
+
+	if (sscanf(val, " even %n", &n) == 0 && n == m)
+		return callback(node, 2, 0);
+	if (sscanf(val, " odd %n", &n) == 0 && n == m)
+		return callback(node, 2, 1);
+
+	if (sscanf(val, " %dn + %d %n", &a, &b, &n) == 2 && n == m)
+		return callback(node, a, b);
+	if (sscanf(val, " %dn - %d %n", &a, &b, &n) == 2 && n == m)
+		return callback(node, a, -b);
+	if (sscanf(val, " %dn %n", &a, &n) == 1 && n == m)
+		return callback(node, a, 0);
+	if (sscanf(val, " %d %n", &b, &n) == 1 && n == m)
+		return callback(node, 0, b);
+
+	if (sscanf(val, " -n + %d %n", &b, &n) == 1 && n == m)
+		return callback(node, -1, b);
+	if (sscanf(val, " -n - %d %n", &b, &n) == 1 && n == m)
+		return callback(node, -1, -b);
+
+	if (sscanf(val, " n + %d %n", &b, &n) == 1 && n == m)
+		return callback(node, 1, b);
+	if (sscanf(val, " n - %d %n", &b, &n) == 1 && n == m)
+		return callback(node, 1, -b);
+
+	if (sscanf(val, " n %n", &n) == 0 && n == m)
+		return 1;
+
+	return 0;
+}
+
+static int
+match_pseudo_condition(fz_xml *node, const char *key, const char *val)
+{
+	if (!strcmp(key, "link"))
+		return fz_xml_att(node, "href") != NULL;
+
+	if (!strcmp(key, "empty"))
+		return fz_xml_down(node) == NULL;
+	if (!strcmp(key, "root"))
+		return fz_xml_up(node) == NULL;
+
+	if (!strcmp(key, "first-child"))
+		return count_siblings_before(node) == 0;
+	if (!strcmp(key, "last-child"))
+		return count_siblings_after(node) == 0;
+	if (!strcmp(key, "only-child"))
+		return count_siblings_before(node) == 0 && count_siblings_after(node) == 0;
+	if (!strcmp(key, "nth-child") && val != NULL)
+		return match_an_plus_b_microsyntax(node, val, match_nth_child);
+	if (!strcmp(key, "nth-last-child") && val != NULL)
+		return match_an_plus_b_microsyntax(node, val, match_nth_last_child);
+
+	if (!strcmp(key, "first-of-type"))
+		return count_siblings_of_type_before(node) == 0;
+	if (!strcmp(key, "last-of-type"))
+		return count_siblings_of_type_after(node) == 0;
+	if (!strcmp(key, "only-of-type"))
+		return count_siblings_of_type_before(node) == 0 && count_siblings_of_type_after(node) == 0;
+	if (!strcmp(key, "nth-of-type") && val != NULL)
+		return match_an_plus_b_microsyntax(node, val, match_nth_of_type);
+	if (!strcmp(key, "nth-last-of-type") && val != NULL)
+		return match_an_plus_b_microsyntax(node, val, match_nth_last_of_type);
+
+	return 0;
+}
+
 static int
 match_condition(fz_css_condition *cond, fz_xml *node)
 {
@@ -316,7 +653,7 @@ match_condition(fz_css_condition *cond, fz_xml *node)
 
 	switch (cond->type) {
 	default: return 0;
-	case ':': return 0; /* don't support pseudo-classes */
+	case ':': if (!match_pseudo_condition(node, cond->key, cond->val)) return 0; break;
 	case '#': if (!match_att_is_condition(node, "id", cond->val)) return 0; break;
 	case '.': if (!match_att_has_condition(node, "class", cond->val)) return 0; break;
 	case '[': if (!match_att_exists_condition(node, cond->key)) return 0; break;
@@ -465,6 +802,16 @@ add_shorthand_trbl(fz_css_match *match, fz_css_value *value, int spec,
 	}
 }
 
+static void
+add_shorthand_inset(fz_css_match *match, fz_css_value *value, int spec)
+{
+	add_shorthand_trbl(match, value, spec,
+		PRO_INSET_TOP,
+		PRO_INSET_RIGHT,
+		PRO_INSET_BOTTOM,
+		PRO_INSET_LEFT);
+}
+
 static void
 add_shorthand_margin(fz_css_match *match, fz_css_value *value, int spec)
 {
@@ -543,7 +890,7 @@ add_shorthand_border(fz_css_match *match, fz_css_value *value, int spec, int T,
 				if (B) add_property(match, PRO_BORDER_BOTTOM_STYLE, value, spec);
 				if (L) add_property(match, PRO_BORDER_LEFT_STYLE, value, spec);
 			}
-			else if (keyword_in_list(value->data, color_kw, nelem(color_kw)))
+			else if (lookup_named_color(value->data) != NULL)
 			{
 				if (T) add_property(match, PRO_BORDER_TOP_COLOR, value, spec);
 				if (R) add_property(match, PRO_BORDER_RIGHT_COLOR, value, spec);
@@ -648,12 +995,22 @@ add_shorthand_font(fz_css_match *match, fz_css_value *value, int spec)
 		add_property(match, PRO_FONT_WEIGHT, &static_value_normal, spec);
 }
 
+static void
+add_shorthand_background(fz_css_match *match, fz_css_value *value, int spec)
+{
+	/* TODO: background-image, -origin, -size, -repeat */
+	add_property(match, PRO_BACKGROUND_COLOR, value, spec);
+}
+
 static void
 add_property(fz_css_match *match, int name, fz_css_value *value, int spec)
 {
 	/* shorthand expansions: */
 	switch (name)
 	{
+	case PRO_INSET:
+		add_shorthand_inset(match, value, spec);
+		return;
 	case PRO_MARGIN:
 		add_shorthand_margin(match, value, spec);
 		return;
@@ -690,7 +1047,9 @@ add_property(fz_css_match *match, int name, fz_css_value *value, int spec)
 	case PRO_FONT:
 		add_shorthand_font(match, value, spec);
 		return;
-	/* TODO: background */
+	case PRO_BACKGROUND:
+		add_shorthand_background(match, value, spec);
+		return;
 	}
 
 	if (name < NUM_PROPERTIES && match->spec[name] <= spec)
@@ -916,6 +1275,7 @@ is_inheritable_property(int name)
 		name == PRO_WHITE_SPACE ||
 		name == PRO_WIDOWS ||
 		name == PRO_WORD_SPACING ||
+		name == PRO_HYPHENS ||
 		// Strictly speaking, text-decoration is not an inherited property,
 		// but since when drawing an underlined element, all children are also underlined,
 		// we may as well make it inherited.
@@ -925,15 +1285,14 @@ is_inheritable_property(int name)
 static fz_css_value *
 value_from_inheritable_property(fz_css_match *match, int name)
 {
-	fz_css_value *value = match->value[name];
-	if (match->up)
+	while (match)
 	{
-		if (value && !strcmp(value->data, "inherit"))
-			return value_from_inheritable_property(match->up, name);
-		if (!value)
-			return value_from_inheritable_property(match->up, name);
+		fz_css_value *value = match->value[name];
+		if (value && strcmp(value->data, "inherit") != 0)
+			return value;
+		match = match->up;
 	}
-	return value;
+	return NULL;
 }
 
 static fz_css_value *
@@ -1080,21 +1439,48 @@ number_from_property(fz_css_match *match, int property, float initial, int initi
 	return number_from_value(value_from_property(match, property), initial, initial_unit);
 }
 
+static int
+has_number_from_property(fz_css_match *match, int property, float initial, int initial_unit, fz_css_number *num)
+{
+	fz_css_value *val = value_from_property(match, property);
+
+	if (val == NULL)
+		return 0;
+
+	*num = number_from_value(val, initial, initial_unit);
+
+	return 1;
+}
+
+static unsigned int
+has_border_width_from_property(fz_css_match *match, int property, fz_css_number *num)
+{
+	fz_css_value *value = value_from_property(match, property);
+	if (!value)
+	{
+		*num = make_number(2, N_LENGTH); /* initial: 'medium' */
+		return 0;
+	}
+	if (!strcmp(value->data, "thin"))
+		*num = make_number(1, N_LENGTH);
+	else if (!strcmp(value->data, "medium"))
+		*num = make_number(2, N_LENGTH);
+	else if (!strcmp(value->data, "thick"))
+		*num = make_number(4, N_LENGTH);
+	else
+		*num = number_from_value(value, 0, N_LENGTH);
+
+	return 1;
+}
+
 static fz_css_number
 border_width_from_property(fz_css_match *match, int property)
 {
-	fz_css_value *value = value_from_property(match, property);
-	if (value)
-	{
-		if (!strcmp(value->data, "thin"))
-			return make_number(1, N_LENGTH);
-		if (!strcmp(value->data, "medium"))
-			return make_number(2, N_LENGTH);
-		if (!strcmp(value->data, "thick"))
-			return make_number(4, N_LENGTH);
-		return number_from_value(value, 0, N_LENGTH);
-	}
-	return make_number(2, N_LENGTH); /* initial: 'medium' */
+	fz_css_number num;
+
+	(void)has_border_width_from_property(match, property, &num);
+
+	return num;
 }
 
 static int
@@ -1106,6 +1492,13 @@ border_style_from_property(fz_css_match *match, int property)
 		if (!strcmp(value->data, "none")) return BS_NONE;
 		else if (!strcmp(value->data, "hidden")) return BS_NONE;
 		else if (!strcmp(value->data, "solid")) return BS_SOLID;
+		else if (!strcmp(value->data, "dotted")) return BS_DOTTED;
+		else if (!strcmp(value->data, "dashed")) return BS_DASHED;
+		else if (!strcmp(value->data, "double")) return BS_DOUBLE;
+		else if (!strcmp(value->data, "groove")) return BS_GROOVE;
+		else if (!strcmp(value->data, "ridge")) return BS_RIDGE;
+		else if (!strcmp(value->data, "inset")) return BS_INSET;
+		else if (!strcmp(value->data, "outset")) return BS_OUTSET;
 	}
 	return BS_NONE;
 }
@@ -1115,6 +1508,11 @@ int fz_css_number_defined(fz_css_number number)
 	return number.unit != N_UNDEFINED;
 }
 
+int fz_css_number_defined_not_auto(fz_css_number number)
+{
+	return number.unit != N_UNDEFINED && number.unit != N_AUTO;
+}
+
 float
 fz_from_css_number(fz_css_number number, float em, float percent_value, float auto_value)
 {
@@ -1156,9 +1554,68 @@ static int tohex(int c)
 {
 	if (c - '0' < 10)
 		return c - '0';
+	if (c <= 'F')
+		c += 'a' - 'A';
 	return (c | 32) - 'a' + 10;
 }
 
+static size_t
+hexlen(const char *data)
+{
+	size_t n = 0;
+
+	while ((*data >= '0' && *data <= '9') ||
+		(*data >= 'a' && *data <= 'f') ||
+		(*data >= 'A' && *data <= 'F'))
+	{
+		data++, n++;
+	}
+
+	return n;
+}
+
+static fz_css_color
+hash_color(const char *data)
+{
+	int r, g, b, a;
+	size_t n = hexlen(data);
+
+	if (n == 3)
+	{
+		r = tohex(data[0]) * 17;
+		g = tohex(data[1]) * 17;
+		b = tohex(data[2]) * 17;
+		a = 255;
+	}
+	else if (n == 4)
+	{
+		r = tohex(data[0]) * 17;
+		g = tohex(data[1]) * 17;
+		b = tohex(data[2]) * 17;
+		a = tohex(data[3]) * 17;
+	}
+	else if (n == 6)
+	{
+		r = tohex(data[0]) * 16 + tohex(data[1]);
+		g = tohex(data[2]) * 16 + tohex(data[3]);
+		b = tohex(data[4]) * 16 + tohex(data[5]);
+		a = 255;
+	}
+	else if (n == 8)
+	{
+		r = tohex(data[0]) * 16 + tohex(data[1]);
+		g = tohex(data[2]) * 16 + tohex(data[3]);
+		b = tohex(data[4]) * 16 + tohex(data[5]);
+		a = tohex(data[6]) * 16 + tohex(data[7]);
+	}
+	else
+	{
+		r = g = b = 0;
+		a = 255;
+	}
+	return make_color(r, g, b, a);
+}
+
 static fz_css_color
 color_from_value(fz_css_value *value, fz_css_color initial)
 {
@@ -1167,44 +1624,8 @@ color_from_value(fz_css_value *value, fz_css_color initial)
 
 	if (value->type == CSS_HASH)
 	{
-		int r, g, b, a;
-		size_t n;
 hex_color:
-		n = strlen(value->data);
-		if (n == 3)
-		{
-			r = tohex(value->data[0]) * 16 + tohex(value->data[0]);
-			g = tohex(value->data[1]) * 16 + tohex(value->data[1]);
-			b = tohex(value->data[2]) * 16 + tohex(value->data[2]);
-			a = 255;
-		}
-		else if (n == 4)
-		{
-			r = tohex(value->data[0]) * 16 + tohex(value->data[0]);
-			g = tohex(value->data[1]) * 16 + tohex(value->data[1]);
-			b = tohex(value->data[2]) * 16 + tohex(value->data[2]);
-			a = tohex(value->data[3]) * 16 + tohex(value->data[3]);
-		}
-		else if (n == 6)
-		{
-			r = tohex(value->data[0]) * 16 + tohex(value->data[1]);
-			g = tohex(value->data[2]) * 16 + tohex(value->data[3]);
-			b = tohex(value->data[4]) * 16 + tohex(value->data[5]);
-			a = 255;
-		}
-		else if (n == 8)
-		{
-			r = tohex(value->data[0]) * 16 + tohex(value->data[1]);
-			g = tohex(value->data[2]) * 16 + tohex(value->data[3]);
-			b = tohex(value->data[4]) * 16 + tohex(value->data[5]);
-			a = tohex(value->data[6]) * 16 + tohex(value->data[7]);
-		}
-		else
-		{
-			r = g = b = 0;
-			a = 255;
-		}
-		return make_color(r, g, b, a);
+		return hash_color(value->data);
 	}
 
 	if (value->type == '(' && !strcmp(value->data, "rgb"))
@@ -1237,53 +1658,59 @@ hex_color:
 
 	if (value->type == CSS_KEYWORD)
 	{
-		if (!strcmp(value->data, "transparent"))
-			return make_color(0, 0, 0, 0);
-		if (!strcmp(value->data, "maroon"))
-			return make_color(0x80, 0x00, 0x00, 255);
-		if (!strcmp(value->data, "red"))
-			return make_color(0xFF, 0x00, 0x00, 255);
-		if (!strcmp(value->data, "orange"))
-			return make_color(0xFF, 0xA5, 0x00, 255);
-		if (!strcmp(value->data, "yellow"))
-			return make_color(0xFF, 0xFF, 0x00, 255);
-		if (!strcmp(value->data, "olive"))
-			return make_color(0x80, 0x80, 0x00, 255);
-		if (!strcmp(value->data, "purple"))
-			return make_color(0x80, 0x00, 0x80, 255);
-		if (!strcmp(value->data, "fuchsia"))
-			return make_color(0xFF, 0x00, 0xFF, 255);
-		if (!strcmp(value->data, "white"))
-			return make_color(0xFF, 0xFF, 0xFF, 255);
-		if (!strcmp(value->data, "lime"))
-			return make_color(0x00, 0xFF, 0x00, 255);
-		if (!strcmp(value->data, "green"))
-			return make_color(0x00, 0x80, 0x00, 255);
-		if (!strcmp(value->data, "navy"))
-			return make_color(0x00, 0x00, 0x80, 255);
-		if (!strcmp(value->data, "blue"))
-			return make_color(0x00, 0x00, 0xFF, 255);
-		if (!strcmp(value->data, "aqua"))
-			return make_color(0x00, 0xFF, 0xFF, 255);
-		if (!strcmp(value->data, "teal"))
-			return make_color(0x00, 0x80, 0x80, 255);
-		if (!strcmp(value->data, "black"))
-			return make_color(0x00, 0x00, 0x00, 255);
-		if (!strcmp(value->data, "silver"))
-			return make_color(0xC0, 0xC0, 0xC0, 255);
-		if (!strcmp(value->data, "gray"))
-			return make_color(0x80, 0x80, 0x80, 255);
+		const fz_css_color *named = lookup_named_color(value->data);
+		if (named)
+			return *named;
 		goto hex_color; /* last ditch attempt: maybe it's a #XXXXXX color without the # */
 	}
+
 	return initial;
 }
 
+fz_css_color
+fz_css_color_from_string(const char *str)
+{
+	const fz_css_color *named;
+
+	if (*str == '#')
+		return hash_color(str+1);
+	named = lookup_named_color(str);
+	if (named)
+		return *named;
+	return hash_color(str);
+}
+
 static fz_css_color
 color_from_property(fz_css_match *match, int property, fz_css_color initial)
 {
 	return color_from_value(value_from_property(match, property), initial);
 }
 
+static unsigned int
+has_color_from_property(fz_css_match *match, int property, fz_css_color *col)
+{
+	fz_css_color transparent = { 0, 0, 0, 0 };
+	fz_css_value *val = value_from_property(match, property);
+
+	if (val == NULL)
+		return 0;
+
+	*col = color_from_value(val, transparent);
+
+	return 1;
+}
+
+static fz_css_color
+color_from_properties(fz_css_match *match, int property, int property2, fz_css_color initial)
+{
+	fz_css_value *value = value_from_property(match, property);
+
+	if (value == NULL)
+		value = value_from_property(match, property2);
+
+	return color_from_value(value, initial);
+}
+
 int
 fz_get_css_match_display(fz_css_match *match)
 {
@@ -1313,9 +1740,12 @@ fz_get_css_match_display(fz_css_match *match)
 		if (!strcmp(value->data, "table-footer-group"))
 			return DIS_TABLE_GROUP;
 		if (!strcmp(value->data, "table-column-group"))
-			return DIS_NONE;
+			return DIS_TABLE_COLGROUP;
 		if (!strcmp(value->data, "table-column"))
-			return DIS_NONE;
+			return DIS_TABLE_COL;
+		/* Not ideal, but it'll do. */
+		if (!strcmp(value->data, "inline-table"))
+			return DIS_TABLE;
 	}
 	return DIS_INLINE;
 }
@@ -1347,17 +1777,31 @@ text_decoration_from_property(fz_css_match *match)
 	return TD_NONE;
 }
 
+static unsigned int
+has_visibility_from_property(fz_css_match *match, unsigned int *val)
+{
+	fz_css_value *value = value_from_property(match, PRO_VISIBILITY);
+	*val = V_VISIBLE;
+	if (!value)
+		return 0;
+	if (!strcmp(value->data, "visible"))
+		*val = V_VISIBLE;
+	else if (!strcmp(value->data, "hidden"))
+		*val = V_HIDDEN;
+	else if (!strcmp(value->data, "collapse"))
+		*val = V_COLLAPSE;
+
+	return 1;
+}
+
 static int
 visibility_from_property(fz_css_match *match)
 {
-	fz_css_value *value = value_from_property(match, PRO_VISIBILITY);
-	if (value)
-	{
-		if (!strcmp(value->data, "visible")) return V_VISIBLE;
-		else if (!strcmp(value->data, "hidden")) return V_HIDDEN;
-		else if (!strcmp(value->data, "collapse")) return V_COLLAPSE;
-	}
-	return V_VISIBLE;
+	unsigned int val;
+
+	(void)has_visibility_from_property(match, &val);
+
+	return val;
 }
 
 static int
@@ -1379,11 +1823,13 @@ void
 fz_default_css_style(fz_context *ctx, fz_css_style *style)
 {
 	memset(style, 0, sizeof *style);
+	style->direction = FZ_BIDI_UNSET;
 	style->visibility = V_VISIBLE;
 	style->text_align = TA_LEFT;
 	style->vertical_align = VA_BASELINE;
 	style->white_space = WS_NORMAL;
 	style->list_style_type = LST_DISC;
+	style->hyphens = HYP_MANUAL;
 	style->font_size = make_number(1, N_SCALE);
 	style->width = make_number(0, N_AUTO);
 	style->height = make_number(0, N_AUTO);
@@ -1415,6 +1861,15 @@ fz_apply_css_style(fz_context *ctx, fz_html_font_set *set, fz_css_style *style,
 		else if (!strcmp(value->data, "justify")) style->text_align = TA_JUSTIFY;
 	}
 
+	value = value_from_property(match, PRO_POSITION);
+	if (value)
+	{
+		if (!strcmp(value->data, "static")) style->position = POS_STATIC;
+		else if (!strcmp(value->data, "fixed")) style->position = POS_FIXED;
+		else if (!strcmp(value->data, "relative")) style->position = POS_RELATIVE;
+		else if (!strcmp(value->data, "absolute")) style->position = POS_ABSOLUTE;
+	}
+
 	value = value_from_property(match, PRO_VERTICAL_ALIGN);
 	if (value)
 	{
@@ -1481,14 +1936,36 @@ fz_apply_css_style(fz_context *ctx, fz_html_font_set *set, fz_css_style *style,
 		else style->overflow_wrap = OVERFLOW_WRAP_NORMAL;
 	}
 
+	value = value_from_property(match, PRO_BORDER_COLLAPSE);
+	if (value)
+	{
+		if (!strcmp(value->data, "collapse")) style->border_collapse = 1;
+		else style->border_collapse = 0;
+	}
+
 	style->line_height = number_from_property(match, PRO_LINE_HEIGHT, 1.2f, N_SCALE);
 	style->leading = number_from_property(match, PRO_LEADING, 0, N_UNDEFINED);
 
 	style->text_indent = number_from_property(match, PRO_TEXT_INDENT, 0, N_LENGTH);
+	style->text_stroke_width = number_from_property(match, PRO_TEXT_STROKE_WIDTH, 0, N_LENGTH);
 
 	style->width = number_from_property(match, PRO_WIDTH, 0, N_AUTO);
 	style->height = number_from_property(match, PRO_HEIGHT, 0, N_AUTO);
 
+	value = value_from_property(match, PRO_DIRECTION);
+	if (value)
+	{
+		if (!strcmp(value->data, "rtl")) style->direction = FZ_BIDI_RTL;
+	}
+
+	value = value_from_property(match, PRO_HYPHENS);
+	if (value)
+	{
+		if (!strcmp(value->data, "none")) style->hyphens = HYP_NONE;
+		else if (!strcmp(value->data, "manual")) style->hyphens = HYP_MANUAL;
+		else if (!strcmp(value->data, "auto")) style->hyphens = HYP_AUTO;
+	}
+
 	style->margin[0] = number_from_property(match, PRO_MARGIN_TOP, 0, N_LENGTH);
 	style->margin[1] = number_from_property(match, PRO_MARGIN_RIGHT, 0, N_LENGTH);
 	style->margin[2] = number_from_property(match, PRO_MARGIN_BOTTOM, 0, N_LENGTH);
@@ -1499,7 +1976,14 @@ fz_apply_css_style(fz_context *ctx, fz_html_font_set *set, fz_css_style *style,
 	style->padding[2] = number_from_property(match, PRO_PADDING_BOTTOM, 0, N_LENGTH);
 	style->padding[3] = number_from_property(match, PRO_PADDING_LEFT, 0, N_LENGTH);
 
+	style->inset[0] = number_from_property(match, PRO_INSET_TOP, 0, N_UNDEFINED);
+	style->inset[1] = number_from_property(match, PRO_INSET_RIGHT, 0, N_UNDEFINED);
+	style->inset[2] = number_from_property(match, PRO_INSET_BOTTOM, 0, N_UNDEFINED);
+	style->inset[3] = number_from_property(match, PRO_INSET_LEFT, 0, N_UNDEFINED);
+
 	style->color = color_from_property(match, PRO_COLOR, black);
+	style->text_fill_color = color_from_properties(match, PRO_TEXT_FILL_COLOR, PRO_COLOR, black);
+	style->text_stroke_color = color_from_property(match, PRO_TEXT_STROKE_COLOR, transparent);
 	style->background_color = color_from_property(match, PRO_BACKGROUND_COLOR, transparent);
 
 	style->border_spacing = number_from_property(match, PRO_BORDER_SPACING, 0, N_LENGTH);
@@ -1542,6 +2026,24 @@ fz_apply_css_style(fz_context *ctx, fz_html_font_set *set, fz_css_style *style,
 	}
 }
 
+void
+fz_css_colstyle(col_style *cs, fz_css_match *match)
+{
+	cs->has_bg_col = has_color_from_property(match, PRO_BACKGROUND_COLOR, &cs->background_color);
+	cs->has_border_col = has_color_from_property(match, PRO_BORDER_TOP_COLOR, &cs->border_color[0]);
+	cs->has_border_col |= has_color_from_property(match, PRO_BORDER_RIGHT_COLOR, &cs->border_color[1])<<1;
+	cs->has_border_col |= has_color_from_property(match, PRO_BORDER_BOTTOM_COLOR, &cs->border_color[2])<<2;
+	cs->has_border_col |= has_color_from_property(match, PRO_BORDER_LEFT_COLOR, &cs->border_color[3])<<3;
+
+	cs->has_border_width = has_border_width_from_property(match, PRO_BORDER_TOP_WIDTH, &cs->border_width[0]);
+	cs->has_border_width |= has_border_width_from_property(match, PRO_BORDER_RIGHT_WIDTH, &cs->border_width[1])<<1;
+	cs->has_border_width |= has_border_width_from_property(match, PRO_BORDER_BOTTOM_WIDTH, &cs->border_width[2])<<2;
+	cs->has_border_width |= has_border_width_from_property(match, PRO_BORDER_LEFT_WIDTH, &cs->border_width[3])<<3;
+
+	cs->has_visibility = has_visibility_from_property(match, &cs->visibility);
+	cs->has_width = has_number_from_property(match, PRO_WIDTH, 0, N_AUTO, &cs->width);
+}
+
 #ifdef DEBUG_CSS_SPLAY
 static void
 do_verify_splay(const fz_css_style_splay *x)
@@ -1761,6 +2263,10 @@ static void print_condition(fz_css_condition *cond)
 		printf("[%s=%s]", cond->key, cond->val);
 	else if (cond->type == '[')
 		printf("[%s]", cond->key);
+	else if (cond->type == ':' && cond->val)
+		printf(":%s(%s)", cond->key, cond->val);
+	else if (cond->type == ':' && !cond->val)
+		printf(":%s", cond->key);
 	else
 		printf("%c%s", cond->type, cond->val);
 	if (cond->next)
