@@ -32,12 +32,12 @@
 #line 1 "source/html/css-properties.gperf"
 struct css_property_info { const char *name; int key; };
 
-#define TOTAL_KEYWORDS 65
+#define TOTAL_KEYWORDS 80
 #define MIN_WORD_LENGTH 3
-#define MAX_WORD_LENGTH 19
-#define MIN_HASH_VALUE 10
-#define MAX_HASH_VALUE 140
-/* maximum key range = 131, duplicates = 0 */
+#define MAX_WORD_LENGTH 25
+#define MIN_HASH_VALUE 8
+#define MAX_HASH_VALUE 189
+/* maximum key range = 182, duplicates = 0 */
 
 #ifdef __GNUC__
 __inline
@@ -51,199 +51,231 @@ css_property_hash (register const char *str, register size_t len)
 {
   static unsigned char asso_values[] =
     {
-      141, 141, 141, 141, 141, 141, 141, 141, 141, 141,
-      141, 141, 141, 141, 141, 141, 141, 141, 141, 141,
-      141, 141, 141, 141, 141, 141, 141, 141, 141, 141,
-      141, 141, 141, 141, 141, 141, 141, 141, 141, 141,
-      141, 141, 141, 141, 141,   0, 141, 141, 141, 141,
-      141, 141, 141, 141, 141, 141, 141, 141, 141, 141,
-      141, 141, 141, 141, 141, 141, 141, 141, 141, 141,
-      141, 141, 141, 141, 141, 141, 141, 141, 141, 141,
-      141, 141, 141, 141, 141, 141, 141, 141, 141, 141,
-      141, 141, 141, 141, 141, 141, 141,  35,   5,   5,
-        5,  10,  65,  85,  55,  75, 141, 141,   0,  20,
-        5,   0,   0,   0,   0,  35,  20,   0,   0,   5,
-      141,  35, 141, 141, 141, 141, 141, 141, 141, 141,
-      141, 141, 141, 141, 141, 141, 141, 141, 141, 141,
-      141, 141, 141, 141, 141, 141, 141, 141, 141, 141,
-      141, 141, 141, 141, 141, 141, 141, 141, 141, 141,
-      141, 141, 141, 141, 141, 141, 141, 141, 141, 141,
-      141, 141, 141, 141, 141, 141, 141, 141, 141, 141,
-      141, 141, 141, 141, 141, 141, 141, 141, 141, 141,
-      141, 141, 141, 141, 141, 141, 141, 141, 141, 141,
-      141, 141, 141, 141, 141, 141, 141, 141, 141, 141,
-      141, 141, 141, 141, 141, 141, 141, 141, 141, 141,
-      141, 141, 141, 141, 141, 141, 141, 141, 141, 141,
-      141, 141, 141, 141, 141, 141, 141, 141, 141, 141,
-      141, 141, 141, 141, 141, 141, 141, 141, 141, 141,
-      141, 141, 141, 141, 141, 141
+      190, 190, 190, 190, 190, 190, 190, 190, 190, 190,
+      190, 190, 190, 190, 190, 190, 190, 190, 190, 190,
+      190, 190, 190, 190, 190, 190, 190, 190, 190, 190,
+      190, 190, 190, 190, 190, 190, 190, 190, 190, 190,
+      190, 190, 190, 190, 190,  35, 190, 190, 190, 190,
+      190, 190, 190, 190, 190, 190, 190, 190, 190, 190,
+      190, 190, 190, 190, 190, 190, 190, 190, 190, 190,
+      190, 190, 190, 190, 190, 190, 190, 190, 190, 190,
+      190, 190, 190, 190, 190, 190, 190, 190, 190, 190,
+      190, 190, 190, 190, 190, 190, 190,  25,   5,  40,
+       60,   5, 110,  50,  75,  70, 190, 190,   0,  25,
+       50,   5,   0,   5,  30,  35,   0,   5,  25,   5,
+      190,   5, 190, 190, 190, 190, 190, 190, 190, 190,
+      190, 190, 190, 190, 190, 190, 190, 190, 190, 190,
+      190, 190, 190, 190, 190, 190, 190, 190, 190, 190,
+      190, 190, 190, 190, 190, 190, 190, 190, 190, 190,
+      190, 190, 190, 190, 190, 190, 190, 190, 190, 190,
+      190, 190, 190, 190, 190, 190, 190, 190, 190, 190,
+      190, 190, 190, 190, 190, 190, 190, 190, 190, 190,
+      190, 190, 190, 190, 190, 190, 190, 190, 190, 190,
+      190, 190, 190, 190, 190, 190, 190, 190, 190, 190,
+      190, 190, 190, 190, 190, 190, 190, 190, 190, 190,
+      190, 190, 190, 190, 190, 190, 190, 190, 190, 190,
+      190, 190, 190, 190, 190, 190, 190, 190, 190, 190,
+      190, 190, 190, 190, 190, 190, 190, 190, 190, 190,
+      190, 190, 190, 190, 190, 190
     };
   return len + asso_values[(unsigned char)str[1]] + asso_values[(unsigned char)str[0]] + asso_values[(unsigned char)str[len - 1]];
 }
 
 static struct css_property_info css_property_list[] =
   {
-    {""}, {""}, {""}, {""}, {""}, {""}, {""}, {""}, {""},
-    {""},
-#line 31 "source/html/css-properties.gperf"
-    {"color",PRO_COLOR},
-#line 10 "source/html/css-properties.gperf"
-    {"border",PRO_BORDER},
-    {""},
-#line 53 "source/html/css-properties.gperf"
-    {"overflow-wrap",PRO_OVERFLOW_WRAP},
-    {""},
-#line 26 "source/html/css-properties.gperf"
-    {"border-top",PRO_BORDER_TOP},
-    {""},
-#line 15 "source/html/css-properties.gperf"
-    {"border-color",PRO_BORDER_COLOR},
+    {""}, {""}, {""}, {""}, {""}, {""}, {""}, {""},
+#line 81 "source/html/css-properties.gperf"
+    {"top",PRO_INSET_TOP},
+#line 52 "source/html/css-properties.gperf"
+    {"left",PRO_INSET_LEFT},
+    {""}, {""}, {""}, {""}, {""}, {""},
+#line 79 "source/html/css-properties.gperf"
+    {"text-indent",PRO_TEXT_INDENT},
     {""}, {""}, {""},
-#line 27 "source/html/css-properties.gperf"
-    {"border-top-color",PRO_BORDER_TOP_COLOR},
-#line 17 "source/html/css-properties.gperf"
-    {"border-left-color",PRO_BORDER_LEFT_COLOR},
+#line 31 "source/html/css-properties.gperf"
+    {"border-top",PRO_BORDER_TOP},
 #line 21 "source/html/css-properties.gperf"
-    {"border-right-color",PRO_BORDER_RIGHT_COLOR},
-#line 12 "source/html/css-properties.gperf"
-    {"border-bottom-color",PRO_BORDER_BOTTOM_COLOR},
-    {""}, {""},
+    {"border-left",PRO_BORDER_LEFT},
 #line 25 "source/html/css-properties.gperf"
+    {"border-right",PRO_BORDER_RIGHT},
+    {""}, {""}, {""}, {""},
+#line 30 "source/html/css-properties.gperf"
     {"border-style",PRO_BORDER_STYLE},
-    {""},
-#line 67 "source/html/css-properties.gperf"
-    {"vertical-align",PRO_VERTICAL_ALIGN},
-    {""},
-#line 28 "source/html/css-properties.gperf"
+    {""}, {""},
+#line 20 "source/html/css-properties.gperf"
+    {"border-collapse",PRO_BORDER_COLLAPSE},
+#line 33 "source/html/css-properties.gperf"
     {"border-top-style",PRO_BORDER_TOP_STYLE},
-#line 18 "source/html/css-properties.gperf"
+#line 23 "source/html/css-properties.gperf"
     {"border-left-style",PRO_BORDER_LEFT_STYLE},
-#line 22 "source/html/css-properties.gperf"
+#line 27 "source/html/css-properties.gperf"
     {"border-right-style",PRO_BORDER_RIGHT_STYLE},
-#line 13 "source/html/css-properties.gperf"
+#line 17 "source/html/css-properties.gperf"
     {"border-bottom-style",PRO_BORDER_BOTTOM_STYLE},
     {""},
-#line 16 "source/html/css-properties.gperf"
-    {"border-left",PRO_BORDER_LEFT},
-#line 20 "source/html/css-properties.gperf"
-    {"border-right",PRO_BORDER_RIGHT},
-#line 11 "source/html/css-properties.gperf"
-    {"border-bottom",PRO_BORDER_BOTTOM},
-    {""}, {""},
-#line 61 "source/html/css-properties.gperf"
-    {"quotes",PRO_QUOTES},
-#line 52 "source/html/css-properties.gperf"
-    {"orphans",PRO_ORPHANS},
-#line 62 "source/html/css-properties.gperf"
-    {"src",PRO_SRC},
-    {""},
-#line 63 "source/html/css-properties.gperf"
-    {"text-align",PRO_TEXT_ALIGN},
-#line 58 "source/html/css-properties.gperf"
+#line 70 "source/html/css-properties.gperf"
     {"padding-top",PRO_PADDING_TOP},
-    {""}, {""}, {""},
-#line 64 "source/html/css-properties.gperf"
-    {"text-decoration",PRO_TEXT_DECORATION},
-#line 59 "source/html/css-properties.gperf"
-    {"page-break-after",PRO_PAGE_BREAK_AFTER},
-    {""}, {""}, {""}, {""},
-#line 9 "source/html/css-properties.gperf"
-    {"background-color",PRO_BACKGROUND_COLOR},
-    {""}, {""}, {""}, {""},
-#line 65 "source/html/css-properties.gperf"
-    {"text-indent",PRO_TEXT_INDENT},
-#line 60 "source/html/css-properties.gperf"
-    {"page-break-before",PRO_PAGE_BREAK_BEFORE},
-    {""},
-#line 66 "source/html/css-properties.gperf"
-    {"text-transform",PRO_TEXT_TRANSFORM},
-#line 51 "source/html/css-properties.gperf"
-    {"margin-top",PRO_MARGIN_TOP},
-#line 47 "source/html/css-properties.gperf"
-    {"margin",PRO_MARGIN},
-#line 56 "source/html/css-properties.gperf"
+#line 68 "source/html/css-properties.gperf"
     {"padding-left",PRO_PADDING_LEFT},
-#line 57 "source/html/css-properties.gperf"
-    {"padding-right",PRO_PADDING_RIGHT},
-#line 55 "source/html/css-properties.gperf"
-    {"padding-bottom",PRO_PADDING_BOTTOM},
-    {""}, {""},
-#line 30 "source/html/css-properties.gperf"
-    {"border-width",PRO_BORDER_WIDTH},
-    {""}, {""}, {""},
-#line 29 "source/html/css-properties.gperf"
-    {"border-top-width",PRO_BORDER_TOP_WIDTH},
-#line 19 "source/html/css-properties.gperf"
-    {"border-left-width",PRO_BORDER_LEFT_WIDTH},
-#line 23 "source/html/css-properties.gperf"
-    {"border-right-width",PRO_BORDER_RIGHT_WIDTH},
-#line 14 "source/html/css-properties.gperf"
-    {"border-bottom-width",PRO_BORDER_BOTTOM_WIDTH},
-    {""},
 #line 69 "source/html/css-properties.gperf"
-    {"white-space",PRO_WHITE_SPACE},
+    {"padding-right",PRO_PADDING_RIGHT},
     {""}, {""},
 #line 36 "source/html/css-properties.gperf"
-    {"font-size",PRO_FONT_SIZE},
-#line 37 "source/html/css-properties.gperf"
-    {"font-style",PRO_FONT_STYLE},
-#line 49 "source/html/css-properties.gperf"
-    {"margin-left",PRO_MARGIN_LEFT},
-#line 50 "source/html/css-properties.gperf"
-    {"margin-right",PRO_MARGIN_RIGHT},
-#line 48 "source/html/css-properties.gperf"
-    {"margin-bottom",PRO_MARGIN_BOTTOM},
-#line 34 "source/html/css-properties.gperf"
-    {"font",PRO_FONT},
+    {"bottom",PRO_INSET_BOTTOM},
     {""},
-#line 40 "source/html/css-properties.gperf"
-    {"height",PRO_HEIGHT},
+#line 65 "source/html/css-properties.gperf"
+    {"overflow-wrap",PRO_OVERFLOW_WRAP},
+#line 80 "source/html/css-properties.gperf"
+    {"text-transform",PRO_TEXT_TRANSFORM},
+    {""},
+#line 14 "source/html/css-properties.gperf"
+    {"border",PRO_BORDER},
+#line 72 "source/html/css-properties.gperf"
+    {"page-break-before",PRO_PAGE_BREAK_BEFORE},
+#line 15 "source/html/css-properties.gperf"
+    {"border-bottom",PRO_BORDER_BOTTOM},
     {""}, {""},
+#line 74 "source/html/css-properties.gperf"
+    {"quotes",PRO_QUOTES},
+#line 19 "source/html/css-properties.gperf"
+    {"border-color",PRO_BORDER_COLOR},
+    {""}, {""}, {""},
 #line 32 "source/html/css-properties.gperf"
-    {"direction",PRO_DIRECTION},
-#line 43 "source/html/css-properties.gperf"
-    {"list-style",PRO_LIST_STYLE},
-#line 39 "source/html/css-properties.gperf"
-    {"font-weight",PRO_FONT_WEIGHT},
+    {"border-top-color",PRO_BORDER_TOP_COLOR},
+#line 22 "source/html/css-properties.gperf"
+    {"border-left-color",PRO_BORDER_LEFT_COLOR},
+#line 26 "source/html/css-properties.gperf"
+    {"border-right-color",PRO_BORDER_RIGHT_COLOR},
+#line 16 "source/html/css-properties.gperf"
+    {"border-bottom-color",PRO_BORDER_BOTTOM_COLOR},
+#line 63 "source/html/css-properties.gperf"
+    {"margin-top",PRO_MARGIN_TOP},
+#line 61 "source/html/css-properties.gperf"
+    {"margin-left",PRO_MARGIN_LEFT},
+#line 62 "source/html/css-properties.gperf"
+    {"margin-right",PRO_MARGIN_RIGHT},
+#line 73 "source/html/css-properties.gperf"
+    {"position",PRO_POSITION},
+#line 67 "source/html/css-properties.gperf"
+    {"padding-bottom",PRO_PADDING_BOTTOM},
+#line 77 "source/html/css-properties.gperf"
+    {"text-align",PRO_TEXT_ALIGN},
+    {""}, {""}, {""},
+#line 53 "source/html/css-properties.gperf"
+    {"letter-spacing",PRO_LETTER_SPACING},
+#line 78 "source/html/css-properties.gperf"
+    {"text-decoration",PRO_TEXT_DECORATION},
+#line 71 "source/html/css-properties.gperf"
+    {"page-break-after",PRO_PAGE_BREAK_AFTER},
+#line 87 "source/html/css-properties.gperf"
+    {"word-spacing",PRO_WORD_SPACING},
+    {""},
+#line 29 "source/html/css-properties.gperf"
+    {"border-spacing",PRO_BORDER_SPACING},
+#line 37 "source/html/css-properties.gperf"
+    {"clear",PRO_CLEAR},
+#line 13 "source/html/css-properties.gperf"
+    {"background-color",PRO_BACKGROUND_COLOR},
+#line 64 "source/html/css-properties.gperf"
+    {"orphans",PRO_ORPHANS},
+    {""}, {""},
 #line 38 "source/html/css-properties.gperf"
+    {"color",PRO_COLOR},
+#line 54 "source/html/css-properties.gperf"
+    {"line-height",PRO_LINE_HEIGHT},
+#line 66 "source/html/css-properties.gperf"
+    {"padding",PRO_PADDING},
+    {""}, {""},
+#line 55 "source/html/css-properties.gperf"
+    {"list-style",PRO_LIST_STYLE},
+#line 49 "source/html/css-properties.gperf"
+    {"height",PRO_HEIGHT},
+#line 39 "source/html/css-properties.gperf"
+    {"columns",PRO_COLUMNS},
+#line 60 "source/html/css-properties.gperf"
+    {"margin-bottom",PRO_MARGIN_BOTTOM},
+    {""},
+#line 58 "source/html/css-properties.gperf"
+    {"list-style-type",PRO_LIST_STYLE_TYPE},
+#line 56 "source/html/css-properties.gperf"
+    {"list-style-image",PRO_LIST_STYLE_IMAGE},
+    {""},
+#line 9 "source/html/css-properties.gperf"
+    {"-webkit-text-fill-color",PRO_TEXT_FILL_COLOR},
+#line 82 "source/html/css-properties.gperf"
+    {"vertical-align",PRO_VERTICAL_ALIGN},
+#line 10 "source/html/css-properties.gperf"
+    {"-webkit-text-stroke-color",PRO_TEXT_STROKE_COLOR},
+#line 84 "source/html/css-properties.gperf"
+    {"white-space",PRO_WHITE_SPACE},
+#line 35 "source/html/css-properties.gperf"
+    {"border-width",PRO_BORDER_WIDTH},
+    {""}, {""},
+#line 12 "source/html/css-properties.gperf"
+    {"background",PRO_BACKGROUND},
+#line 34 "source/html/css-properties.gperf"
+    {"border-top-width",PRO_BORDER_TOP_WIDTH},
+#line 24 "source/html/css-properties.gperf"
+    {"border-left-width",PRO_BORDER_LEFT_WIDTH},
+#line 28 "source/html/css-properties.gperf"
+    {"border-right-width",PRO_BORDER_RIGHT_WIDTH},
+#line 18 "source/html/css-properties.gperf"
+    {"border-bottom-width",PRO_BORDER_BOTTOM_WIDTH},
+#line 75 "source/html/css-properties.gperf"
+    {"right",PRO_INSET_RIGHT},
+#line 59 "source/html/css-properties.gperf"
+    {"margin",PRO_MARGIN},
+    {""},
+#line 76 "source/html/css-properties.gperf"
+    {"src",PRO_SRC},
+    {""},
+#line 83 "source/html/css-properties.gperf"
+    {"visibility",PRO_VISIBILITY},
+    {""}, {""}, {""}, {""},
+#line 42 "source/html/css-properties.gperf"
+    {"float",PRO_FLOAT},
+#line 85 "source/html/css-properties.gperf"
+    {"widows",PRO_WIDOWS},
+    {""}, {""},
+#line 43 "source/html/css-properties.gperf"
+    {"font",PRO_FONT},
+    {""}, {""},
+#line 50 "source/html/css-properties.gperf"
+    {"hyphens",PRO_HYPHENS},
+    {""},
+#line 8 "source/html/css-properties.gperf"
+    {"-mupdf-leading",PRO_LEADING},
+#line 51 "source/html/css-properties.gperf"
+    {"inset",PRO_INSET},
+#line 48 "source/html/css-properties.gperf"
+    {"font-weight",PRO_FONT_WEIGHT},
+#line 47 "source/html/css-properties.gperf"
     {"font-variant",PRO_FONT_VARIANT},
     {""},
 #line 45 "source/html/css-properties.gperf"
-    {"list-style-position",PRO_LIST_STYLE_POSITION},
+    {"font-size",PRO_FONT_SIZE},
 #line 46 "source/html/css-properties.gperf"
-    {"list-style-type",PRO_LIST_STYLE_TYPE},
+    {"font-style",PRO_FONT_STYLE},
 #line 44 "source/html/css-properties.gperf"
-    {"list-style-image",PRO_LIST_STYLE_IMAGE},
-#line 72 "source/html/css-properties.gperf"
-    {"word-spacing",PRO_WORD_SPACING},
-    {""},
-#line 24 "source/html/css-properties.gperf"
-    {"border-spacing",PRO_BORDER_SPACING},
-    {""},
-#line 42 "source/html/css-properties.gperf"
-    {"line-height",PRO_LINE_HEIGHT},
-    {""}, {""},
-#line 41 "source/html/css-properties.gperf"
-    {"letter-spacing",PRO_LETTER_SPACING},
-    {""},
-#line 35 "source/html/css-properties.gperf"
     {"font-family",PRO_FONT_FAMILY},
     {""}, {""}, {""}, {""}, {""}, {""}, {""},
-#line 8 "source/html/css-properties.gperf"
-    {"-mupdf-leading",PRO_LEADING},
-#line 68 "source/html/css-properties.gperf"
-    {"visibility",PRO_VISIBILITY},
-#line 70 "source/html/css-properties.gperf"
-    {"widows",PRO_WIDOWS},
-#line 33 "source/html/css-properties.gperf"
+#line 57 "source/html/css-properties.gperf"
+    {"list-style-position",PRO_LIST_STYLE_POSITION},
+#line 11 "source/html/css-properties.gperf"
+    {"-webkit-text-stroke-width",PRO_TEXT_STROKE_WIDTH},
+    {""},
+#line 41 "source/html/css-properties.gperf"
     {"display",PRO_DISPLAY},
-    {""}, {""}, {""}, {""},
-#line 54 "source/html/css-properties.gperf"
-    {"padding",PRO_PADDING},
     {""}, {""}, {""}, {""}, {""}, {""}, {""}, {""}, {""},
     {""}, {""}, {""},
-#line 71 "source/html/css-properties.gperf"
-    {"width",PRO_WIDTH}
+#line 86 "source/html/css-properties.gperf"
+    {"width",PRO_WIDTH},
+    {""}, {""}, {""}, {""}, {""}, {""}, {""}, {""}, {""},
+    {""}, {""}, {""}, {""}, {""}, {""}, {""}, {""}, {""},
+    {""}, {""}, {""}, {""}, {""}, {""}, {""}, {""}, {""},
+    {""}, {""}, {""}, {""}, {""}, {""},
+#line 40 "source/html/css-properties.gperf"
+    {"direction",PRO_DIRECTION}
   };
 
 struct css_property_info *
