@@ -1,4 +1,4 @@
-// Copyright (C) 2004-2024 Artifex Software, Inc.
+// Copyright (C) 2004-2025 Artifex Software, Inc.
 //
 // This file is part of MuPDF.
 //
@@ -35,12 +35,16 @@ pdf_filter_type3(fz_context *ctx, pdf_document *doc, pdf_obj *obj, pdf_obj *page
 static void
 pdf_filter_resources(fz_context *ctx, pdf_document *doc, pdf_obj *in_res, pdf_obj *res, pdf_filter_options *options, pdf_cycle_list *cycle_up)
 {
+	pdf_cycle_list cycle;
 	pdf_obj *obj;
 	int i, n;
 
 	if (!options->recurse)
 		return;
 
+	if (pdf_cycle(ctx, &cycle, cycle_up, in_res))
+		return;
+
 	/* ExtGState */
 	obj = pdf_dict_get(ctx, res, PDF_NAME(ExtGState));
 	if (obj)
@@ -55,7 +59,7 @@ pdf_filter_resources(fz_context *ctx, pdf_document *doc, pdf_obj *in_res, pdf_ob
 				if (g)
 				{
 					/* Transparency group XObject */
-					pdf_filter_xobject(ctx, doc, g, in_res, options, cycle_up);
+					pdf_filter_xobject(ctx, doc, g, in_res, options, &cycle);
 				}
 			}
 		}
@@ -71,7 +75,7 @@ pdf_filter_resources(fz_context *ctx, pdf_document *doc, pdf_obj *in_res, pdf_ob
 			pdf_obj *pat = pdf_dict_get_val(ctx, obj, i);
 			if (pat && pdf_dict_get_int(ctx, pat, PDF_NAME(PatternType)) == 1)
 			{
-				pdf_filter_xobject(ctx, doc, pat, in_res, options, cycle_up);
+				pdf_filter_xobject(ctx, doc, pat, in_res, options, &cycle);
 			}
 		}
 	}
@@ -88,7 +92,7 @@ pdf_filter_resources(fz_context *ctx, pdf_document *doc, pdf_obj *in_res, pdf_ob
 				pdf_obj *xobj = pdf_dict_get_val(ctx, obj, i);
 				if (xobj && pdf_dict_get(ctx, xobj, PDF_NAME(Subtype)) == PDF_NAME(Form))
 				{
-					pdf_filter_xobject(ctx, doc, xobj, in_res, options, cycle_up);
+					pdf_filter_xobject(ctx, doc, xobj, in_res, options, &cycle);
 				}
 			}
 		}
@@ -104,7 +108,7 @@ pdf_filter_resources(fz_context *ctx, pdf_document *doc, pdf_obj *in_res, pdf_ob
 			pdf_obj *font = pdf_dict_get_val(ctx, obj, i);
 			if (font && pdf_dict_get(ctx, font, PDF_NAME(Subtype)) == PDF_NAME(Type3))
 			{
-				pdf_filter_type3(ctx, doc, font, in_res, options, cycle_up);
+				pdf_filter_type3(ctx, doc, font, in_res, options, &cycle);
 			}
 		}
 	}
@@ -213,6 +217,7 @@ pdf_filter_type3(fz_context *ctx, pdf_document *doc, pdf_obj *obj, pdf_obj *page
 	fz_var(buffer);
 	fz_var(res);
 	fz_var(new_buf);
+	fz_var(top);
 
 	/* We cannot combine instancing with type3 fonts. The new names for
 	 * instanced form/image resources would clash, since they start over for
@@ -255,10 +260,10 @@ pdf_filter_type3(fz_context *ctx, pdf_document *doc, pdf_obj *obj, pdf_obj *page
 
 			if (i > 0)
 			{
+				// reset all chained processors (and clear the buffer)
 				pdf_reset_processor(ctx, top);
-				fz_clear_buffer(ctx, buffer);
 			}
-			pdf_process_raw_contents(ctx, top, doc, in_res, val, NULL);
+			pdf_process_raw_contents(ctx, top, doc, val, NULL);
 
 			pdf_close_processor(ctx, top);
 
@@ -361,7 +366,7 @@ pdf_filter_xobject_instance(fz_context *ctx, pdf_obj *old_xobj, pdf_obj *page_re
 		return pdf_keep_obj(ctx, old_xobj);
 
 	matrix = pdf_dict_get_matrix(ctx, old_xobj, PDF_NAME(Matrix));
-	transform = fz_concat(transform, matrix);
+	transform = fz_concat(matrix, transform);
 
 	fz_try(ctx)
 	{
@@ -454,6 +459,28 @@ struct redact_filter_state {
 	int text;
 };
 
+
+static void pdf_run_obj_to_buf(fz_context *ctx, fz_buffer *buffer, pdf_obj *obj, pdf_page *page)
+{
+	pdf_processor *proc = pdf_new_buffer_processor(ctx, buffer, 0, 0);
+	pdf_obj *res;
+
+
+	fz_try(ctx)
+	{
+		res = pdf_xobject_resources(ctx, obj);
+		if (res == NULL)
+			res = pdf_page_resources(ctx, page);
+
+		pdf_process_contents(ctx, proc, page->doc, res, obj, NULL, NULL);
+		pdf_close_processor(ctx, proc);
+	}
+	fz_always(ctx)
+		pdf_drop_processor(ctx, proc);
+	fz_catch(ctx)
+		fz_rethrow(ctx);
+}
+
 static void
 pdf_redact_end_page(fz_context *ctx, fz_buffer *buf, void *opaque)
 {
@@ -471,35 +498,43 @@ pdf_redact_end_page(fz_context *ctx, fz_buffer *buf, void *opaque)
 			continue;
 		if (pdf_dict_get(ctx, annot->obj, PDF_NAME(Subtype)) == PDF_NAME(Redact))
 		{
-			qp = pdf_dict_get(ctx, annot->obj, PDF_NAME(QuadPoints));
-			n = pdf_array_len(ctx, qp);
-			if (n > 0)
+			pdf_obj *ro = pdf_dict_get(ctx, annot->obj, PDF_NAME(RO));
+			if (ro)
 			{
-				for (i = 0; i < n; i += 8)
-				{
-					fz_quad q = pdf_to_quad(ctx, qp, i);
-					fz_append_printf(ctx, buf, "%g %g m\n", q.ll.x, q.ll.y);
-					fz_append_printf(ctx, buf, "%g %g l\n", q.lr.x, q.lr.y);
-					fz_append_printf(ctx, buf, "%g %g l\n", q.ur.x, q.ur.y);
-					fz_append_printf(ctx, buf, "%g %g l\n", q.ul.x, q.ul.y);
-					fz_append_string(ctx, buf, "f\n");
-				}
+				pdf_run_obj_to_buf(ctx, buf, ro, page);
 			}
 			else
 			{
-				fz_rect r = pdf_dict_get_rect(ctx, annot->obj, PDF_NAME(Rect));
-				fz_append_printf(ctx, buf, "%g %g m\n", r.x0, r.y0);
-				fz_append_printf(ctx, buf, "%g %g l\n", r.x1, r.y0);
-				fz_append_printf(ctx, buf, "%g %g l\n", r.x1, r.y1);
-				fz_append_printf(ctx, buf, "%g %g l\n", r.x0, r.y1);
-				fz_append_string(ctx, buf, "f\n");
+				qp = pdf_dict_get(ctx, annot->obj, PDF_NAME(QuadPoints));
+				n = pdf_array_len(ctx, qp);
+				if (n > 0)
+				{
+					for (i = 0; i < n; i += 8)
+					{
+						fz_quad q = pdf_to_quad(ctx, qp, i);
+						fz_append_printf(ctx, buf, "%g %g m\n", q.ll.x, q.ll.y);
+						fz_append_printf(ctx, buf, "%g %g l\n", q.lr.x, q.lr.y);
+						fz_append_printf(ctx, buf, "%g %g l\n", q.ur.x, q.ur.y);
+						fz_append_printf(ctx, buf, "%g %g l\n", q.ul.x, q.ul.y);
+						fz_append_string(ctx, buf, "f\n");
+					}
+				}
+				else
+				{
+					fz_rect r = pdf_dict_get_rect(ctx, annot->obj, PDF_NAME(Rect));
+					fz_append_printf(ctx, buf, "%g %g m\n", r.x0, r.y0);
+					fz_append_printf(ctx, buf, "%g %g l\n", r.x1, r.y0);
+					fz_append_printf(ctx, buf, "%g %g l\n", r.x1, r.y1);
+					fz_append_printf(ctx, buf, "%g %g l\n", r.x0, r.y1);
+					fz_append_string(ctx, buf, "f\n");
+				}
 			}
 		}
 	}
 }
 
 static int
-pdf_redact_text_filter(fz_context *ctx, void *opaque, int *ucsbuf, int ucslen, fz_matrix trm, fz_matrix ctm, fz_rect bbox)
+pdf_redact_text_filter(fz_context *ctx, void *opaque, int *ucsbuf, int ucslen, fz_matrix trm, fz_matrix ctm, fz_rect bbox, int tr, float ca, float CA)
 {
 	struct redact_filter_state *red = opaque;
 	pdf_page *page = red->page;
@@ -555,6 +590,33 @@ pdf_redact_text_filter(fz_context *ctx, void *opaque, int *ucsbuf, int ucslen, f
 	return 0;
 }
 
+static int
+pdf_redact_invisible_text_filter(fz_context *ctx, void *opaque, int *ucsbuf, int ucslen, fz_matrix trm, fz_matrix ctm, fz_rect bbox, int tr, float ca, float CA)
+{
+	int invisible = 0;
+
+	switch (tr)
+	{
+		case 0: /* Fill */
+			invisible = (ca == 0);
+			break;
+		case 1: /* Stroke */
+			invisible = (CA == 0);
+			break;
+		case 2: /* Fill + Stroke */
+			invisible = (ca == 0 && CA == 0);
+			break;
+		case 3: /* Neither Fill nor stroke */
+			invisible = 1;
+			break;
+	}
+
+	if (!invisible)
+		return 0;
+
+	return pdf_redact_text_filter(ctx, opaque, ucsbuf, ucslen, trm, ctm, bbox, tr, ca, CA);
+}
+
 static fz_pixmap *
 pdf_redact_image_imp(fz_context *ctx, fz_matrix ctm, fz_image *image, fz_pixmap *pixmap, fz_pixmap **pmask, fz_quad q)
 {
@@ -921,6 +983,28 @@ rect_touches_redactions(fz_context *ctx, fz_rect area, struct redact_filter_stat
 	return 0;
 }
 
+static void
+remove_page_link(fz_context *ctx, pdf_page *page, pdf_obj *obj)
+{
+	pdf_link **linkp = (pdf_link **)&page->links;
+	pdf_link *link;
+
+	while ((link = *linkp) != NULL)
+	{
+		if (link->obj == obj)
+		{
+			*linkp = (pdf_link *)link->super.next;
+			link->super.next = NULL;
+			fz_drop_link(ctx, &link->super);
+			break;
+		}
+		else
+		{
+			linkp = (pdf_link **)&link->super.next;
+		}
+	}
+}
+
 static void
 pdf_redact_page_links(fz_context *ctx, struct redact_filter_state *red)
 {
@@ -940,6 +1024,7 @@ pdf_redact_page_links(fz_context *ctx, struct redact_filter_state *red)
 			if (rect_touches_redactions(ctx, area, red))
 			{
 				pdf_array_delete(ctx, annots, k);
+				remove_page_link(ctx, red->page, link);
 				continue;
 			}
 		}
@@ -977,6 +1062,9 @@ static int culler(fz_context *ctx, void *opaque, fz_rect bbox, fz_cull_type type
 	case FZ_CULL_PATH_FILL:
 	case FZ_CULL_PATH_STROKE:
 	case FZ_CULL_PATH_FILL_STROKE:
+	case FZ_CULL_CLIP_PATH_FILL:
+	case FZ_CULL_CLIP_PATH_STROKE:
+	case FZ_CULL_CLIP_PATH_FILL_STROKE:
 		if (red->line_art == PDF_REDACT_LINE_ART_REMOVE_IF_COVERED)
 			return (rect_touches_redactions(ctx, bbox, red) == 2);
 		else if (red->line_art == PDF_REDACT_LINE_ART_REMOVE_IF_TOUCHED)
@@ -1011,6 +1099,8 @@ void init_redact_filter(fz_context *ctx, pdf_redact_options *redact_opts, struct
 	red->sanitize_opts.opaque = red;
 	if (text == PDF_REDACT_TEXT_REMOVE)
 		red->sanitize_opts.text_filter = pdf_redact_text_filter;
+	if (text == PDF_REDACT_TEXT_REMOVE_INVISIBLE)
+		red->sanitize_opts.text_filter = pdf_redact_invisible_text_filter;
 	if (image_method == PDF_REDACT_IMAGE_PIXELS)
 		red->sanitize_opts.image_filter = pdf_redact_image_filter_pixels;
 	if (image_method == PDF_REDACT_IMAGE_REMOVE)
@@ -1025,7 +1115,7 @@ void init_redact_filter(fz_context *ctx, pdf_redact_options *redact_opts, struct
 	red->filter_list[1].options = NULL;
 
 	red->page = page;
-	red->target = NULL;
+	red->target = target;
 }
 
 static int
@@ -1098,3 +1188,191 @@ pdf_apply_redaction(fz_context *ctx, pdf_annot *annot, pdf_redact_options *redac
 {
 	return pdf_apply_redaction_imp(ctx, annot->page, annot, redact_opts);
 }
+
+/* Hard clipping of pages */
+
+struct clip_filter_state {
+	pdf_filter_options filter_opts;
+	pdf_sanitize_filter_options sanitize_opts;
+	pdf_filter_factory filter_list[2];
+	pdf_page *page;
+	fz_rect clip;
+};
+
+static int clip_culler(fz_context *ctx, void *opaque, fz_rect bbox, fz_cull_type type)
+{
+	struct clip_filter_state *hc = opaque;
+
+	switch (type)
+	{
+	case FZ_CULL_PATH_FILL:
+	case FZ_CULL_PATH_STROKE:
+	case FZ_CULL_PATH_FILL_STROKE:
+	case FZ_CULL_CLIP_PATH_FILL:
+	case FZ_CULL_CLIP_PATH_STROKE:
+	case FZ_CULL_CLIP_PATH_FILL_STROKE:
+	case FZ_CULL_GLYPH:
+	case FZ_CULL_IMAGE:
+	case FZ_CULL_SHADING:
+		return (fz_is_empty_rect(fz_intersect_rect(bbox, hc->clip)));
+	default:
+		return 0;
+	}
+}
+
+static
+void init_clip_filter(fz_context *ctx, struct clip_filter_state *hc, pdf_page *page, fz_rect *clip)
+{
+	memset(&hc->filter_opts, 0, sizeof hc->filter_opts);
+	memset(&hc->sanitize_opts, 0, sizeof hc->sanitize_opts);
+
+	hc->filter_opts.recurse = 0; /* don't redact patterns, softmasks, and type3 fonts */
+	hc->filter_opts.instance_forms = 1; /* redact xobjects with instancing */
+	hc->filter_opts.ascii = 0;
+	hc->filter_opts.opaque = hc;
+	hc->filter_opts.filters = hc->filter_list;
+	hc->clip = *clip;
+
+	hc->sanitize_opts.opaque = hc;
+	hc->sanitize_opts.culler = clip_culler;
+
+	hc->filter_list[0].filter = pdf_new_sanitize_filter;
+	hc->filter_list[0].options = &hc->sanitize_opts;
+	hc->filter_list[1].filter = NULL;
+	hc->filter_list[1].options = NULL;
+
+	hc->page = page;
+}
+
+static void
+pdf_clip_page_links(fz_context *ctx, struct clip_filter_state *hc)
+{
+	pdf_obj *annots;
+	pdf_obj *link;
+	fz_rect area;
+	int k;
+
+	annots = pdf_dict_get(ctx, hc->page->obj, PDF_NAME(Annots));
+	k = 0;
+	while (k < pdf_array_len(ctx, annots))
+	{
+		link = pdf_array_get(ctx, annots, k);
+		if (pdf_dict_get(ctx, link, PDF_NAME(Subtype)) == PDF_NAME(Link))
+		{
+			area = pdf_dict_get_rect(ctx, link, PDF_NAME(Rect));
+			if (fz_is_empty_rect(fz_intersect_rect(area, hc->clip)))
+			{
+				pdf_array_delete(ctx, annots, k);
+				continue;
+			}
+		}
+		++k;
+	}
+}
+
+static void
+pdf_clip_page_annotations(fz_context *ctx, struct clip_filter_state *hc)
+{
+	pdf_annot *annot;
+	fz_rect area;
+
+restart:
+	for (annot = pdf_first_annot(ctx, hc->page); annot; annot = pdf_next_annot(ctx, annot))
+	{
+		if (pdf_annot_type(ctx, annot) == PDF_ANNOT_FREE_TEXT)
+		{
+			area = pdf_dict_get_rect(ctx, pdf_annot_obj(ctx, annot), PDF_NAME(Rect));
+			if (fz_is_empty_rect(fz_intersect_rect(area, hc->clip)))
+			{
+				pdf_delete_annot(ctx, hc->page, annot);
+				goto restart;
+			}
+		}
+	}
+}
+
+void
+pdf_clip_page(fz_context *ctx, pdf_page *page, fz_rect *clip)
+{
+	pdf_document *doc;
+	struct clip_filter_state hc;
+
+	if (page == NULL)
+		return;
+
+	doc = page->doc;
+
+	init_clip_filter(ctx, &hc, page, clip);
+
+	pdf_begin_operation(ctx, doc, "Apply hard clip to page");
+	fz_try(ctx)
+	{
+		pdf_filter_page_contents(ctx, doc, page, &hc.filter_opts);
+		pdf_clip_page_links(ctx, &hc);
+		pdf_clip_page_annotations(ctx, &hc);
+		pdf_end_operation(ctx, doc);
+	}
+	fz_catch(ctx)
+	{
+		pdf_abandon_operation(ctx, doc);
+		fz_rethrow(ctx);
+	}
+}
+
+/* Vectorisation of pages */
+
+struct vectorize_filter_state {
+	pdf_filter_options filter_opts;
+	pdf_vectorize_filter_options vectorize_opts;
+	pdf_filter_factory filter_list[2];
+	pdf_page *page;
+};
+
+static
+void init_vectorize_filter(fz_context *ctx, struct vectorize_filter_state *hc, pdf_page *page)
+{
+	memset(&hc->filter_opts, 0, sizeof hc->filter_opts);
+	memset(&hc->vectorize_opts, 0, sizeof hc->vectorize_opts);
+
+	hc->filter_opts.recurse = 0;
+	hc->filter_opts.instance_forms = 0;
+	hc->filter_opts.ascii = 0;
+	hc->filter_opts.opaque = hc;
+	hc->filter_opts.filters = hc->filter_list;
+	hc->filter_opts.recurse = 1;
+
+	hc->vectorize_opts.opaque = hc;
+
+	hc->filter_list[0].filter = pdf_new_vectorize_filter;
+	hc->filter_list[0].options = &hc->vectorize_opts;
+	hc->filter_list[1].filter = NULL;
+	hc->filter_list[1].options = NULL;
+
+	hc->page = page;
+}
+
+void
+pdf_vectorize_page(fz_context *ctx, pdf_page *page)
+{
+	pdf_document *doc;
+	struct vectorize_filter_state hv;
+
+	if (page == NULL)
+		return;
+
+	doc = page->doc;
+
+	init_vectorize_filter(ctx, &hv, page);
+
+	pdf_begin_operation(ctx, doc, "Vectorize text to page");
+	fz_try(ctx)
+	{
+		pdf_filter_page_contents(ctx, doc, page, &hv.filter_opts);
+		pdf_end_operation(ctx, doc);
+	}
+	fz_catch(ctx)
+	{
+		pdf_abandon_operation(ctx, doc);
+		fz_rethrow(ctx);
+	}
+}
