@@ -1,4 +1,4 @@
-// Copyright (C) 2004-2024 Artifex Software, Inc.
+// Copyright (C) 2004-2025 Artifex Software, Inc.
 //
 // This file is part of MuPDF.
 //
@@ -83,13 +83,13 @@ typedef struct
 	pdf_obj super;
 	char *text; /* utf8 encoded text string */
 	size_t len;
-	char buf[1];
+	char buf[FZ_FLEXIBLE_ARRAY];
 } pdf_obj_string;
 
 typedef struct
 {
 	pdf_obj super;
-	char n[1];
+	char n[FZ_FLEXIBLE_ARRAY];
 } pdf_obj_name;
 
 typedef struct
@@ -659,13 +659,14 @@ do_objcmp(fz_context *ctx, pdf_obj *a, pdf_obj *b, int check_streams)
 		 * Otherwise, do the full, painful, comparison. */
 		{
 			/* Slightly convoluted to know if something is a stream. */
-			pdf_document *doc = DICT(a)->doc;
+			pdf_document *doca = DICT(a)->doc;
+			pdf_document *docb = DICT(b)->doc;
 			int ap = pdf_obj_parent_num(ctx, a);
 			int bp;
 			int a_is_stream = 0;
-			pdf_xref_entry *entrya = pdf_get_xref_entry_no_change(ctx, doc, ap);
+			pdf_xref_entry *entrya = pdf_get_xref_entry_no_change(ctx, doca, ap);
 			pdf_xref_entry *entryb;
-			if (entrya != NULL && entrya->obj == a && pdf_obj_num_is_stream(ctx, doc, ap))
+			if (entrya != NULL && entrya->obj == a && pdf_obj_num_is_stream(ctx, doca, ap))
 			{
 				/* It's a stream, and we know a != b from above. */
 				if (!check_streams)
@@ -673,8 +674,8 @@ do_objcmp(fz_context *ctx, pdf_obj *a, pdf_obj *b, int check_streams)
 				a_is_stream = 1;
 			}
 			bp = pdf_obj_parent_num(ctx, b);
-			entryb = pdf_get_xref_entry_no_change(ctx, doc, bp);
-			if (entryb != NULL && entryb->obj == b && pdf_obj_num_is_stream(ctx, doc, bp))
+			entryb = pdf_get_xref_entry_no_change(ctx, docb, bp);
+			if (entryb != NULL && entryb->obj == b && pdf_obj_num_is_stream(ctx, docb, bp))
 			{
 				/* It's a stream, and we know a != b from above. So mismatch. */
 				if (!check_streams || !a_is_stream)
@@ -699,8 +700,8 @@ do_objcmp(fz_context *ctx, pdf_obj *a, pdf_obj *b, int check_streams)
 				{
 					unsigned char *dataa, *datab;
 					size_t lena, lenb;
-					sa = pdf_load_raw_stream_number(ctx, doc, ap);
-					sb = pdf_load_raw_stream_number(ctx, doc, bp);
+					sa = pdf_load_raw_stream_number(ctx, doca, ap);
+					sb = pdf_load_raw_stream_number(ctx, docb, bp);
 					lena = fz_buffer_storage(ctx, sa, &dataa);
 					lenb = fz_buffer_storage(ctx, sb, &datab);
 					if (lena == lenb && memcmp(dataa, datab, lena) == 0)
@@ -718,8 +719,6 @@ do_objcmp(fz_context *ctx, pdf_obj *a, pdf_obj *b, int check_streams)
 				return differ;
 			}
 		}
-
-		return 0;
 	}
 	return 1;
 }
@@ -1055,7 +1054,7 @@ dump_changes(fz_context *ctx, pdf_document *doc, pdf_journal_entry *entry)
 }
 #endif
 
-/* We build up journal entries as being a list of changes (framents) that
+/* We build up journal entries as being a list of changes (fragments) that
  * happen all together as part of a single step. When we reach pdf_end_operation
  * we have all the changes that have happened during this operation in a list
  * that basically boils down to being:
@@ -1352,7 +1351,7 @@ swap_fragments(fz_context *ctx, pdf_document *doc, pdf_journal_entry *entry)
 		obuf = xre->stm_buf;
 		xre->obj = frag->inactive;
 		type = xre->type;
-		xre->type = frag->newobj ? 0 : 'o';
+		xre->type = frag->newobj ? 0 : 'n';
 		frag->newobj = type == 0;
 		xre->stm_buf = frag->stream;
 		frag->inactive = old;
@@ -1422,15 +1421,15 @@ void pdf_undo(fz_context *ctx, pdf_document *doc)
 	fz_write_printf(ctx, fz_stddbg(ctx), "Undo!\n");
 #endif
 
+	doc->journal->current = entry->prev;
+
+	swap_fragments(ctx, doc, entry);
+
 	// nuke all caches
 	pdf_drop_page_tree_internal(ctx, doc);
 	pdf_sync_open_pages(ctx, doc);
 	for (frag = entry->head; frag; frag = frag->next)
 		pdf_purge_object_from_store(ctx, doc, frag->obj_num);
-
-	doc->journal->current = entry->prev;
-
-	swap_fragments(ctx, doc, entry);
 }
 
 /* Move forwards in the undo history. Throws an error if we are at the
@@ -1468,15 +1467,15 @@ void pdf_redo(fz_context *ctx, pdf_document *doc)
 	if (entry == NULL)
 		fz_throw(ctx, FZ_ERROR_ARGUMENT, "Already at end of history");
 
+	doc->journal->current = entry;
+
+	swap_fragments(ctx, doc, entry);
+
 	// nuke all caches
 	pdf_drop_page_tree_internal(ctx, doc);
 	pdf_sync_open_pages(ctx, doc);
 	for (frag = entry->head; frag; frag = frag->next)
 		pdf_purge_object_from_store(ctx, doc, frag->obj_num);
-
-	doc->journal->current = entry;
-
-	swap_fragments(ctx, doc, entry);
 }
 
 void pdf_discard_journal(fz_context *ctx, pdf_journal *journal)
@@ -1708,8 +1707,8 @@ void pdf_deserialise_journal(fz_context *ctx, pdf_document *doc, fz_stream *stm)
 		if (fz_skip_string(ctx, stm, "entry\n") == 0)
 		{
 			/* Read the fragment title. */
-			pdf_token tok = pdf_lex(ctx, stm, &doc->lexbuf.base);
 			char *title;
+			tok = pdf_lex(ctx, stm, &doc->lexbuf.base);
 
 			if (tok != PDF_TOK_STRING)
 				fz_throw(ctx, FZ_ERROR_FORMAT, "Bad string in journal");
@@ -2467,7 +2466,7 @@ pdf_obj *
 pdf_dict_geta(fz_context *ctx, pdf_obj *obj, pdf_obj *key, pdf_obj *abbrev)
 {
 	pdf_obj *v;
-	/* ISO 32000-2:2020 (PDF 2.0) - abbreviated names take precendence. */
+	/* ISO 32000-2:2020 (PDF 2.0) - abbreviated names take precedence. */
 	v = pdf_dict_get(ctx, obj, abbrev);
 	if (v)
 		return v;
@@ -3682,7 +3681,8 @@ pdf_sprint_encrypted_obj(fz_context *ctx, char *buf, size_t cap, size_t *len, pd
 	}
 	fz_catch(ctx)
 	{
-		fz_free(ctx, fmt.ptr);
+		if (!buf || cap == 0)
+			fz_free(ctx, fmt.ptr);
 		fz_rethrow(ctx);
 	}
 
