@@ -1,4 +1,4 @@
-// Copyright (C) 2004-2024 Artifex Software, Inc.
+// Copyright (C) 2004-2025 Artifex Software, Inc.
 //
 // This file is part of MuPDF.
 //
@@ -96,12 +96,6 @@ struct pdf_gstate
 	int luminosity;
 };
 
-typedef struct resources_stack
-{
-	struct resources_stack *next;
-	pdf_obj *resources;
-} resources_stack;
-
 typedef struct marked_content_stack
 {
 	struct marked_content_stack *next;
@@ -125,8 +119,6 @@ struct pdf_run_processor
 
 	fz_default_colorspaces *default_cs;
 
-	resources_stack *rstack;
-
 	/* path object state */
 	fz_path *path;
 	int clip;
@@ -160,6 +152,12 @@ struct pdf_run_processor
 	begin_layer_stack **next_begin_layer;
 
 	int mc_depth;
+	/* The nest_mark array holds a record of the way in which clips and
+	 * marked content are nested to ensure we pop stuff in the same order
+	 * that we push it - i.e. to keep calls nested nicely. An entry x,
+	 * where x >= 0 represents that a push has happened for mc_depth == x.
+	 * An entry x, where x < 0 means that -x clips have happened at this
+	 * position. */
 	int nest_depth;
 	int nest_mark[1024];
 };
@@ -208,13 +206,20 @@ flush_begin_layer(fz_context *ctx, pdf_run_processor *proc)
 	proc->next_begin_layer = &proc->begin_layer;
 }
 
-#define CLIP_MARK -1
-
 static void nest_layer_clip(fz_context *ctx, pdf_run_processor *proc)
 {
 	if (proc->nest_depth == nelem(proc->nest_mark))
 		fz_throw(ctx, FZ_ERROR_LIMIT, "layer/clip nesting too deep");
-	proc->nest_mark[proc->nest_depth++] = CLIP_MARK;
+	if (proc->nest_depth > 0 && proc->nest_mark[proc->nest_depth-1] < 0)
+	{
+		/* The last mark was a clip. Just increase that count. */
+		proc->nest_mark[proc->nest_depth-1]--;
+	}
+	else
+	{
+		/* Create a new entry for a single clip. */
+		proc->nest_mark[proc->nest_depth++] = -1;
+	}
 }
 
 static void
@@ -255,7 +260,7 @@ load_transfer_function(fz_context *ctx, pdf_obj *obj)
 }
 
 static pdf_gstate *
-begin_softmask(fz_context *ctx, pdf_run_processor *pr, softmask_save *save)
+begin_softmask(fz_context *ctx, pdf_run_processor *pr, softmask_save *save, fz_rect bbox)
 {
 	pdf_gstate *gstate = pr->gstate + pr->gtop;
 	pdf_obj *softmask = gstate->softmask;
@@ -266,6 +271,8 @@ begin_softmask(fz_context *ctx, pdf_run_processor *pr, softmask_save *save)
 	fz_colorspace *mask_colorspace;
 	int saved_blendmode;
 	fz_function *tr = NULL;
+	float save_alpha_fill;
+	float save_alpha_stroke;
 
 	fz_var(tr);
 
@@ -293,10 +300,15 @@ begin_softmask(fz_context *ctx, pdf_run_processor *pr, softmask_save *save)
 		mask_bbox = fz_transform_rect(mask_bbox, mask_matrix);
 		mask_bbox = fz_transform_rect(mask_bbox, gstate->softmask_ctm);
 	}
+	mask_bbox = fz_intersect_rect(mask_bbox, bbox);
 	gstate->softmask = NULL;
 	gstate->softmask_cs = NULL;
 	gstate->softmask_resources = NULL;
 	gstate->ctm = gstate->softmask_ctm;
+	save_alpha_fill = gstate->fill.alpha;
+	save_alpha_stroke = gstate->stroke.alpha;
+	gstate->fill.alpha = 1;
+	gstate->stroke.alpha = 1;
 
 	saved_blendmode = gstate->blendmode;
 
@@ -327,6 +339,8 @@ begin_softmask(fz_context *ctx, pdf_run_processor *pr, softmask_save *save)
 
 	gstate = pr->gstate + pr->gtop;
 	gstate->ctm = save_ctm;
+	gstate->fill.alpha = save_alpha_fill;
+	gstate->stroke.alpha = save_alpha_stroke;
 
 	return gstate;
 }
@@ -352,7 +366,7 @@ end_softmask(fz_context *ctx, pdf_run_processor *pr, softmask_save *save)
 static pdf_gstate *
 pdf_begin_group(fz_context *ctx, pdf_run_processor *pr, fz_rect bbox, softmask_save *softmask)
 {
-	pdf_gstate *gstate = begin_softmask(ctx, pr, softmask);
+	pdf_gstate *gstate = begin_softmask(ctx, pr, softmask, bbox);
 
 	if (gstate->blendmode)
 		fz_begin_group(ctx, pr->dev, bbox, NULL, 0, 0, gstate->blendmode, 1);
@@ -406,11 +420,11 @@ static pdf_material *
 pdf_keep_material(fz_context *ctx, pdf_material *mat)
 {
 	if (mat->colorspace)
-		fz_keep_colorspace(ctx, mat->colorspace);
+		mat->colorspace = fz_keep_colorspace(ctx, mat->colorspace);
 	if (mat->pattern)
-		pdf_keep_pattern(ctx, mat->pattern);
+		mat->pattern = pdf_keep_pattern(ctx, mat->pattern);
 	if (mat->shade)
-		fz_keep_shade(ctx, mat->shade);
+		mat->shade = fz_keep_shade(ctx, mat->shade);
 	return mat;
 }
 
@@ -418,33 +432,52 @@ static pdf_material *
 pdf_drop_material(fz_context *ctx, pdf_material *mat)
 {
 	fz_drop_colorspace(ctx, mat->colorspace);
+	mat->colorspace = NULL;
 	pdf_drop_pattern(ctx, mat->pattern);
+	mat->pattern = NULL;
 	fz_drop_shade(ctx, mat->shade);
+	mat->shade = NULL;
 	return mat;
 }
 
 static void
 pdf_copy_pattern_gstate(fz_context *ctx, pdf_gstate *dst, const pdf_gstate *src)
 {
-	pdf_font_desc *old_font = dst->text.font;
-
 	dst->ctm = src->ctm;
 
+	// stroke state
+	fz_drop_stroke_state(ctx, dst->stroke_state);
+	dst->stroke_state = fz_keep_stroke_state(ctx, src->stroke_state);
+
+	// text state
+	pdf_drop_font(ctx, dst->text.font);
+	fz_drop_string(ctx, dst->text.fontname);
 	dst->text = src->text;
 	pdf_keep_font(ctx, src->text.font);
-	pdf_drop_font(ctx, old_font);
+	fz_keep_string(ctx, src->text.fontname);
 
+	// extgstate
+	dst->fill.alpha = src->fill.alpha;
+	dst->stroke.alpha = src->stroke.alpha;
+
+	// transparency
+	dst->blendmode = src->blendmode;
 	pdf_drop_obj(ctx, dst->softmask);
 	dst->softmask = pdf_keep_obj(ctx, src->softmask);
 
 	pdf_drop_obj(ctx, dst->softmask_resources);
 	dst->softmask_resources = pdf_keep_obj(ctx, src->softmask_resources);
 
+	pdf_drop_obj(ctx, dst->softmask_tr);
+	dst->softmask_tr = pdf_keep_obj(ctx, src->softmask_tr);
+
+	dst->softmask_ctm = src->softmask_ctm;
+
 	fz_drop_colorspace(ctx, dst->softmask_cs);
 	dst->softmask_cs = fz_keep_colorspace(ctx, src->softmask_cs);
+	memcpy(dst->softmask_bc, src->softmask_bc, sizeof dst->softmask_bc);
 
-	fz_drop_stroke_state(ctx, dst->stroke_state);
-	dst->stroke_state = fz_keep_stroke_state(ctx, src->stroke_state);
+	dst->luminosity = src->luminosity;
 }
 
 static void
@@ -467,15 +500,15 @@ pdf_keep_gstate(fz_context *ctx, pdf_gstate *gs)
 	pdf_keep_material(ctx, &gs->stroke);
 	pdf_keep_material(ctx, &gs->fill);
 	if (gs->text.font)
-		pdf_keep_font(ctx, gs->text.font);
+		gs->text.font = pdf_keep_font(ctx, gs->text.font);
 	if (gs->softmask)
-		pdf_keep_obj(ctx, gs->softmask);
+		gs->softmask = pdf_keep_obj(ctx, gs->softmask);
 	if (gs->softmask_cs)
-		fz_keep_colorspace(ctx, gs->softmask_cs);
+		gs->softmask_cs = fz_keep_colorspace(ctx, gs->softmask_cs);
 	if (gs->softmask_resources)
-		pdf_keep_obj(ctx, gs->softmask_resources);
-	fz_keep_stroke_state(ctx, gs->stroke_state);
-	pdf_keep_obj(ctx, gs->softmask_tr);
+		gs->softmask_resources = pdf_keep_obj(ctx, gs->softmask_resources);
+	gs->stroke_state = fz_keep_stroke_state(ctx, gs->stroke_state);
+	gs->softmask_tr = pdf_keep_obj(ctx, gs->softmask_tr);
 }
 
 static void
@@ -484,11 +517,17 @@ pdf_drop_gstate(fz_context *ctx, pdf_gstate *gs)
 	pdf_drop_material(ctx, &gs->stroke);
 	pdf_drop_material(ctx, &gs->fill);
 	pdf_drop_font(ctx, gs->text.font);
+	gs->text.font = NULL;
 	pdf_drop_obj(ctx, gs->softmask);
+	gs->softmask = NULL;
 	fz_drop_colorspace(ctx, gs->softmask_cs);
+	gs->softmask_cs = NULL;
 	pdf_drop_obj(ctx, gs->softmask_resources);
+	gs->softmask_resources = NULL;
 	fz_drop_stroke_state(ctx, gs->stroke_state);
+	gs->stroke_state = NULL;
 	pdf_drop_obj(ctx, gs->softmask_tr);
+	gs->softmask_tr = NULL;
 }
 
 static void
@@ -530,14 +569,21 @@ pdf_grestore(fz_context *ctx, pdf_run_processor *pr)
 		fz_try(ctx)
 		{
 			// End layer early (before Q) if unbalanced Q appears between BMC and EMC.
-			while (pr->nest_depth > 0 && pr->nest_mark[pr->nest_depth-1] != CLIP_MARK)
+			while (pr->nest_depth > 0 && pr->nest_mark[pr->nest_depth-1] >= 0)
 			{
 				fz_end_layer(ctx, pr->dev);
 				pr->nest_depth--;
 			}
 
-			fz_pop_clip(ctx, pr->dev);
-			pr->nest_depth--;
+			if (pr->nest_depth > 0)
+			{
+				/* So this one must be a clip record. */
+				fz_pop_clip(ctx, pr->dev);
+				/* Pop a single clip record off. */
+				pr->nest_mark[pr->nest_depth-1]++;
+				if (pr->nest_mark[pr->nest_depth-1] == 0)
+					pr->nest_depth--;
+			}
 
 			// End layer late (after Q) if unbalanced EMC appears between q and Q.
 			while (pr->nest_depth > 0 && pr->nest_mark[pr->nest_depth-1] > pr->mc_depth)
@@ -609,6 +655,10 @@ pdf_show_pattern(fz_context *ctx, pdf_run_processor *pr, pdf_pattern *pat, int p
 	{
 		pdf_drop_obj(ctx, gstate->softmask);
 		gstate->softmask = NULL;
+		pdf_drop_obj(ctx, gstate->softmask_resources);
+		gstate->softmask_resources = NULL;
+		fz_drop_colorspace(ctx, gstate->softmask_cs);
+		gstate->softmask_cs = NULL;
 	}
 
 	ptm = fz_concat(pat->matrix, pat_gstate->ctm);
@@ -648,7 +698,7 @@ pdf_show_pattern(fz_context *ctx, pdf_run_processor *pr, pdf_pattern *pat, int p
 	if (0)
 #endif
 	{
-		int cached = fz_begin_tile_id(ctx, pr->dev, local_area, pat->bbox, pat->xstep, pat->ystep, ptm, id);
+		int cached = fz_begin_tile_tid(ctx, pr->dev, local_area, pat->bbox, pat->xstep, pat->ystep, ptm, id, pat->document->super.id);
 		if (!cached)
 		{
 			gstate->ctm = ptm;
@@ -724,14 +774,18 @@ static void
 pdf_show_image_imp(fz_context *ctx, pdf_run_processor *pr, fz_image *image, fz_matrix image_ctm, fz_rect bbox)
 {
 	pdf_gstate *gstate = pr->gstate + pr->gtop;
+	fz_color_params cp = gstate->fill.color_params;
+
+	if (image->has_intent)
+		cp.ri = image->intent;
 
 	if (image->colorspace)
 	{
-		fz_fill_image(ctx, pr->dev, image, image_ctm, gstate->fill.alpha, gstate->fill.color_params);
+		fz_fill_image(ctx, pr->dev, image, image_ctm, gstate->fill.alpha, cp);
 	}
 	else if (gstate->fill.kind == PDF_MAT_COLOR)
 	{
-		fz_fill_image_mask(ctx, pr->dev, image, image_ctm, gstate->fill.colorspace, gstate->fill.v, gstate->fill.alpha, gstate->fill.color_params);
+		fz_fill_image_mask(ctx, pr->dev, image, image_ctm, gstate->fill.colorspace, gstate->fill.v, gstate->fill.alpha, cp);
 	}
 	else if (gstate->fill.kind == PDF_MAT_PATTERN && gstate->fill.pattern)
 	{
@@ -742,7 +796,7 @@ pdf_show_image_imp(fz_context *ctx, pdf_run_processor *pr, fz_image *image, fz_m
 	else if (gstate->fill.kind == PDF_MAT_SHADE && gstate->fill.shade)
 	{
 		fz_clip_image_mask(ctx, pr->dev, image, image_ctm, bbox);
-		fz_fill_shade(ctx, pr->dev, gstate->fill.shade, pr->gstate[gstate->fill.gstate_num].ctm, gstate->fill.alpha, gstate->fill.color_params);
+		fz_fill_shade(ctx, pr->dev, gstate->fill.shade, pr->gstate[gstate->fill.gstate_num].ctm, gstate->fill.alpha, cp);
 		fz_pop_clip(ctx, pr->dev);
 	}
 }
@@ -817,7 +871,7 @@ pdf_show_path(fz_context *ctx, pdf_run_processor *pr, int doclose, int dofill, i
 	flush_begin_layer(ctx, pr);
 
 	if (dostroke) {
-		if (pr->dev->flags & (FZ_DEVFLAG_STROKECOLOR_UNDEFINED | FZ_DEVFLAG_LINEJOIN_UNDEFINED | FZ_DEVFLAG_LINEWIDTH_UNDEFINED))
+		if (pr->dev->flags & (FZ_DEVFLAG_STROKECOLOR_UNDEFINED | FZ_DEVFLAG_LINEJOIN_UNDEFINED | FZ_DEVFLAG_LINEWIDTH_UNDEFINED | FZ_DEVFLAG_DASH_PATTERN_UNDEFINED))
 			pr->dev->flags |= FZ_DEVFLAG_UNCACHEABLE;
 		else if (gstate->stroke_state->dash_len != 0 && pr->dev->flags & (FZ_DEVFLAG_STARTCAP_UNDEFINED | FZ_DEVFLAG_DASHCAP_UNDEFINED | FZ_DEVFLAG_ENDCAP_UNDEFINED))
 			pr->dev->flags |= FZ_DEVFLAG_UNCACHEABLE;
@@ -874,12 +928,17 @@ pdf_show_path(fz_context *ctx, pdf_run_processor *pr, int doclose, int dofill, i
 					gstate->fill.colorspace, gstate->fill.v, gstate->fill.alpha, gstate->fill.color_params);
 				break;
 			case PDF_MAT_PATTERN:
-				if (gstate->fill.pattern)
-				{
-					fz_clip_path(ctx, pr->dev, path, even_odd, gstate->ctm, bbox);
-					gstate = pdf_show_pattern(ctx, pr, gstate->fill.pattern, gstate->fill.gstate_num, bbox, PDF_FILL);
-					fz_pop_clip(ctx, pr->dev);
-				}
+				if (!gstate->fill.pattern)
+					break;
+				if (gstate->fill.alpha == 0)
+					break;
+				if (gstate->fill.alpha != 1)
+					fz_begin_group(ctx, pr->dev, bbox, NULL, 0, 0, FZ_BLEND_NORMAL, gstate->fill.alpha);
+				fz_clip_path(ctx, pr->dev, path, even_odd, gstate->ctm, bbox);
+				gstate = pdf_show_pattern(ctx, pr, gstate->fill.pattern, gstate->fill.gstate_num, bbox, PDF_FILL);
+				fz_pop_clip(ctx, pr->dev);
+				if (gstate->fill.alpha != 1)
+					fz_end_group(ctx, pr->dev);
 				break;
 			case PDF_MAT_SHADE:
 				if (gstate->fill.shade)
@@ -904,12 +963,17 @@ pdf_show_path(fz_context *ctx, pdf_run_processor *pr, int doclose, int dofill, i
 					gstate->stroke.colorspace, gstate->stroke.v, gstate->stroke.alpha, gstate->stroke.color_params);
 				break;
 			case PDF_MAT_PATTERN:
-				if (gstate->stroke.pattern)
-				{
-					fz_clip_stroke_path(ctx, pr->dev, path, gstate->stroke_state, gstate->ctm, bbox);
-					gstate = pdf_show_pattern(ctx, pr, gstate->stroke.pattern, gstate->stroke.gstate_num, bbox, PDF_STROKE);
-					fz_pop_clip(ctx, pr->dev);
-				}
+				if (!gstate->stroke.pattern)
+					break;
+				if (gstate->stroke.alpha == 0)
+					break;
+				if (gstate->stroke.alpha != 1)
+					fz_begin_group(ctx, pr->dev, bbox, NULL, 0, 0, FZ_BLEND_NORMAL, gstate->stroke.alpha);
+				fz_clip_stroke_path(ctx, pr->dev, path, gstate->stroke_state, gstate->ctm, bbox);
+				gstate = pdf_show_pattern(ctx, pr, gstate->stroke.pattern, gstate->stroke.gstate_num, bbox, PDF_STROKE);
+				fz_pop_clip(ctx, pr->dev);
+				if (gstate->stroke.alpha != 1)
+					fz_end_group(ctx, pr->dev);
 				break;
 			case PDF_MAT_SHADE:
 				if (gstate->stroke.shade)
@@ -1157,30 +1221,58 @@ pdf_show_char(fz_context *ctx, pdf_run_processor *pr, int cid, fz_text_language
 	int ucslen;
 	int i;
 	int render_direct;
+	int type3_hitr;
 
 	gid = pdf_tos_make_trm(ctx, &pr->tos, &gstate->text, fontdesc, cid, &trm, &adv);
 
 	/* If we are uncachable, then render direct. */
 	render_direct = !fz_glyph_cacheable(ctx, fontdesc->font, gid);
 
+	/* PDF spec: ISO 32000-2 latest version at the time of writing:
+	 * Section 9.3.6:
+	 * Where text is drawn using a Type 3 font:
+	 *  + if text rendering mode is set to a value of 3 or 7, the text shall not be rendered.
+	 *  + if text rendering mode is set to a value other than 3 or 7, the text shall be rendered using the glyph descriptions in the Type 3 font.
+	 *  + If text rendering mode is set to a value of 4, 5, 6 or 7, nothing shall be added to the clipping path.
+	 */
+	type3_hitr = (fontdesc->font->t3procs && gstate->text.render >= 4);
+
 	/* flush buffered text if rendermode has changed */
-	if (!pr->tos.text || gstate->text.render != pr->tos.text_mode || render_direct)
+	if (!pr->tos.text || gstate->text.render != pr->tos.text_mode || render_direct || type3_hitr)
 	{
 		gstate = pdf_flush_text(ctx, pr);
 		pdf_tos_reset(ctx, &pr->tos, gstate->text.render);
 	}
 
-	if (render_direct)
+	/* If Type3 and tr >= 4, then ignore the clipping path part. */
+	if (type3_hitr)
+		pr->tos.text_mode -= 4;
+
+	/* If Type3 and tr != 3, then use textrender mode 0. */
+	if (fontdesc->font->t3procs && pr->tos.text_mode != 3)
+		pr->tos.text_mode = 0;
+
+	if (render_direct && pr->tos.text_mode != 3 /* or 7, by type3_hitr */)
 	{
 		/* Render the glyph stream direct here (only happens for
 		 * type3 glyphs that seem to inherit current graphics
 		 * attributes, or type 3 glyphs within type3 glyphs). */
 		fz_matrix composed = fz_concat(trm, gstate->ctm);
+		/* Whatever problems the underlying char has is no concern of
+		 * ours. Store the flags, restore them afterwards. */
+		int old_flags = pr->dev->flags;
+		pdf_gstate *fill_gstate = NULL;
+		pdf_gstate *stroke_gstate = NULL;
 		pdf_gsave(ctx, pr);
 		gstate = pr->gstate + pr->gtop;
+		if (gstate->fill.kind == PDF_MAT_PATTERN && gstate->fill.gstate_num >= 0)
+			fill_gstate = pr->gstate + gstate->fill.gstate_num;
+		if (gstate->stroke.kind == PDF_MAT_PATTERN && gstate->stroke.gstate_num >= 0)
+			stroke_gstate = pr->gstate + gstate->stroke.gstate_num;
 		pdf_drop_font(ctx, gstate->text.font);
 		gstate->text.font = NULL; /* don't inherit the current font... */
-		fz_render_t3_glyph_direct(ctx, pr->dev, fontdesc->font, gid, composed, gstate, pr->default_cs);
+		fz_render_t3_glyph_direct(ctx, pr->dev, fontdesc->font, gid, composed, gstate, pr->default_cs, fill_gstate, stroke_gstate);
+		pr->dev->flags = old_flags;
 		pdf_grestore(ctx, pr);
 		/* Render text invisibly so that it can still be extracted. */
 		pr->tos.text_mode = 3;
@@ -1234,6 +1326,54 @@ pdf_show_space(fz_context *ctx, pdf_run_processor *pr, float tadj)
 		pr->tos.tm = fz_pre_translate(pr->tos.tm, 0, tadj);
 }
 
+static int
+int_in_singleton_or_array(fz_context *ctx, pdf_obj *k, int id)
+{
+	/* In the most common case the /K value will be id. */
+	if (pdf_is_int(ctx, k) && pdf_to_int(ctx, k) == id)
+		return 1;
+
+	/* In the next most common case, there will be an array of
+	 * items, one of which is k. */
+	if (pdf_is_array(ctx, k))
+	{
+		int i, n = pdf_array_len(ctx, k);
+
+		for (i = 0; i < n; i++)
+		{
+			pdf_obj *o = pdf_array_get(ctx, k, i);
+			if (pdf_is_int(ctx, o) && pdf_to_int(ctx, o) == id)
+				return 1;
+		}
+	}
+
+	return 0;
+}
+
+pdf_obj *
+pdf_lookup_mcid_in_mcids(fz_context *ctx, int id, pdf_obj *mcids)
+{
+	pdf_obj *mcid = pdf_array_get(ctx, mcids, id);
+	pdf_obj *k = pdf_dict_get(ctx, mcid, PDF_NAME(K));
+	int i, n;
+
+	if (int_in_singleton_or_array(ctx, k, id))
+		return mcid;
+
+	/* At this point, something has gone wrong. One common case that
+	 * appears to fail is where the MCIDs array has the right things
+	 * in, but at the wrong indexes. So do some searching. */
+	n = pdf_array_len(ctx, mcids);
+	for (i = 0; i < n; i++)
+	{
+		pdf_obj *o = pdf_array_get(ctx, mcids, i);
+		if (int_in_singleton_or_array(ctx, pdf_dict_get(ctx, o, PDF_NAME(K)), id))
+			return o;
+	}
+
+	return NULL;
+}
+
 static pdf_obj *
 lookup_mcid(fz_context *ctx, pdf_run_processor *proc, pdf_obj *val)
 {
@@ -1253,7 +1393,7 @@ lookup_mcid(fz_context *ctx, pdf_run_processor *proc, pdf_obj *val)
 
 	id = pdf_to_int(ctx, mcid);
 	mcids = pdf_lookup_number(ctx, pdf_dict_getl(ctx, pdf_trailer(ctx, proc->doc), PDF_NAME(Root), PDF_NAME(StructTreeRoot), PDF_NAME(ParentTree), NULL), proc->struct_parent);
-	return pdf_array_get(ctx, mcids, id);
+	return pdf_lookup_mcid_in_mcids(ctx, id, mcids);
 }
 
 static fz_text_language
@@ -1350,12 +1490,16 @@ pdf_show_text(fz_context *ctx, pdf_run_processor *pr, pdf_obj *text)
 			if (pdf_is_string(ctx, item))
 				show_string(ctx, pr, (unsigned char *)pdf_to_str_buf(ctx, item), pdf_to_str_len(ctx, item));
 			else
+			{
+				/* Bug 708615: pdf_show_char inside show_string can realloc gstate. */
+				gstate = pr->gstate + pr->gtop;
 				pdf_show_space(ctx, pr, - pdf_to_real(ctx, item) * gstate->text.size * 0.001f);
+			}
 		}
 	}
 	else if (pdf_is_string(ctx, text))
 	{
-		pdf_show_string(ctx, pr, (unsigned char *)pdf_to_str_buf(ctx, text), pdf_to_str_len(ctx, text));
+		show_string(ctx, pr, (unsigned char *)pdf_to_str_buf(ctx, text), pdf_to_str_len(ctx, text));
 	}
 }
 
@@ -1717,57 +1861,89 @@ pop_any_pending_mcid_changes(fz_context *ctx, pdf_run_processor *pr)
 	pr->pending_mcid_pop = NULL;
 }
 
-struct line
+static pdf_obj **
+get_lineage(fz_context *ctx, pdf_obj *a, int *lenp)
 {
-	pdf_obj *obj;
-	struct line *child;
-};
+	int max = 0;
+	int len = 0;
+	pdf_obj **line = NULL;
 
-static pdf_obj *
-find_most_recent_common_ancestor_imp(fz_context *ctx, pdf_obj *a, struct line *line_a, pdf_obj *b, struct line *line_b, pdf_cycle_list *cycle_up_a, pdf_cycle_list *cycle_up_b)
-{
-	struct line line;
-	pdf_obj *common = NULL;
-	pdf_cycle_list cycle;
-	pdf_obj *parent;
+	fz_var(line);
 
-	/* First ascend one lineage. */
-	if (pdf_is_dict(ctx, a))
+	fz_try(ctx)
 	{
-		if (pdf_cycle(ctx, &cycle, cycle_up_a, a))
-			fz_throw(ctx, FZ_ERROR_FORMAT, "cycle in structure tree");
-		line.obj = a;
-		line.child = line_a;
-		parent = pdf_dict_get(ctx, a, PDF_NAME(P));
-		return find_most_recent_common_ancestor_imp(ctx, parent, &line, b, NULL, &cycle, NULL);
+		while (1)
+		{
+			/* Put a into lineage. */
+			if (max == len)
+			{
+				max *= 2;
+				if (max == 0)
+					max = 32;
+				line = fz_realloc(ctx, line, sizeof(*line) * max);
+			}
+			line[len++] = a;
+
+			a = pdf_dict_get(ctx, a, PDF_NAME(P));
+			if (a == NULL)
+				break;
+
+			if (a == line[len>>1])
+				fz_throw(ctx, FZ_ERROR_FORMAT, "cycle in structure tree");
+		}
 	}
-	/* Then ascend the other lineage. */
-	else if (pdf_is_dict(ctx, b))
+	fz_catch(ctx)
 	{
-		if (pdf_cycle(ctx, &cycle, cycle_up_b, b))
-			fz_throw(ctx, FZ_ERROR_FORMAT, "cycle in structure tree");
-		line.obj = b;
-		line.child = line_b;
-		parent = pdf_dict_get(ctx, b, PDF_NAME(P));
-		return find_most_recent_common_ancestor_imp(ctx, a, line_a, parent, &line, cycle_up_a, &cycle);
+		fz_free(ctx, line);
+		fz_rethrow(ctx);
 	}
 
-	/* Once both lineages are know, traverse top-down to find most recent common ancestor. */
-	while (line_a && line_b && !pdf_objcmp(ctx, line_a->obj, line_b->obj))
-	{
-		common = line_a->obj;
-		line_a = line_a->child;
-		line_b = line_b->child;
-	}
-	return common;
+	*lenp = len;
+
+	return line;
 }
 
-static pdf_obj *
+pdf_obj *
 find_most_recent_common_ancestor(fz_context *ctx, pdf_obj *a, pdf_obj *b)
 {
+	/* First ascend one lineage. */
+	int a_len, b_len;
+	pdf_obj **line_a = NULL;
+	pdf_obj **line_b = NULL;
+	pdf_obj *common;
+
 	if (!pdf_is_dict(ctx, a) || !pdf_is_dict(ctx, b))
 		return NULL;
-	return find_most_recent_common_ancestor_imp(ctx, a, NULL, b, NULL, NULL, NULL);
+
+	fz_var(line_a);
+	fz_var(line_b);
+
+	fz_try(ctx)
+	{
+		line_a = get_lineage(ctx, a, &a_len);
+		line_b = get_lineage(ctx, b, &b_len);
+
+		assert(a_len > 0 && b_len > 0);
+
+		/* Once both lineages are know, traverse top-down to find most recent common ancestor. */
+		if (line_a[a_len-1] != line_b[b_len-1])
+			fz_throw(ctx, FZ_ERROR_FORMAT, "No common ancestor in structure tree");
+
+		while (a_len > 0 && b_len > 0 && line_a[a_len-1] == line_b[b_len-1])
+			a_len--, b_len--;
+
+		common = line_a[a_len];
+	}
+	fz_always(ctx)
+	{
+		fz_free(ctx, line_a);
+		fz_free(ctx, line_b);
+	}
+	fz_catch(ctx)
+	{
+		fz_rethrow(ctx);
+	}
+	return common;
 }
 
 static int
@@ -1778,7 +1954,7 @@ get_struct_index(fz_context *ctx, pdf_obj *send)
 	int i, n;
 
 	if (p == NULL)
-		return 0; /* Presumably the StructTreeToot */
+		return 0; /* Presumably the StructTreeRoot */
 
 	/* So, get the kids array. */
 	k = pdf_dict_get(ctx, p, PDF_NAME(K));
@@ -2138,7 +2314,7 @@ pdf_run_xobject(fz_context *ctx, pdf_run_processor *pr, pdf_obj *xobj, pdf_obj *
 
 			fz_rect bbox = fz_transform_rect(xobj_bbox, gstate->ctm);
 
-			gstate = begin_softmask(ctx, pr, &softmask);
+			gstate = begin_softmask(ctx, pr, &softmask, bbox);
 
 			if (isolated)
 				cs = pdf_xobject_colorspace(ctx, xobj);
@@ -2301,9 +2477,9 @@ static void pdf_run_d(fz_context *ctx, pdf_processor *proc, pdf_obj *array, floa
 	pdf_gstate *gstate = pdf_flush_text(ctx, pr);
 	int len, i;
 
+	pr->dev->flags &= ~FZ_DEVFLAG_DASH_PATTERN_UNDEFINED;
 	len = pdf_array_len(ctx, array);
 	gstate->stroke_state = fz_unshare_stroke_state_with_dash_len(ctx, gstate->stroke_state, len);
-	gstate->stroke_state->dash_len = len;
 	for (i = 0; i < len; i++)
 		gstate->stroke_state->dash_list[i] = pdf_array_get_real(ctx, array, i);
 	gstate->stroke_state->dash_phase = phase;
@@ -2408,7 +2584,7 @@ static void pdf_run_gs_SMask(fz_context *ctx, pdf_processor *proc, pdf_obj *smas
 		gstate->softmask_ctm = gstate->ctm;
 		gstate->softmask = pdf_keep_obj(ctx, smask);
 		gstate->softmask_cs = fz_keep_colorspace(ctx, smask_cs);
-		gstate->softmask_resources = pdf_keep_obj(ctx, pr->rstack->resources);
+		gstate->softmask_resources = pdf_keep_obj(ctx, proc->rstack->resources);
 		pdf_drop_obj(ctx, gstate->softmask_tr);
 		gstate->softmask_tr = NULL;
 		if (tr)
@@ -2717,14 +2893,7 @@ static void pdf_run_d1(fz_context *ctx, pdf_processor *proc, float wx, float wy,
 {
 	pdf_run_processor *pr = (pdf_run_processor *)proc;
 	pr->dev->flags |= FZ_DEVFLAG_MASK | FZ_DEVFLAG_BBOX_DEFINED;
-	pr->dev->flags &= ~(FZ_DEVFLAG_FILLCOLOR_UNDEFINED |
-				FZ_DEVFLAG_STROKECOLOR_UNDEFINED |
-				FZ_DEVFLAG_STARTCAP_UNDEFINED |
-				FZ_DEVFLAG_DASHCAP_UNDEFINED |
-				FZ_DEVFLAG_ENDCAP_UNDEFINED |
-				FZ_DEVFLAG_LINEJOIN_UNDEFINED |
-				FZ_DEVFLAG_MITERLIMIT_UNDEFINED |
-				FZ_DEVFLAG_LINEWIDTH_UNDEFINED);
+	pr->dev->flags &= ~(FZ_DEVFLAG_FILLCOLOR_UNDEFINED | FZ_DEVFLAG_STROKECOLOR_UNDEFINED);
 	pr->dev->d1_rect.x0 = fz_min(llx, urx);
 	pr->dev->d1_rect.y0 = fz_min(lly, ury);
 	pr->dev->d1_rect.x1 = fz_max(llx, urx);
@@ -2737,6 +2906,8 @@ static void pdf_run_CS(fz_context *ctx, pdf_processor *proc, const char *name, f
 {
 	pdf_run_processor *pr = (pdf_run_processor *)proc;
 	pr->dev->flags &= ~FZ_DEVFLAG_STROKECOLOR_UNDEFINED;
+	if (pr->dev->flags & FZ_DEVFLAG_MASK)
+		return;
 	if (!strcmp(name, "Pattern"))
 		pdf_set_pattern(ctx, pr, PDF_STROKE, NULL, NULL);
 	else
@@ -2747,6 +2918,8 @@ static void pdf_run_cs(fz_context *ctx, pdf_processor *proc, const char *name, f
 {
 	pdf_run_processor *pr = (pdf_run_processor *)proc;
 	pr->dev->flags &= ~FZ_DEVFLAG_FILLCOLOR_UNDEFINED;
+	if (pr->dev->flags & FZ_DEVFLAG_MASK)
+		return;
 	if (!strcmp(name, "Pattern"))
 		pdf_set_pattern(ctx, pr, PDF_FILL, NULL, NULL);
 	else
@@ -2757,6 +2930,8 @@ static void pdf_run_SC_color(fz_context *ctx, pdf_processor *proc, int n, float
 {
 	pdf_run_processor *pr = (pdf_run_processor *)proc;
 	pr->dev->flags &= ~FZ_DEVFLAG_STROKECOLOR_UNDEFINED;
+	if (pr->dev->flags & FZ_DEVFLAG_MASK)
+		return;
 	pdf_set_color(ctx, pr, PDF_STROKE, color);
 }
 
@@ -2764,6 +2939,8 @@ static void pdf_run_sc_color(fz_context *ctx, pdf_processor *proc, int n, float
 {
 	pdf_run_processor *pr = (pdf_run_processor *)proc;
 	pr->dev->flags &= ~FZ_DEVFLAG_FILLCOLOR_UNDEFINED;
+	if (pr->dev->flags & FZ_DEVFLAG_MASK)
+		return;
 	pdf_set_color(ctx, pr, PDF_FILL, color);
 }
 
@@ -2771,6 +2948,8 @@ static void pdf_run_SC_pattern(fz_context *ctx, pdf_processor *proc, const char
 {
 	pdf_run_processor *pr = (pdf_run_processor *)proc;
 	pr->dev->flags &= ~FZ_DEVFLAG_STROKECOLOR_UNDEFINED;
+	if (pr->dev->flags & FZ_DEVFLAG_MASK)
+		return;
 	pdf_set_pattern(ctx, pr, PDF_STROKE, pat, color);
 }
 
@@ -2778,6 +2957,8 @@ static void pdf_run_sc_pattern(fz_context *ctx, pdf_processor *proc, const char
 {
 	pdf_run_processor *pr = (pdf_run_processor *)proc;
 	pr->dev->flags &= ~FZ_DEVFLAG_FILLCOLOR_UNDEFINED;
+	if (pr->dev->flags & FZ_DEVFLAG_MASK)
+		return;
 	pdf_set_pattern(ctx, pr, PDF_FILL, pat, color);
 }
 
@@ -2785,6 +2966,8 @@ static void pdf_run_SC_shade(fz_context *ctx, pdf_processor *proc, const char *n
 {
 	pdf_run_processor *pr = (pdf_run_processor *)proc;
 	pr->dev->flags &= ~FZ_DEVFLAG_STROKECOLOR_UNDEFINED;
+	if (pr->dev->flags & FZ_DEVFLAG_MASK)
+		return;
 	pdf_set_shade(ctx, pr, PDF_STROKE, shade);
 }
 
@@ -2792,6 +2975,8 @@ static void pdf_run_sc_shade(fz_context *ctx, pdf_processor *proc, const char *n
 {
 	pdf_run_processor *pr = (pdf_run_processor *)proc;
 	pr->dev->flags &= ~FZ_DEVFLAG_FILLCOLOR_UNDEFINED;
+	if (pr->dev->flags & FZ_DEVFLAG_MASK)
+		return;
 	pdf_set_shade(ctx, pr, PDF_FILL, shade);
 }
 
@@ -2799,6 +2984,8 @@ static void pdf_run_G(fz_context *ctx, pdf_processor *proc, float g)
 {
 	pdf_run_processor *pr = (pdf_run_processor *)proc;
 	pr->dev->flags &= ~FZ_DEVFLAG_STROKECOLOR_UNDEFINED;
+	if (pr->dev->flags & FZ_DEVFLAG_MASK)
+		return;
 	pdf_set_colorspace(ctx, pr, PDF_STROKE, fz_device_gray(ctx));
 	pdf_set_color(ctx, pr, PDF_STROKE, &g);
 }
@@ -2807,6 +2994,8 @@ static void pdf_run_g(fz_context *ctx, pdf_processor *proc, float g)
 {
 	pdf_run_processor *pr = (pdf_run_processor *)proc;
 	pr->dev->flags &= ~FZ_DEVFLAG_FILLCOLOR_UNDEFINED;
+	if (pr->dev->flags & FZ_DEVFLAG_MASK)
+		return;
 	pdf_set_colorspace(ctx, pr, PDF_FILL, fz_device_gray(ctx));
 	pdf_set_color(ctx, pr, PDF_FILL, &g);
 }
@@ -2816,6 +3005,8 @@ static void pdf_run_K(fz_context *ctx, pdf_processor *proc, float c, float m, fl
 	pdf_run_processor *pr = (pdf_run_processor *)proc;
 	float color[4] = {c, m, y, k};
 	pr->dev->flags &= ~FZ_DEVFLAG_STROKECOLOR_UNDEFINED;
+	if (pr->dev->flags & FZ_DEVFLAG_MASK)
+		return;
 	pdf_set_colorspace(ctx, pr, PDF_STROKE, fz_device_cmyk(ctx));
 	pdf_set_color(ctx, pr, PDF_STROKE, color);
 }
@@ -2825,6 +3016,8 @@ static void pdf_run_k(fz_context *ctx, pdf_processor *proc, float c, float m, fl
 	pdf_run_processor *pr = (pdf_run_processor *)proc;
 	float color[4] = {c, m, y, k};
 	pr->dev->flags &= ~FZ_DEVFLAG_FILLCOLOR_UNDEFINED;
+	if (pr->dev->flags & FZ_DEVFLAG_MASK)
+		return;
 	pdf_set_colorspace(ctx, pr, PDF_FILL, fz_device_cmyk(ctx));
 	pdf_set_color(ctx, pr, PDF_FILL, color);
 }
@@ -2834,6 +3027,8 @@ static void pdf_run_RG(fz_context *ctx, pdf_processor *proc, float r, float g, f
 	pdf_run_processor *pr = (pdf_run_processor *)proc;
 	float color[3] = {r, g, b};
 	pr->dev->flags &= ~FZ_DEVFLAG_STROKECOLOR_UNDEFINED;
+	if (pr->dev->flags & FZ_DEVFLAG_MASK)
+		return;
 	pdf_set_colorspace(ctx, pr, PDF_STROKE, fz_device_rgb(ctx));
 	pdf_set_color(ctx, pr, PDF_STROKE, color);
 }
@@ -2843,6 +3038,8 @@ static void pdf_run_rg(fz_context *ctx, pdf_processor *proc, float r, float g, f
 	pdf_run_processor *pr = (pdf_run_processor *)proc;
 	float color[3] = {r, g, b};
 	pr->dev->flags &= ~FZ_DEVFLAG_FILLCOLOR_UNDEFINED;
+	if (pr->dev->flags & FZ_DEVFLAG_MASK)
+		return;
 	pdf_set_colorspace(ctx, pr, PDF_FILL, fz_device_rgb(ctx));
 	pdf_set_color(ctx, pr, PDF_FILL, color);
 }
@@ -2872,8 +3069,7 @@ static void pdf_run_Do_image(fz_context *ctx, pdf_processor *proc, const char *n
 
 static void pdf_run_Do_form(fz_context *ctx, pdf_processor *proc, const char *name, pdf_obj *xobj)
 {
-	pdf_run_processor *pr = (pdf_run_processor *)proc;
-	pdf_run_xobject(ctx, (pdf_run_processor*)proc, xobj, pr->rstack->resources, fz_identity, 0);
+	pdf_run_xobject(ctx, (pdf_run_processor*)proc, xobj, proc->rstack->resources, fz_identity, 0);
 }
 
 /* marked content */
@@ -2935,11 +3131,20 @@ pdf_close_run_processor(fz_context *ctx, pdf_processor *proc)
 
 	while (pr->nest_depth > 0)
 	{
-		if (pr->nest_mark[pr->nest_depth-1] == CLIP_MARK)
+		if (pr->nest_mark[pr->nest_depth-1] < 0)
+		{
+			/* It's a clip. */
 			fz_pop_clip(ctx, pr->dev);
+			pr->nest_mark[pr->nest_depth-1]++;
+			if (pr->nest_mark[pr->nest_depth-1] == 0)
+				pr->nest_depth--;
+		}
 		else
+		{
+			/* It's a layer. */
 			fz_end_layer(ctx, pr->dev);
-		pr->nest_depth--;
+			pr->nest_depth--;
+		}
 	}
 
 	pop_structure_to(ctx, pr, NULL);
@@ -2965,14 +3170,6 @@ pdf_drop_run_processor(fz_context *ctx, pdf_processor *proc)
 
 	fz_free(ctx, pr->gstate);
 
-	while (pr->rstack)
-	{
-		resources_stack *stk = pr->rstack;
-		pr->rstack = stk->next;
-		pdf_drop_obj(ctx, stk->resources);
-		fz_free(ctx, stk);
-	}
-
 	while (pr->begin_layer)
 	{
 		begin_layer_stack *stk = pr->begin_layer;
@@ -2990,33 +3187,6 @@ pdf_drop_run_processor(fz_context *ctx, pdf_processor *proc)
 	pdf_drop_obj(ctx, pr->role_map);
 }
 
-static void
-pdf_run_push_resources(fz_context *ctx, pdf_processor *proc, pdf_obj *resources)
-{
-	pdf_run_processor *pr = (pdf_run_processor *)proc;
-	resources_stack *stk = fz_malloc_struct(ctx, resources_stack);
-
-	stk->next = pr->rstack;
-	pr->rstack = stk;
-	stk->resources = pdf_keep_obj(ctx, resources);
-}
-
-static pdf_obj *
-pdf_run_pop_resources(fz_context *ctx, pdf_processor *proc)
-{
-	pdf_run_processor *pr = (pdf_run_processor *)proc;
-	resources_stack *stk = pr->rstack;
-
-	if (stk)
-	{
-		pr->rstack = stk->next;
-		pdf_drop_obj(ctx, stk->resources);
-		fz_free(ctx, stk);
-	}
-
-	return NULL;
-}
-
 /*
 	Create a new "run" processor. This maps
 	from PDF operators to fz_device level calls.
@@ -3033,7 +3203,7 @@ pdf_run_pop_resources(fz_context *ctx, pdf_processor *proc)
 	gstate: The initial graphics state.
 */
 pdf_processor *
-pdf_new_run_processor(fz_context *ctx, pdf_document *doc, fz_device *dev, fz_matrix ctm, int struct_parent, const char *usage, pdf_gstate *gstate, fz_default_colorspaces *default_cs, fz_cookie *cookie)
+pdf_new_run_processor(fz_context *ctx, pdf_document *doc, fz_device *dev, fz_matrix ctm, int struct_parent, const char *usage, pdf_gstate *gstate, fz_default_colorspaces *default_cs, fz_cookie *cookie, pdf_gstate *fill_gstate, pdf_gstate *stroke_gstate)
 {
 	pdf_run_processor *proc = pdf_new_processor(ctx, sizeof *proc);
 	{
@@ -3042,9 +3212,6 @@ pdf_new_run_processor(fz_context *ctx, pdf_document *doc, fz_device *dev, fz_mat
 		proc->super.close_processor = pdf_close_run_processor;
 		proc->super.drop_processor = pdf_drop_run_processor;
 
-		proc->super.push_resources = pdf_run_push_resources;
-		proc->super.pop_resources = pdf_run_pop_resources;
-
 		/* general graphics state */
 		proc->super.op_w = pdf_run_w;
 		proc->super.op_j = pdf_run_j;
@@ -3202,14 +3369,31 @@ pdf_new_run_processor(fz_context *ctx, pdf_document *doc, fz_device *dev, fz_mat
 		proc->gtop = 0;
 		pdf_init_gstate(ctx, &proc->gstate[0], ctm);
 
+		if (fill_gstate)
+		{
+			pdf_copy_gstate(ctx, &proc->gstate[0], fill_gstate);
+			proc->gstate[0].clip_depth = 0;
+			proc->gtop++;
+		}
+		if (stroke_gstate)
+		{
+			pdf_copy_gstate(ctx, &proc->gstate[proc->gtop], stroke_gstate);
+			proc->gstate[proc->gtop].clip_depth = 0;
+			proc->gtop++;
+		}
 		if (gstate)
 		{
-			pdf_copy_gstate(ctx, &proc->gstate[0], gstate);
-			proc->gstate[0].clip_depth = 0;
-			proc->gstate[0].ctm = ctm;
+			pdf_copy_gstate(ctx, &proc->gstate[proc->gtop], gstate);
+			proc->gstate[proc->gtop].clip_depth = 0;
+			proc->gstate[proc->gtop].ctm = ctm;
 		}
+		proc->gparent = proc->gtop;
+		if (fill_gstate)
+			proc->gstate[proc->gtop].fill.gstate_num = 0;
+		if (stroke_gstate)
+			proc->gstate[proc->gtop].fill.gstate_num = (fill_gstate != NULL);
 
-		/* We need to save an extra level to allow for level 0 to be the parent gstate level. */
+		/* We need to save an extra level to allow for the parent gstate level. */
 		pdf_gsave(ctx, proc);
 
 		/* Structure details */
