@@ -1,4 +1,4 @@
-// Copyright (C) 2004-2024 Artifex Software, Inc.
+// Copyright (C) 2004-2025 Artifex Software, Inc.
 //
 // This file is part of MuPDF.
 //
@@ -50,13 +50,16 @@ pdf_test_outline(fz_context *ctx, pdf_document *doc, pdf_obj *dict, pdf_mark_bit
 		if (pdf_mark_bits_set(ctx, marks, dict))
 			fz_throw(ctx, FZ_ERROR_FORMAT, "Cycle detected in outlines");
 
+		if (!pdf_is_indirect(ctx, dict))
+			  fz_throw(ctx, FZ_ERROR_FORMAT, "Non-indirect outline entry discovered");
+
 		parent = pdf_dict_get(ctx, dict, PDF_NAME(Parent));
 		prev = pdf_dict_get(ctx, dict, PDF_NAME(Prev));
 		next = pdf_dict_get(ctx, dict, PDF_NAME(Next));
 
 		parent_diff = pdf_objcmp(ctx, parent, expected_parent);
 		prev_diff = pdf_objcmp(ctx, prev, expected_prev);
-		last_diff = next == NULL && pdf_objcmp_resolve(ctx, last, dict);
+		last_diff = next == NULL && pdf_objcmp(ctx, last, dict);
 
 		if (fixed == NULL)
 		{
@@ -77,12 +80,10 @@ pdf_test_outline(fz_context *ctx, pdf_document *doc, pdf_obj *dict, pdf_mark_bit
 			{
 				if (parent_diff)
 				{
-					fz_warn(ctx, "Bad or missing parent pointer in outline tree, repairing");
 					pdf_dict_put(ctx, dict, PDF_NAME(Parent), expected_parent);
 				}
 				if (prev_diff)
 				{
-					fz_warn(ctx, "Bad or missing prev pointer in outline tree, repairing");
 					if (expected_prev)
 						pdf_dict_put(ctx, dict, PDF_NAME(Prev), expected_prev);
 					else
@@ -90,7 +91,6 @@ pdf_test_outline(fz_context *ctx, pdf_document *doc, pdf_obj *dict, pdf_mark_bit
 				}
 				if (last_diff)
 				{
-					fz_warn(ctx, "Bad or missing last pointer in outline tree, repairing");
 					pdf_dict_put(ctx, expected_parent, PDF_NAME(Last), dict);
 				}
 			}
@@ -256,6 +256,8 @@ do_outline_update(fz_context *ctx, pdf_obj *obj, fz_outline_item *item, int is_n
 		pdf_dict_del(ctx, obj, PDF_NAME(Title));
 
 	pdf_dict_del(ctx, obj, PDF_NAME(A));
+	pdf_dict_del(ctx, obj, PDF_NAME(C));
+	pdf_dict_del(ctx, obj, PDF_NAME(F));
 	pdf_dict_del(ctx, obj, PDF_NAME(Dest));
 	if (item->uri)
 	{
@@ -271,6 +273,15 @@ do_outline_update(fz_context *ctx, pdf_obj *obj, fz_outline_item *item, int is_n
 			pdf_dict_put_drop(ctx, obj, PDF_NAME(A),
 				pdf_new_action_from_link(ctx, doc, item->uri));
 	}
+	if (item->r != 0 || item->g != 0 || item->b != 0)
+	{
+		pdf_obj *color = pdf_dict_put_array(ctx, obj, PDF_NAME(C), 3);
+		pdf_array_put_real(ctx, color, 0, item->r / 255.0);
+		pdf_array_put_real(ctx, color, 1, item->g / 255.0);
+		pdf_array_put_real(ctx, color, 2, item->b / 255.0);
+	}
+	if (item->flags != 0)
+		pdf_dict_put_int(ctx, obj, PDF_NAME(F), item->flags);
 }
 
 static int
@@ -500,6 +511,13 @@ pdf_outline_iterator_item(fz_context *ctx, fz_outline_iterator *iter_)
 
 	iter->item.is_open = pdf_dict_get_int(ctx, iter->current, PDF_NAME(Count)) > 0;
 
+	obj = pdf_dict_get(ctx, iter->current, PDF_NAME(C));
+	iter->item.r = (int)(0.5 + 255 * pdf_array_get_real(ctx, obj, 0));
+	iter->item.g = (int)(0.5 + 255 * pdf_array_get_real(ctx, obj, 1));
+	iter->item.b = (int)(0.5 + 255 * pdf_array_get_real(ctx, obj, 2));
+
+	iter->item.flags = pdf_dict_get_int(ctx, iter->current, PDF_NAME(F)) & 127;
+
 	return &iter->item;
 }
 
@@ -539,9 +557,9 @@ fz_outline_iterator *pdf_new_outline_iterator(fz_context *ctx, pdf_document *doc
 			{
 				/* Pass through the outlines once, fixing inconsistencies */
 				pdf_test_outline(ctx, doc, first, marks, obj, &fixed);
-
 				if (fixed)
 				{
+					fz_warn(ctx, "repaired broken tree structure in outline");
 					/* If a fix was performed, pass through again,
 					 * this time throwing if it's still not correct. */
 					pdf_mark_bits_reset(ctx, marks);
