@@ -1,4 +1,4 @@
-// Copyright (C) 2004-2024 Artifex Software, Inc.
+// Copyright (C) 2004-2025 Artifex Software, Inc.
 //
 // This file is part of MuPDF.
 //
@@ -29,6 +29,20 @@
 
 static void pdf_adjust_page_labels(fz_context *ctx, pdf_document *doc, int index, int adjust);
 
+void
+pdf_set_page_tree_cache(fz_context *ctx, pdf_document *doc, int enabled)
+{
+	if (enabled)
+	{
+		doc->use_page_tree_map = 1;
+	}
+	else
+	{
+		pdf_drop_page_tree_internal(ctx, doc);
+		doc->use_page_tree_map = 0;
+	}
+}
+
 int
 pdf_count_pages(fz_context *ctx, pdf_document *doc)
 {
@@ -43,6 +57,10 @@ pdf_count_pages(fz_context *ctx, pdf_document *doc)
 		pages = pdf_to_int(ctx, pdf_dict_getp(ctx, pdf_trailer(ctx, doc), "Root/Pages/Count"));
 	if (pages < 0)
 		fz_throw(ctx, FZ_ERROR_FORMAT, "Invalid number of pages");
+	/* set an upper bound to cope with malicious /Count entries; as there
+	 * can't possibly be more pages than objects in the file! */
+	if (pages >= pdf_xref_len(ctx, doc))
+		fz_throw(ctx, FZ_ERROR_FORMAT, "Invalid number of pages");
 	return pages;
 }
 
@@ -252,13 +270,13 @@ pdf_lookup_page_loc(fz_context *ctx, pdf_document *doc, int needle, pdf_obj **pa
 pdf_obj *
 pdf_lookup_page_obj(fz_context *ctx, pdf_document *doc, int needle)
 {
-	if (doc->fwd_page_map == NULL && !doc->page_tree_broken)
+	if (doc->fwd_page_map == NULL && doc->use_page_tree_map)
 	{
 		fz_try(ctx)
 			pdf_load_page_tree_internal(ctx, doc);
 		fz_catch(ctx)
 		{
-			doc->page_tree_broken = 1;
+			doc->use_page_tree_map = 0;
 			fz_rethrow_if(ctx, FZ_ERROR_SYSTEM);
 			fz_report_error(ctx);
 			fz_warn(ctx, "Page tree load failed. Falling back to slow lookup");
@@ -290,14 +308,14 @@ pdf_count_pages_before_kid(fz_context *ctx, pdf_document *doc, pdf_obj *parent,
 		{
 			pdf_obj *count = pdf_dict_get(ctx, kid, PDF_NAME(Count));
 			int n = pdf_to_int(ctx, count);
-			if (!pdf_is_int(ctx, count) || n < 0)
+			if (!pdf_is_int(ctx, count) || n < 0 || INT_MAX - total <= n)
 				fz_throw(ctx, FZ_ERROR_FORMAT, "illegal or missing count in pages tree");
 			total += n;
 		}
 		else
 			total++;
 	}
-	fz_throw(ctx, FZ_ERROR_FORMAT, "kid not found in parent's kids array");
+	return -1; // the page we're looking for is not in the page tree (it has been deleted)
 }
 
 static int
@@ -306,6 +324,7 @@ pdf_lookup_page_number_slow(fz_context *ctx, pdf_document *doc, pdf_obj *node)
 	pdf_mark_list mark_list;
 	int needle = pdf_to_num(ctx, node);
 	int total = 0;
+	int n;
 	pdf_obj *parent;
 
 	if (!pdf_name_eq(ctx, pdf_dict_get(ctx, node, PDF_NAME(Type)), PDF_NAME(Page)))
@@ -322,7 +341,18 @@ pdf_lookup_page_number_slow(fz_context *ctx, pdf_document *doc, pdf_obj *node)
 		{
 			if (pdf_mark_list_push(ctx, &mark_list, parent))
 				fz_throw(ctx, FZ_ERROR_FORMAT, "cycle in page tree (parents)");
-			total += pdf_count_pages_before_kid(ctx, doc, parent, needle);
+			n = pdf_count_pages_before_kid(ctx, doc, parent, needle);
+
+			// Page was not found in page tree!
+			if (n < 0)
+			{
+				total = -1;
+				break;
+			}
+			if (INT_MAX - total <= n)
+				fz_throw(ctx, FZ_ERROR_FORMAT, "illegal or missing count in pages tree");
+
+			total += n;
 			needle = pdf_to_num(ctx, parent);
 			parent = pdf_dict_get(ctx, parent, PDF_NAME(Parent));
 		}
@@ -357,13 +387,14 @@ pdf_lookup_page_number_fast(fz_context *ctx, pdf_document *doc, int needle)
 int
 pdf_lookup_page_number(fz_context *ctx, pdf_document *doc, pdf_obj *page)
 {
-	if (doc->rev_page_map == NULL && !doc->page_tree_broken)
+	if (doc->rev_page_map == NULL && doc->use_page_tree_map)
 	{
 		fz_try(ctx)
 			pdf_load_page_tree_internal(ctx, doc);
 		fz_catch(ctx)
 		{
-			doc->page_tree_broken = 1;
+			doc->use_page_tree_map = 0;
+			fz_report_error(ctx);
 			fz_warn(ctx, "Page tree load failed. Falling back to slow lookup.");
 		}
 	}
@@ -764,6 +795,7 @@ pdf_page_obj_transform_box(fz_context *ctx, pdf_obj *pageobj, fz_rect *outbox, f
 	if (!pdf_is_array(ctx, obj))
 		obj = pdf_dict_get_inheritable(ctx, pageobj, PDF_NAME(MediaBox));
 	cropbox = pdf_to_rect(ctx, obj);
+	cropbox = fz_intersect_rect(cropbox, mediabox);
 	if (fz_is_empty_rect(cropbox))
 		cropbox = fz_make_rect(0, 0, 612, 792);
 	cropbox.x0 = fz_min(cropbox.x0, cropbox.x1);
@@ -1269,6 +1301,7 @@ void pdf_nuke_page(fz_context *ctx, pdf_page *page)
 	pdf_nuke_annots(ctx, page);
 	pdf_drop_obj(ctx, page->obj);
 	page->obj = NULL;
+	page->super.in_doc = 0;
 }
 
 void pdf_sync_page(fz_context *ctx, pdf_page *page)
@@ -1478,11 +1511,10 @@ void
 pdf_delete_page_range(fz_context *ctx, pdf_document *doc, int start, int end)
 {
 	int count = pdf_count_pages(ctx, doc);
-
-	if (end < 0 || end > count)
-		end = count+1;
-	if (start < 0)
-		start = 0;
+	if (end < 0)
+		end = count;
+	start = fz_clampi(start, 0, count);
+	end = fz_clampi(end, 0, count);
 	while (start < end)
 	{
 		pdf_delete_page(ctx, doc, start);
@@ -1648,9 +1680,10 @@ pdf_lookup_page_label_imp(fz_context *ctx, pdf_obj *node, int index, struct page
 static struct page_label_range
 pdf_lookup_page_label(fz_context *ctx, pdf_document *doc, int index)
 {
-	struct page_label_range range = { 0, NULL };
 	pdf_obj *root = pdf_dict_get(ctx, pdf_trailer(ctx, doc), PDF_NAME(Root));
 	pdf_obj *labels = pdf_dict_get(ctx, root, PDF_NAME(PageLabels));
+	pdf_obj *nums = pdf_dict_get(ctx, labels, PDF_NAME(Nums));
+	struct page_label_range range = { 0, NULL, 0, nums };
 	pdf_lookup_page_label_imp(ctx, labels, index, &range);
 	return range;
 }
@@ -1943,13 +1976,13 @@ pdf_format_page_label(fz_context *ctx, int index, pdf_obj *dict, char *buf, size
 	if (style == PDF_NAME(D))
 		fz_snprintf(buf, size, "%d", index + start);
 	else if (style == PDF_NAME(R))
-		pdf_format_roman_page_label(buf, size, index + start, roman_uc, "M");
+		pdf_format_roman_page_label(buf, (int)size, index + start, roman_uc, "M");
 	else if (style == PDF_NAME(r))
-		pdf_format_roman_page_label(buf, size, index + start, roman_lc, "m");
+		pdf_format_roman_page_label(buf, (int)size, index + start, roman_lc, "m");
 	else if (style == PDF_NAME(A))
-		pdf_format_alpha_page_label(buf, size, index + start, 'A');
+		pdf_format_alpha_page_label(buf, (int)size, index + start, 'A');
 	else if (style == PDF_NAME(a))
-		pdf_format_alpha_page_label(buf, size, index + start, 'a');
+		pdf_format_alpha_page_label(buf, (int)size, index + start, 'a');
 }
 
 void
