@@ -67,6 +67,15 @@ pdf_to_quad(fz_context *ctx, pdf_obj *array, int offset)
 	return q;
 }
 
+fz_point
+pdf_to_point(fz_context *ctx, pdf_obj *array, int offset)
+{
+	fz_point p;
+	p.x = pdf_array_get_real(ctx, array, offset+0);
+	p.y = pdf_array_get_real(ctx, array, offset+1);
+	return p;
+}
+
 fz_matrix
 pdf_to_matrix(fz_context *ctx, pdf_obj *array)
 {
@@ -304,7 +313,10 @@ is_valid_utf8(const unsigned char *s, const unsigned char *end)
 {
 	for (; s < end; ++s)
 	{
-		int skip = *s < 0x80 ? 0 : *s < 0xC0 ? -1 : *s < 0xE0 ? 1 : *s < 0xF0 ? 2 : *s < 0xF5 ? 3 : -1;
+		int c = *s;
+		int skip = c < 0x80 ? 0 : c < 0xC0 ? -1 : c < 0xE0 ? 1 : c < 0xF0 ? 2 : c < 0xF5 ? 3 : -1;
+		if (c >= 0x18 && c <= 0x1f) // Reserved control characters used by PDFDocEncoding.
+			return 0;
 		if (skip == -1)
 			return 0;
 		while (skip-- > 0)
@@ -702,10 +714,7 @@ pdf_parse_dict(fz_context *ctx, pdf_document *doc, fz_stream *file, pdf_lexbuf *
 				if (tok == PDF_TOK_CLOSE_DICT || tok == PDF_TOK_NAME ||
 					(tok == PDF_TOK_KEYWORD && !strcmp(buf->scratch, "ID")))
 				{
-					val = pdf_new_int(ctx, a);
-					pdf_dict_put(ctx, dict, key, val);
-					pdf_drop_obj(ctx, val);
-					val = NULL;
+					pdf_dict_put_int(ctx, dict, key, a);
 					pdf_drop_obj(ctx, key);
 					key = NULL;
 					goto skip;
