@@ -1,4 +1,4 @@
-// Copyright (C) 2004-2021 Artifex Software, Inc.
+// Copyright (C) 2004-2025 Artifex Software, Inc.
 //
 // This file is part of MuPDF.
 //
@@ -36,7 +36,7 @@ typedef struct
 	int funcs;
 } recolor_details;
 
-#define FUNSEGS 64 /* size of sampled mesh for function-based shadings */
+#define FUNSEGS 256 /* size of sampled mesh for function-based shadings */
 #define FUNBPS 8 /* Bits per sample in output functions */
 
 static void
@@ -621,10 +621,8 @@ fz_recolor_shade_type6(fz_context *ctx, pdf_obj *shade, recolor_details *rd)
 
 			for (i = startpt; i < 12; i++)
 			{
-				unsigned int x_bits = fz_read_bits(ctx, stream, bpcoord);
-				unsigned int y_bits = fz_read_bits(ctx, stream, bpcoord);
-				fz_write_bits(ctx, out, x_bits, bpcoord);
-				fz_write_bits(ctx, out, y_bits, bpcoord);
+				(void) fz_read_bits(ctx, stream, bpcoord);
+				(void) fz_read_bits(ctx, stream, bpcoord);
 			}
 
 			for (i = startcolor; i < 4; i++)
@@ -865,8 +863,6 @@ pdf_new_colorspace(fz_context *ctx, fz_colorspace *cs)
 	default:
 		fz_throw(ctx, FZ_ERROR_ARGUMENT, "Unimplemented colorspace");
 	}
-
-	return NULL;
 }
 
 pdf_obj *
@@ -908,7 +904,7 @@ pdf_recolor_shade(fz_context *ctx, pdf_obj *shade, pdf_shade_recolorer *reshade,
 		background = pdf_dict_get(ctx, shade, PDF_NAME(Background));
 		if (background)
 		{
-			int i, n = pdf_array_len(ctx, background);
+			int n = pdf_array_len(ctx, background);
 			float bg[FZ_MAX_COLORS];
 			float nbg[FZ_MAX_COLORS];
 
@@ -950,7 +946,7 @@ pdf_recolor_shade(fz_context *ctx, pdf_obj *shade, pdf_shade_recolorer *reshade,
 		}
 		else if (pdf_is_array(ctx, function))
 		{
-			int in, i;
+			int in;
 
 			rd.funcs = pdf_array_len(ctx, function);
 
