@@ -1,4 +1,4 @@
-// Copyright (C) 2004-2024 Artifex Software, Inc.
+// Copyright (C) 2004-2025 Artifex Software, Inc.
 //
 // This file is part of MuPDF.
 //
@@ -36,7 +36,7 @@
 typedef struct svg_state
 {
 	fz_matrix transform;
-	fz_stroke_state stroke;
+	fz_stroke_state *stroke;
 	int use_depth;
 
 	float viewport_w, viewport_h;
@@ -63,6 +63,38 @@ typedef struct svg_state
 static void svg_parse_common(fz_context *ctx, svg_document *doc, fz_xml *node, svg_state *state);
 static void svg_run_element(fz_context *ctx, fz_device *dev, svg_document *doc, fz_xml *root, const svg_state *state);
 
+static int svg_push_use(fz_context *ctx, svg_document *doc, svg_cycle_list *here, fz_xml *symbol)
+{
+	svg_cycle_list *x = doc->cycle;
+	while (x)
+	{
+		if (x->symbol == symbol)
+			return 0;
+		x = x->up;
+	}
+	here->up = doc->cycle;
+	here->symbol = symbol;
+	doc->cycle = here;
+	return 1;
+}
+
+static void svg_pop_use(fz_context *ctx, svg_document *doc)
+{
+	if (doc->cycle)
+		doc->cycle = doc->cycle->up;
+}
+
+void svg_begin_state(fz_context *ctx, svg_state *child, const svg_state *parent)
+{
+	memcpy(child, parent, sizeof(svg_state));
+	child->stroke = fz_clone_stroke_state(ctx, parent->stroke);
+}
+
+void svg_end_state(fz_context *ctx, svg_state *child)
+{
+	fz_drop_stroke_state(ctx, child->stroke);
+}
+
 static void svg_fill(fz_context *ctx, fz_device *dev, svg_document *doc, fz_path *path, svg_state *state)
 {
 	float opacity = state->opacity * state->fill_opacity;
@@ -74,7 +106,7 @@ static void svg_stroke(fz_context *ctx, fz_device *dev, svg_document *doc, fz_pa
 {
 	float opacity = state->opacity * state->stroke_opacity;
 	if (path)
-		fz_stroke_path(ctx, dev, path, &state->stroke, state->transform, fz_device_rgb(ctx), state->stroke_color, opacity, fz_default_color_params);
+		fz_stroke_path(ctx, dev, path, state->stroke, state->transform, fz_device_rgb(ctx), state->stroke_color, opacity, fz_default_color_params);
 }
 
 static void svg_draw_path(fz_context *ctx, fz_device *dev, svg_document *doc, fz_path *path, svg_state *state)
@@ -107,7 +139,7 @@ static void approx_circle(fz_context *ctx, fz_path *path, float cx, float cy, fl
 static void
 svg_run_rect(fz_context *ctx, fz_device *dev, svg_document *doc, fz_xml *node, const svg_state *inherit_state)
 {
-	svg_state local_state = *inherit_state;
+	svg_state local_state;
 
 	char *x_att = fz_xml_att(node, "x");
 	char *y_att = fz_xml_att(node, "y");
@@ -123,32 +155,35 @@ svg_run_rect(fz_context *ctx, fz_device *dev, svg_document *doc, fz_xml *node, c
 	float rx = 0;
 	float ry = 0;
 
-	fz_path *path;
+	fz_path *path = NULL;
 
-	svg_parse_common(ctx, doc, node, &local_state);
+	fz_var(path);
 
-	if (x_att) x = svg_parse_length(x_att, local_state.viewbox_w, local_state.fontsize);
-	if (y_att) y = svg_parse_length(y_att, local_state.viewbox_h, local_state.fontsize);
-	if (w_att) w = svg_parse_length(w_att, local_state.viewbox_w, local_state.fontsize);
-	if (h_att) h = svg_parse_length(h_att, local_state.viewbox_h, local_state.fontsize);
-	if (rx_att) rx = svg_parse_length(rx_att, local_state.viewbox_w, local_state.fontsize);
-	if (ry_att) ry = svg_parse_length(ry_att, local_state.viewbox_h, local_state.fontsize);
-
-	if (rx_att && !ry_att)
-		ry = rx;
-	if (ry_att && !rx_att)
-		rx = ry;
-	if (rx > w * 0.5f)
-		rx = w * 0.5f;
-	if (ry > h * 0.5f)
-		ry = h * 0.5f;
-
-	if (w <= 0 || h <= 0)
-		return;
-
-	path = fz_new_path(ctx);
+	svg_begin_state(ctx, &local_state, inherit_state);
 	fz_try(ctx)
 	{
+		svg_parse_common(ctx, doc, node, &local_state);
+
+		if (x_att) x = svg_parse_length(x_att, local_state.viewbox_w, local_state.fontsize);
+		if (y_att) y = svg_parse_length(y_att, local_state.viewbox_h, local_state.fontsize);
+		if (w_att) w = svg_parse_length(w_att, local_state.viewbox_w, local_state.fontsize);
+		if (h_att) h = svg_parse_length(h_att, local_state.viewbox_h, local_state.fontsize);
+		if (rx_att) rx = svg_parse_length(rx_att, local_state.viewbox_w, local_state.fontsize);
+		if (ry_att) ry = svg_parse_length(ry_att, local_state.viewbox_h, local_state.fontsize);
+
+		if (rx_att && !ry_att)
+			ry = rx;
+		if (ry_att && !rx_att)
+			rx = ry;
+		if (rx > w * 0.5f)
+			rx = w * 0.5f;
+		if (ry > h * 0.5f)
+			ry = h * 0.5f;
+
+		if (w <= 0 || h <= 0)
+			break;
+
+		path = fz_new_path(ctx);
 		if (rx == 0 || ry == 0)
 		{
 			fz_moveto(ctx, path, x, y);
@@ -174,7 +209,10 @@ svg_run_rect(fz_context *ctx, fz_device *dev, svg_document *doc, fz_xml *node, c
 		svg_draw_path(ctx, dev, doc, path, &local_state);
 	}
 	fz_always(ctx)
+	{
 		fz_drop_path(ctx, path);
+		svg_end_state(ctx, &local_state);
+	}
 	fz_catch(ctx)
 		fz_rethrow(ctx);
 
@@ -183,7 +221,7 @@ svg_run_rect(fz_context *ctx, fz_device *dev, svg_document *doc, fz_xml *node, c
 static void
 svg_run_circle(fz_context *ctx, fz_device *dev, svg_document *doc, fz_xml *node, const svg_state *inherit_state)
 {
-	svg_state local_state = *inherit_state;
+	svg_state local_state;
 
 	char *cx_att = fz_xml_att(node, "cx");
 	char *cy_att = fz_xml_att(node, "cy");
@@ -192,33 +230,40 @@ svg_run_circle(fz_context *ctx, fz_device *dev, svg_document *doc, fz_xml *node,
 	float cx = 0;
 	float cy = 0;
 	float r = 0;
-	fz_path *path;
+	fz_path *path = NULL;
 
-	svg_parse_common(ctx, doc, node, &local_state);
+	fz_var(path);
 
-	if (cx_att) cx = svg_parse_length(cx_att, local_state.viewbox_w, local_state.fontsize);
-	if (cy_att) cy = svg_parse_length(cy_att, local_state.viewbox_h, local_state.fontsize);
-	if (r_att) r = svg_parse_length(r_att, local_state.viewbox_size, 12);
-
-	if (r <= 0)
-		return;
-
-	path = fz_new_path(ctx);
+	svg_begin_state(ctx, &local_state, inherit_state);
 	fz_try(ctx)
 	{
-		approx_circle(ctx, path, cx, cy, r, r);
-		svg_draw_path(ctx, dev, doc, path, &local_state);
+		svg_parse_common(ctx, doc, node, &local_state);
+
+		if (cx_att) cx = svg_parse_length(cx_att, local_state.viewbox_w, local_state.fontsize);
+		if (cy_att) cy = svg_parse_length(cy_att, local_state.viewbox_h, local_state.fontsize);
+		if (r_att) r = svg_parse_length(r_att, local_state.viewbox_size, 12);
+
+		if (r > 0)
+		{
+			path = fz_new_path(ctx);
+			approx_circle(ctx, path, cx, cy, r, r);
+			svg_draw_path(ctx, dev, doc, path, &local_state);
+		}
 	}
 	fz_always(ctx)
+	{
 		fz_drop_path(ctx, path);
+		svg_end_state(ctx, &local_state);
+	}
 	fz_catch(ctx)
 		fz_rethrow(ctx);
+
 }
 
 static void
 svg_run_ellipse(fz_context *ctx, fz_device *dev, svg_document *doc, fz_xml *node, const svg_state *inherit_state)
 {
-	svg_state local_state = *inherit_state;
+	svg_state local_state;
 
 	char *cx_att = fz_xml_att(node, "cx");
 	char *cy_att = fz_xml_att(node, "cy");
@@ -230,26 +275,32 @@ svg_run_ellipse(fz_context *ctx, fz_device *dev, svg_document *doc, fz_xml *node
 	float rx = 0;
 	float ry = 0;
 
-	fz_path *path;
+	fz_path *path = NULL;
 
-	svg_parse_common(ctx, doc, node, &local_state);
+	fz_var(path);
 
-	if (cx_att) cx = svg_parse_length(cx_att, local_state.viewbox_w, local_state.fontsize);
-	if (cy_att) cy = svg_parse_length(cy_att, local_state.viewbox_h, local_state.fontsize);
-	if (rx_att) rx = svg_parse_length(rx_att, local_state.viewbox_w, local_state.fontsize);
-	if (ry_att) ry = svg_parse_length(ry_att, local_state.viewbox_h, local_state.fontsize);
-
-	if (rx <= 0 || ry <= 0)
-		return;
-
-	path = fz_new_path(ctx);
+	svg_begin_state(ctx, &local_state, inherit_state);
 	fz_try(ctx)
 	{
-		approx_circle(ctx, path, cx, cy, rx, ry);
-		svg_draw_path(ctx, dev, doc, path, &local_state);
+		svg_parse_common(ctx, doc, node, &local_state);
+
+		if (cx_att) cx = svg_parse_length(cx_att, local_state.viewbox_w, local_state.fontsize);
+		if (cy_att) cy = svg_parse_length(cy_att, local_state.viewbox_h, local_state.fontsize);
+		if (rx_att) rx = svg_parse_length(rx_att, local_state.viewbox_w, local_state.fontsize);
+		if (ry_att) ry = svg_parse_length(ry_att, local_state.viewbox_h, local_state.fontsize);
+
+		if (rx > 0 && ry > 0)
+		{
+			path = fz_new_path(ctx);
+			approx_circle(ctx, path, cx, cy, rx, ry);
+			svg_draw_path(ctx, dev, doc, path, &local_state);
+		}
 	}
 	fz_always(ctx)
+	{
 		fz_drop_path(ctx, path);
+		svg_end_state(ctx, &local_state);
+	}
 	fz_catch(ctx)
 		fz_rethrow(ctx);
 }
@@ -257,7 +308,8 @@ svg_run_ellipse(fz_context *ctx, fz_device *dev, svg_document *doc, fz_xml *node
 static void
 svg_run_line(fz_context *ctx, fz_device *dev, svg_document *doc, fz_xml *node, const svg_state *inherit_state)
 {
-	svg_state local_state = *inherit_state;
+	svg_state local_state;
+	fz_path *path = NULL;
 
 	char *x1_att = fz_xml_att(node, "x1");
 	char *y1_att = fz_xml_att(node, "y1");
@@ -269,27 +321,33 @@ svg_run_line(fz_context *ctx, fz_device *dev, svg_document *doc, fz_xml *node, c
 	float x2 = 0;
 	float y2 = 0;
 
-	svg_parse_common(ctx, doc, node, &local_state);
+	fz_var(path);
 
-	if (x1_att) x1 = svg_parse_length(x1_att, local_state.viewbox_w, local_state.fontsize);
-	if (y1_att) y1 = svg_parse_length(y1_att, local_state.viewbox_h, local_state.fontsize);
-	if (x2_att) x2 = svg_parse_length(x2_att, local_state.viewbox_w, local_state.fontsize);
-	if (y2_att) y2 = svg_parse_length(y2_att, local_state.viewbox_h, local_state.fontsize);
-
-	if (local_state.stroke_is_set)
+	svg_begin_state(ctx, &local_state, inherit_state);
+	fz_try(ctx)
 	{
-		fz_path *path = fz_new_path(ctx);
-		fz_try(ctx)
+		svg_parse_common(ctx, doc, node, &local_state);
+
+		if (x1_att) x1 = svg_parse_length(x1_att, local_state.viewbox_w, local_state.fontsize);
+		if (y1_att) y1 = svg_parse_length(y1_att, local_state.viewbox_h, local_state.fontsize);
+		if (x2_att) x2 = svg_parse_length(x2_att, local_state.viewbox_w, local_state.fontsize);
+		if (y2_att) y2 = svg_parse_length(y2_att, local_state.viewbox_h, local_state.fontsize);
+
+		if (local_state.stroke_is_set)
 		{
+			path = fz_new_path(ctx);
 			fz_moveto(ctx, path, x1, y1);
 			fz_lineto(ctx, path, x2, y2);
 			svg_stroke(ctx, dev, doc, path, &local_state);
 		}
-		fz_always(ctx)
-			fz_drop_path(ctx, path);
-		fz_catch(ctx)
-			fz_rethrow(ctx);
 	}
+	fz_always(ctx)
+	{
+		fz_drop_path(ctx, path);
+		svg_end_state(ctx, &local_state);
+	}
+	fz_catch(ctx)
+		fz_rethrow(ctx);
 }
 
 static fz_path *
@@ -322,6 +380,12 @@ svg_parse_polygon_imp(fz_context *ctx, svg_document *doc, fz_xml *node, int docl
 				str = svg_lex_number(&number, str);
 				args[nargs++] = number;
 			}
+			else if (*str)
+			{
+				/* Don't know what this is. Just skip it. */
+				fz_warn(ctx, "syntax error in polygon points");
+				str++;
+			}
 
 			if (nargs == 2)
 			{
@@ -350,35 +414,52 @@ svg_parse_polygon_imp(fz_context *ctx, svg_document *doc, fz_xml *node, int docl
 static void
 svg_run_polyline(fz_context *ctx, fz_device *dev, svg_document *doc, fz_xml *node, const svg_state *inherit_state)
 {
-	svg_state local_state = *inherit_state;
+	svg_state local_state;
+	fz_path *path = NULL;
 
-	svg_parse_common(ctx, doc, node, &local_state);
+	fz_var(path);
 
-	if (local_state.stroke_is_set)
+	svg_begin_state(ctx, &local_state, inherit_state);
+	fz_try(ctx)
 	{
-		fz_path *path = svg_parse_polygon_imp(ctx, doc, node, 0);
-		fz_try(ctx)
+		svg_parse_common(ctx, doc, node, &local_state);
+
+		if (local_state.stroke_is_set)
+		{
+			path = svg_parse_polygon_imp(ctx, doc, node, 0);
 			svg_stroke(ctx, dev, doc, path, &local_state);
-		fz_always(ctx)
-			fz_drop_path(ctx, path);
-		fz_catch(ctx)
-			fz_rethrow(ctx);
+		}
 	}
+	fz_always(ctx)
+	{
+		fz_drop_path(ctx, path);
+		svg_end_state(ctx, &local_state);
+	}
+	fz_catch(ctx)
+		fz_rethrow(ctx);
 }
 
 static void
 svg_run_polygon(fz_context *ctx, fz_device *dev, svg_document *doc, fz_xml *node, const svg_state *inherit_state)
 {
-	svg_state local_state = *inherit_state;
-	fz_path *path;
+	svg_state local_state;
+	fz_path *path = NULL;
 
-	svg_parse_common(ctx, doc, node, &local_state);
+	fz_var(path);
 
-	path = svg_parse_polygon_imp(ctx, doc, node, 1);
+	svg_begin_state(ctx, &local_state, inherit_state);
 	fz_try(ctx)
+	{
+		svg_parse_common(ctx, doc, node, &local_state);
+
+		path = svg_parse_polygon_imp(ctx, doc, node, 1);
 		svg_draw_path(ctx, dev, doc, path, &local_state);
+	}
 	fz_always(ctx)
+	{
 		fz_drop_path(ctx, path);
+		svg_end_state(ctx, &local_state);
+	}
 	fz_catch(ctx)
 		fz_rethrow(ctx);
 }
@@ -847,26 +928,33 @@ svg_parse_path_data(fz_context *ctx, fz_path *path, const char *str)
 static void
 svg_run_path(fz_context *ctx, fz_device *dev, svg_document *doc, fz_xml *node, const svg_state *inherit_state)
 {
-	svg_state local_state = *inherit_state;
+	svg_state local_state;
+	fz_path *path = NULL;
 
 	const char *d_att = fz_xml_att(node, "d");
 	/* unused: char *path_length_att = fz_xml_att(node, "pathLength"); */
 
-	svg_parse_common(ctx, doc, node, &local_state);
+	fz_var(path);
 
-	if (d_att)
+	svg_begin_state(ctx, &local_state, inherit_state);
+	fz_try(ctx)
 	{
-		fz_path *path = fz_new_path(ctx);
-		fz_try(ctx)
+		svg_parse_common(ctx, doc, node, &local_state);
+
+		if (d_att)
 		{
+			path = fz_new_path(ctx);
 			svg_parse_path_data(ctx, path, d_att);
 			svg_draw_path(ctx, dev, doc, path, &local_state);
 		}
-		fz_always(ctx)
-			fz_drop_path(ctx, path);
-		fz_catch(ctx)
-			fz_rethrow(ctx);
 	}
+	fz_always(ctx)
+	{
+		fz_drop_path(ctx, path);
+		svg_end_state(ctx, &local_state);
+	}
+	fz_catch(ctx)
+		fz_rethrow(ctx);
 }
 
 /* svg, symbol, image, foreignObject establish new viewports */
@@ -954,7 +1042,7 @@ static const char *linejoin_table[] = { "miter", "round", "bevel" };
 static void
 svg_parse_common(fz_context *ctx, svg_document *doc, fz_xml *node, svg_state *state)
 {
-	fz_stroke_state *stroke = &state->stroke;
+	fz_stroke_state *stroke = state->stroke = fz_unshare_stroke_state(ctx, state->stroke);
 
 	char *transform_att = fz_xml_att(node, "transform");
 
@@ -1059,7 +1147,7 @@ svg_parse_common(fz_context *ctx, svg_document *doc, fz_xml *node, svg_state *st
 	}
 	else
 	{
-		stroke->linewidth = svg_parse_number_from_style(ctx, doc, style_att, "stroke-width", state->stroke.linewidth);
+		stroke->linewidth = svg_parse_number_from_style(ctx, doc, style_att, "stroke-width", state->stroke->linewidth);
 	}
 
 	if (stroke_linecap_att)
@@ -1074,7 +1162,7 @@ svg_parse_common(fz_context *ctx, svg_document *doc, fz_xml *node, svg_state *st
 	else
 	{
 		stroke->start_cap = svg_parse_enum_from_style(ctx, doc, style_att, "stroke-linecap",
-			nelem(linecap_table), linecap_table, FZ_LINECAP_BUTT);
+			nelem(linecap_table), linecap_table, stroke->start_cap);
 	}
 
 	stroke->dash_cap = stroke->start_cap;
@@ -1092,7 +1180,7 @@ svg_parse_common(fz_context *ctx, svg_document *doc, fz_xml *node, svg_state *st
 	else
 	{
 		stroke->linejoin = svg_parse_enum_from_style(ctx, doc, style_att, "stroke-linejoin",
-			nelem(linejoin_table), linejoin_table, FZ_LINEJOIN_MITER);
+			nelem(linejoin_table), linejoin_table, stroke->linejoin);
 	}
 
 	if (stroke_miterlimit_att)
@@ -1104,7 +1192,7 @@ svg_parse_common(fz_context *ctx, svg_document *doc, fz_xml *node, svg_state *st
 	}
 	else
 	{
-		stroke->miterlimit = svg_parse_number_from_style(ctx, doc, style_att, "stroke-miterlimit", state->stroke.miterlimit);
+		stroke->miterlimit = svg_parse_number_from_style(ctx, doc, style_att, "stroke-miterlimit", state->stroke->miterlimit);
 	}
 }
 
@@ -1172,118 +1260,153 @@ svg_parse_font_attributes(fz_context *ctx, svg_document *doc, fz_xml *node, svg_
 static void
 svg_run_svg(fz_context *ctx, fz_device *dev, svg_document *doc, fz_xml *root, const svg_state *inherit_state)
 {
-	svg_state local_state = *inherit_state;
+	svg_state local_state;
 	fz_xml *node;
 
 	char *w_att = fz_xml_att(root, "width");
 	char *h_att = fz_xml_att(root, "height");
 	char *viewbox_att = fz_xml_att(root, "viewBox");
 
-	/* get default viewport from viewBox if width and/or height is missing */
-	if (viewbox_att && (!w_att || !h_att))
+	svg_begin_state(ctx, &local_state, inherit_state);
+	fz_try(ctx)
 	{
-		float x, y;
-		svg_lex_viewbox(viewbox_att, &x, &y, &local_state.viewbox_w, &local_state.viewbox_h);
-		if (!w_att) local_state.viewport_w = local_state.viewbox_w;
-		if (!h_att) local_state.viewport_h = local_state.viewbox_h;
+		/* get default viewport from viewBox if width and/or height is missing */
+		if (viewbox_att && (!w_att || !h_att))
+		{
+			float x, y;
+			svg_lex_viewbox(viewbox_att, &x, &y, &local_state.viewbox_w, &local_state.viewbox_h);
+			if (!w_att) local_state.viewport_w = local_state.viewbox_w;
+			if (!h_att) local_state.viewport_h = local_state.viewbox_h;
+		}
+
+		svg_parse_viewport(ctx, doc, root, &local_state);
+		svg_parse_viewbox(ctx, doc, root, &local_state);
+		svg_parse_common(ctx, doc, root, &local_state);
+
+		for (node = fz_xml_down(root); node; node = fz_xml_next(node))
+			svg_run_element(ctx, dev, doc, node, &local_state);
 	}
-
-	svg_parse_viewport(ctx, doc, root, &local_state);
-	svg_parse_viewbox(ctx, doc, root, &local_state);
-	svg_parse_common(ctx, doc, root, &local_state);
-
-	for (node = fz_xml_down(root); node; node = fz_xml_next(node))
-		svg_run_element(ctx, dev, doc, node, &local_state);
+	fz_always(ctx)
+		svg_end_state(ctx, &local_state);
+	fz_catch(ctx)
+		fz_rethrow(ctx);
 }
 
 static void
 svg_run_g(fz_context *ctx, fz_device *dev, svg_document *doc, fz_xml *root, const svg_state *inherit_state)
 {
-	svg_state local_state = *inherit_state;
+	svg_state local_state;
 	fz_xml *node;
 
-	svg_parse_common(ctx, doc, root, &local_state);
+	svg_begin_state(ctx, &local_state, inherit_state);
+	fz_try(ctx)
+	{
+		svg_parse_common(ctx, doc, root, &local_state);
 
-	for (node = fz_xml_down(root); node; node = fz_xml_next(node))
-		svg_run_element(ctx, dev, doc, node, &local_state);
+		for (node = fz_xml_down(root); node; node = fz_xml_next(node))
+			svg_run_element(ctx, dev, doc, node, &local_state);
+	}
+	fz_always(ctx)
+		svg_end_state(ctx, &local_state);
+	fz_catch(ctx)
+		fz_rethrow(ctx);
 }
 
 static void
 svg_run_use_symbol(fz_context *ctx, fz_device *dev, svg_document *doc, fz_xml *use, fz_xml *symbol, const svg_state *inherit_state)
 {
-	svg_state local_state = *inherit_state;
+	svg_state local_state;
 	fz_xml *node;
 
-	svg_parse_viewport(ctx, doc, use, &local_state);
-	svg_parse_viewbox(ctx, doc, use, &local_state);
-
-	for (node = fz_xml_down(symbol); node; node = fz_xml_next(node))
-		svg_run_element(ctx, dev, doc, node, &local_state);
-}
-
-static int
-is_use_cycle(fz_xml *use, fz_xml *symbol)
-{
-	/* If "use" is a direct child of "symbol", we have a recursive symbol/use definition! */
-	while (use)
+	svg_begin_state(ctx, &local_state, inherit_state);
+	fz_try(ctx)
 	{
-		if (use == symbol)
-			return 1;
-		use = fz_xml_up(use);
+		svg_parse_viewport(ctx, doc, use, &local_state);
+		svg_parse_viewbox(ctx, doc, use, &local_state);
+
+		for (node = fz_xml_down(symbol); node; node = fz_xml_next(node))
+			svg_run_element(ctx, dev, doc, node, &local_state);
 	}
-	return 0;
+	fz_always(ctx)
+		svg_end_state(ctx, &local_state);
+	fz_catch(ctx)
+		fz_rethrow(ctx);
 }
 
 static void
 svg_run_use(fz_context *ctx, fz_device *dev, svg_document *doc, fz_xml *root, const svg_state *inherit_state)
 {
-	svg_state local_state = *inherit_state;
+	svg_state local_state;
+	svg_cycle_list cycle;
 
 	char *href_att = fz_xml_att_alt(root, "xlink:href", "href");
 	char *x_att = fz_xml_att(root, "x");
 	char *y_att = fz_xml_att(root, "y");
+	fz_xml *linked = NULL;
 
 	float x = 0;
 	float y = 0;
 
-	if (++local_state.use_depth > MAX_USE_DEPTH)
+	svg_begin_state(ctx, &local_state, inherit_state);
+	fz_try(ctx)
 	{
-		fz_warn(ctx, "svg: too much recursion");
-		return;
-	}
-
-	svg_parse_common(ctx, doc, root, &local_state);
-	if (x_att) x = svg_parse_length(x_att, local_state.viewbox_w, local_state.fontsize);
-	if (y_att) y = svg_parse_length(y_att, local_state.viewbox_h, local_state.fontsize);
-
-	local_state.transform = fz_concat(fz_translate(x, y), local_state.transform);
-
-	if (href_att && href_att[0] == '#')
-	{
-		fz_xml *linked = fz_tree_lookup(ctx, doc->idmap, href_att + 1);
-		if (is_use_cycle(root, linked))
+		if (++local_state.use_depth > MAX_USE_DEPTH)
 		{
-			fz_warn(ctx, "svg: cyclic <use> reference");
-			return;
+			fz_warn(ctx, "svg: too much recursion");
+			break;
 		}
 
-		if (linked)
+		svg_parse_common(ctx, doc, root, &local_state);
+		if (x_att) x = svg_parse_length(x_att, local_state.viewbox_w, local_state.fontsize);
+		if (y_att) y = svg_parse_length(y_att, local_state.viewbox_h, local_state.fontsize);
+
+		local_state.transform = fz_concat(fz_translate(x, y), local_state.transform);
+
+		if (href_att && href_att[0] == '#')
 		{
-			if (fz_xml_is_tag(linked, "symbol"))
-				svg_run_use_symbol(ctx, dev, doc, root, linked, &local_state);
+			linked = fz_tree_lookup(ctx, doc->idmap, href_att + 1);
+			if (linked)
+			{
+				if (svg_push_use(ctx, doc, &cycle, linked))
+				{
+					fz_try(ctx)
+					{
+						if (fz_xml_is_tag(linked, "symbol"))
+							svg_run_use_symbol(ctx, dev, doc, root, linked, &local_state);
+						else
+							svg_run_element(ctx, dev, doc, linked, &local_state);
+					}
+					fz_always(ctx)
+						svg_pop_use(ctx, doc);
+					fz_catch(ctx)
+						fz_rethrow(ctx);
+				}
+				else
+				{
+					fz_warn(ctx, "svg: mutual recursion in <use> symbol");
+				}
+			}
 			else
-				svg_run_element(ctx, dev, doc, linked, &local_state);
-			return;
+			{
+				fz_warn(ctx, "svg: cannot find linked symbol");
+			}
+		}
+		else
+		{
+			fz_warn(ctx, "svg: cannot find linked symbol");
 		}
-	}
 
-	fz_warn(ctx, "svg: cannot find linked symbol");
+	}
+	fz_always(ctx)
+		svg_end_state(ctx, &local_state);
+	fz_catch(ctx)
+		fz_rethrow(ctx);
 }
 
 static void
 svg_run_image(fz_context *ctx, fz_device *dev, svg_document *doc, fz_xml *root, const svg_state *inherit_state)
 {
-	svg_state local_state = *inherit_state;
+	svg_state local_state;
 	float x=0, y=0, w=0, h=0;
 	const char *data;
 
@@ -1296,94 +1419,103 @@ svg_run_image(fz_context *ctx, fz_device *dev, svg_document *doc, fz_xml *root,
 	char *w_att = fz_xml_att(root, "width");
 	char *h_att = fz_xml_att(root, "height");
 
-	svg_parse_common(ctx, doc, root, &local_state);
-	if (x_att) x = svg_parse_length(x_att, local_state.viewbox_w, local_state.fontsize);
-	if (y_att) y = svg_parse_length(y_att, local_state.viewbox_h, local_state.fontsize);
-	if (w_att) w = svg_parse_length(w_att, local_state.viewbox_w, local_state.fontsize);
-	if (h_att) h = svg_parse_length(h_att, local_state.viewbox_h, local_state.fontsize);
-
-	if (w <= 0 || h <= 0)
-		return;
-
-	if (!href_att)
-		return;
-
-	local_state.transform = fz_concat(fz_translate(x, y), local_state.transform);
-	local_state.transform = fz_concat(fz_scale(w, h), local_state.transform);
-
-	if (!strncmp(href_att, jpeg_uri, strlen(jpeg_uri)))
-		data = href_att + strlen(jpeg_uri);
-	else if (!strncmp(href_att, png_uri, strlen(png_uri)))
-		data = href_att + strlen(png_uri);
-	else
-		data = NULL;
-	if (data)
+	svg_begin_state(ctx, &local_state, inherit_state);
+	fz_try(ctx)
 	{
-		fz_image *img = NULL;
-		fz_buffer *buf;
+		svg_parse_common(ctx, doc, root, &local_state);
+		if (x_att) x = svg_parse_length(x_att, local_state.viewbox_w, local_state.fontsize);
+		if (y_att) y = svg_parse_length(y_att, local_state.viewbox_h, local_state.fontsize);
+		if (w_att) w = svg_parse_length(w_att, local_state.viewbox_w, local_state.fontsize);
+		if (h_att) h = svg_parse_length(h_att, local_state.viewbox_h, local_state.fontsize);
 
-		fz_var(img);
+		if (w <= 0 || h <= 0)
+			break; // out of try-catch
 
-		buf = fz_new_buffer_from_base64(ctx, data, 0);
-		fz_try(ctx)
-		{
-			fz_matrix orient;
-			img = fz_new_image_from_buffer(ctx, buf);
-			orient = fz_image_orientation_matrix(ctx, img);
-			local_state.transform = fz_concat(orient, local_state.transform);
-			fz_fill_image(ctx, dev, img, local_state.transform, 1, fz_default_color_params);
-		}
-		fz_always(ctx)
-		{
-			fz_drop_buffer(ctx, buf);
-			fz_drop_image(ctx, img);
-		}
-		fz_catch(ctx)
-		{
-			fz_rethrow_if(ctx, FZ_ERROR_SYSTEM);
-			fz_report_error(ctx);
-			fz_warn(ctx, "svg: ignoring embedded image '%s'", href_att);
-		}
-	}
-	else if (doc->zip)
-	{
-		char path[2048];
-		fz_buffer *buf = NULL;
-		fz_image *img = NULL;
+		if (!href_att)
+			break; // out of try-catch
 
-		fz_var(buf);
-		fz_var(img);
+		local_state.transform = fz_concat(fz_translate(x, y), local_state.transform);
+		local_state.transform = fz_concat(fz_scale(w, h), local_state.transform);
 
-		fz_strlcpy(path, doc->base_uri, sizeof path);
-		fz_strlcat(path, "/", sizeof path);
-		fz_strlcat(path, href_att, sizeof path);
-		fz_urldecode(path);
+		if (!strncmp(href_att, jpeg_uri, strlen(jpeg_uri)))
+			data = href_att + strlen(jpeg_uri);
+		else if (!strncmp(href_att, png_uri, strlen(png_uri)))
+			data = href_att + strlen(png_uri);
+		else
+			data = NULL;
+		if (data)
+		{
+			fz_image *img = NULL;
+			fz_buffer *buf;
 
-		fz_try(ctx)
-		{
-			fz_matrix orient;
-			buf = fz_read_archive_entry(ctx, doc->zip, path);
-			img = fz_new_image_from_buffer(ctx, buf);
-			orient = fz_image_orientation_matrix(ctx, img);
-			local_state.transform = fz_concat(orient, local_state.transform);
-			fz_fill_image(ctx, dev, img, local_state.transform, 1, fz_default_color_params);
+			fz_var(img);
+
+			buf = fz_new_buffer_from_base64(ctx, data, 0);
+			fz_try(ctx)
+			{
+				fz_matrix orient;
+				img = fz_new_image_from_buffer(ctx, buf);
+				orient = fz_image_orientation_matrix(ctx, img);
+				local_state.transform = fz_concat(orient, local_state.transform);
+				fz_fill_image(ctx, dev, img, local_state.transform, 1, fz_default_color_params);
+			}
+			fz_always(ctx)
+			{
+				fz_drop_buffer(ctx, buf);
+				fz_drop_image(ctx, img);
+			}
+			fz_catch(ctx)
+			{
+				fz_rethrow_if(ctx, FZ_ERROR_SYSTEM);
+				fz_report_error(ctx);
+				fz_warn(ctx, "svg: ignoring embedded image '%s'", href_att);
+			}
 		}
-		fz_always(ctx)
+		else if (doc->zip)
 		{
-			fz_drop_buffer(ctx, buf);
-			fz_drop_image(ctx, img);
+			char path[2048];
+			fz_buffer *buf = NULL;
+			fz_image *img = NULL;
+
+			fz_var(buf);
+			fz_var(img);
+
+			fz_strlcpy(path, doc->base_uri, sizeof path);
+			fz_strlcat(path, "/", sizeof path);
+			fz_strlcat(path, href_att, sizeof path);
+			fz_urldecode(path);
+
+			fz_try(ctx)
+			{
+				fz_matrix orient;
+				buf = fz_read_archive_entry(ctx, doc->zip, path);
+				img = fz_new_image_from_buffer(ctx, buf);
+				orient = fz_image_orientation_matrix(ctx, img);
+				local_state.transform = fz_concat(orient, local_state.transform);
+				fz_fill_image(ctx, dev, img, local_state.transform, 1, fz_default_color_params);
+			}
+			fz_always(ctx)
+			{
+				fz_drop_buffer(ctx, buf);
+				fz_drop_image(ctx, img);
+			}
+			fz_catch(ctx)
+			{
+				fz_rethrow_if(ctx, FZ_ERROR_SYSTEM);
+				fz_report_error(ctx);
+				fz_warn(ctx, "svg: ignoring external image '%s'", href_att);
+			}
 		}
-		fz_catch(ctx)
+		else
 		{
-			fz_rethrow_if(ctx, FZ_ERROR_SYSTEM);
-			fz_report_error(ctx);
 			fz_warn(ctx, "svg: ignoring external image '%s'", href_att);
 		}
+
 	}
-	else
-	{
-		fz_warn(ctx, "svg: ignoring external image '%s'", href_att);
-	}
+	fz_always(ctx)
+		svg_end_state(ctx, &local_state);
+	fz_catch(ctx)
+		fz_rethrow(ctx);
 }
 
 static fz_font *
@@ -1463,7 +1595,7 @@ svg_run_text_string(fz_context *ctx, fz_device *dev, fz_matrix trm, const char *
 				fz_default_color_params);
 		if (state->stroke_is_set)
 			fz_stroke_text(ctx, dev, text,
-				&state->stroke,
+				state->stroke,
 				state->transform,
 				fz_device_rgb(ctx), state->stroke_color,
 				state->opacity,
@@ -1509,7 +1641,7 @@ static fz_matrix
 svg_run_text(fz_context *ctx, fz_device *dev, svg_document *doc, fz_xml *root, const svg_state *inherit_state,
 	float x, float y, int is_first, int is_last)
 {
-	svg_state local_state = *inherit_state;
+	svg_state local_state;
 	char font_family[100];
 	fz_xml *node;
 	fz_matrix trm;
@@ -1521,35 +1653,43 @@ svg_run_text(fz_context *ctx, fz_device *dev, svg_document *doc, fz_xml *root, c
 	char *dx_att = fz_xml_att(root, "dx");
 	char *dy_att = fz_xml_att(root, "dy");
 
-	svg_parse_common(ctx, doc, root, &local_state);
-	svg_parse_font_attributes(ctx, doc, root, &local_state, font_family, sizeof font_family);
-
-	trm = fz_scale(local_state.fontsize, -local_state.fontsize);
-	trm.e = x;
-	trm.f = y;
-
-	if (x_att) trm.e = svg_parse_length(x_att, local_state.viewbox_w, local_state.fontsize);
-	if (y_att) trm.f = svg_parse_length(y_att, local_state.viewbox_h, local_state.fontsize);
-
-	if (dx_att) trm.e += svg_parse_length(dx_att, local_state.viewbox_w, local_state.fontsize);
-	if (dy_att) trm.f += svg_parse_length(dy_att, local_state.viewbox_h, local_state.fontsize);
-
-	cif = is_first;
-	for (node = fz_xml_down(root); node; node = fz_xml_next(node))
+	svg_begin_state(ctx, &local_state, inherit_state);
+	fz_try(ctx)
 	{
-		cil = is_last && !fz_xml_next(node);
-		text = fz_xml_text(node);
-		if (text)
+		svg_parse_common(ctx, doc, root, &local_state);
+		svg_parse_font_attributes(ctx, doc, root, &local_state, font_family, sizeof font_family);
+
+		trm = fz_scale(local_state.fontsize, -local_state.fontsize);
+		trm.e = x;
+		trm.f = y;
+
+		if (x_att) trm.e = svg_parse_length(x_att, local_state.viewbox_w, local_state.fontsize);
+		if (y_att) trm.f = svg_parse_length(y_att, local_state.viewbox_h, local_state.fontsize);
+
+		if (dx_att) trm.e += svg_parse_length(dx_att, local_state.viewbox_w, local_state.fontsize);
+		if (dy_att) trm.f += svg_parse_length(dy_att, local_state.viewbox_h, local_state.fontsize);
+
+		cif = is_first;
+		for (node = fz_xml_down(root); node; node = fz_xml_next(node))
 		{
-			svg_collapse_whitespace(text, cif, cil);
-			trm = svg_run_text_string(ctx, dev, trm, text, &local_state);
+			cil = is_last && !fz_xml_next(node);
+			text = fz_xml_text(node);
+			if (text)
+			{
+				svg_collapse_whitespace(text, cif, cil);
+				trm = svg_run_text_string(ctx, dev, trm, text, &local_state);
+			}
+			else if (fz_xml_is_tag(node, "tspan"))
+				trm = svg_run_text(ctx, dev, doc, node, &local_state, trm.e, trm.f, cif, cil);
+			else if (fz_xml_is_tag(node, "textPath"))
+				trm = svg_run_text(ctx, dev, doc, node, &local_state, trm.e, trm.f, cif, cil);
+			cif = 0;
 		}
-		else if (fz_xml_is_tag(node, "tspan"))
-			trm = svg_run_text(ctx, dev, doc, node, &local_state, trm.e, trm.f, cif, cil);
-		else if (fz_xml_is_tag(node, "textPath"))
-			trm = svg_run_text(ctx, dev, doc, node, &local_state, trm.e, trm.f, cif, cil);
-		cif = 0;
 	}
+	fz_always(ctx)
+		svg_end_state(ctx, &local_state);
+	fz_catch(ctx)
+		fz_rethrow(ctx);
 
 	return trm;
 }
@@ -1656,7 +1796,7 @@ svg_run_document(fz_context *ctx, svg_document *doc, fz_xml *root, fz_device *de
 
 	/* Initial graphics state */
 	state.transform = ctm;
-	state.stroke = fz_default_stroke_state;
+	state.stroke = fz_new_stroke_state(ctx);
 	state.use_depth = 0;
 
 	state.viewport_w = DEF_WIDTH;
@@ -1689,5 +1829,16 @@ svg_run_document(fz_context *ctx, svg_document *doc, fz_xml *root, fz_device *de
 	state.is_italic = 0;
 	state.text_anchor = 0;
 
-	svg_run_svg(ctx, dev, doc, root, &state);
+	fz_try(ctx)
+	{
+		svg_run_svg(ctx, dev, doc, root, &state);
+	}
+	fz_always(ctx)
+	{
+		fz_drop_stroke_state(ctx, state.stroke);
+	}
+	fz_catch(ctx)
+	{
+		fz_rethrow(ctx);
+	}
 }
