@@ -1,4 +1,4 @@
-// Copyright (C) 2004-2023 Artifex Software, Inc.
+// Copyright (C) 2004-2025 Artifex Software, Inc.
 //
 // This file is part of MuPDF.
 //
@@ -44,7 +44,7 @@ static const char *format = NULL;
 static const char *options = "";
 
 static fz_context *ctx;
-static fz_document *doc;
+static fz_document *doc = NULL;
 static fz_document_writer *out;
 static fz_box_type page_box = FZ_CROP_BOX;
 static int count;
@@ -151,7 +151,7 @@ int muconvert_main(int argc, char **argv)
 		case 'U': layout_css = fz_optarg; break;
 		case 'X': layout_use_doc_css = 0; break;
 
-		case 'o': output = fz_optarg; break;
+		case 'o': output = fz_optpath(fz_optarg); break;
 		case 'F': format = fz_optarg; break;
 		case 'O': options = fz_optarg; break;
 
@@ -191,11 +191,7 @@ int muconvert_main(int argc, char **argv)
 	fz_set_aa_level(ctx, alphabits);
 
 	if (layout_css)
-	{
-		fz_buffer *buf = fz_read_file(ctx, layout_css);
-		fz_set_user_css(ctx, fz_string_from_buffer(ctx, buf));
-		fz_drop_buffer(ctx, buf);
-	}
+		fz_load_user_css(ctx, layout_css);
 
 	fz_set_use_document_css(ctx, layout_use_doc_css);
 
