@@ -1,4 +1,4 @@
-// Copyright (C) 2004-2024 Artifex Software, Inc.
+// Copyright (C) 2004-2025 Artifex Software, Inc.
 //
 // This file is part of MuPDF.
 //
@@ -47,7 +47,7 @@ static int usage(void)
 		"\t-gg\tin addition to -g compact xref table\n"
 		"\t-ggg\tin addition to -gg merge duplicate objects\n"
 		"\t-gggg\tin addition to -ggg check streams for duplication\n"
-		"\t-l\tlinearize PDF\n"
+		"\t-l\tlinearize PDF (no longer supported!)\n"
 		"\t-D\tsave file without encryption\n"
 		"\t-E -\tsave file with new encryption (rc4-40, rc4-128, aes-128, or aes-256)\n"
 		"\t-O -\towner password (only if encrypting)\n"
@@ -61,7 +61,10 @@ static int usage(void)
 		"\t-i\tcompress image streams\n"
 		"\t-c\tclean content streams\n"
 		"\t-s\tsanitize content streams\n"
-		"\t-t\t'tighten' objects\n"
+		"\t-t\tcompact object syntax\n"
+		"\t-tt\tindented object syntax\n"
+		"\t-L\twrite object labels\n"
+		"\t-v\tvectorize text\n"
 		"\t-A\tcreate appearance streams for annotations\n"
 		"\t-AA\trecreate appearance streams for annotations\n"
 		"\t-m\tpreserve metadata\n"
@@ -70,6 +73,7 @@ static int usage(void)
 		"\t--{color,gray,bitonal}-{,lossy-,lossless-}image-subsample-method -\n\t\taverage, bicubic\n"
 		"\t--{color,gray,bitonal}-{,lossy-,lossless-}image-subsample-dpi -[,-]\n\t\tDPI at which to subsample [+ target dpi]\n"
 		"\t--{color,gray,bitonal}-{,lossy-,lossless-}image-recompress-method -[:quality]\n\t\tnever, same, lossless, jpeg, j2k, fax, jbig2\n"
+		"\t--structure=keep|drop\tKeep or drop the structure tree\n"
 		"\tpages\tcomma separated list of page numbers and ranges\n"
 		);
 	return 1;
@@ -90,10 +94,11 @@ int pdfclean_main(int argc, char **argv)
 	char *outfile = "out.pdf";
 	char *password = "";
 	int c;
-	int tighten = 0;
+	int pretty = -1;
 	pdf_clean_options opts = { 0 };
 	int errors = 0;
 	fz_context *ctx;
+	int structure;
 	const fz_getopt_long_options longopts[] =
 	{
 		{ "color-lossy-image-subsample-method=average|bicubic", &opts.image.color_lossy_image_subsample_method, (void *)1 },
@@ -120,6 +125,8 @@ int pdfclean_main(int argc, char **argv)
 		{ "bitonal-image-subsample-dpi:", &opts.image.bitonal_image_subsample_threshold, (void *)20 },
 		{ "bitonal-image-recompress-method=never|same|lossless|jpeg:|j2k:|fax|jbig2", &opts.image.bitonal_image_recompress_method, (void *)21 },
 
+		{ "structure=drop|keep", &structure, (void *)22 },
+
 		{ NULL, NULL, NULL }
 	};
 
@@ -127,7 +134,7 @@ int pdfclean_main(int argc, char **argv)
 	opts.write = pdf_default_write_options;
 	opts.write.dont_regenerate_id = 1;
 
-	while ((c = fz_getopt_long(argc, argv, "ade:fgilmp:stczDAE:O:U:P:SZ", longopts)) != -1)
+	while ((c = fz_getopt_long(argc, argv, "ade:fgilmp:stcvzDAE:LO:U:P:SZ", longopts)) != -1)
 	{
 		switch (c)
 		{
@@ -143,8 +150,9 @@ int pdfclean_main(int argc, char **argv)
 		case 'l': opts.write.do_linear += 1; break;
 		case 'c': opts.write.do_clean += 1; break;
 		case 's': opts.write.do_sanitize += 1; break;
-		case 't': tighten = 1; break;
+		case 't': pretty = (pretty < 0) ? 0 : 1; break;
 		case 'A': opts.write.do_appearance += 1; break;
+		case 'L': opts.write.do_labels = 1; break;
 
 		case 'D': opts.write.do_encrypt = PDF_ENCRYPT_NONE; break;
 		case 'E': opts.write.do_encrypt = encrypt_method_from_string(fz_optarg); break;
@@ -154,6 +162,7 @@ int pdfclean_main(int argc, char **argv)
 		case 'm': opts.write.do_preserve_metadata = 1; break;
 		case 'S': opts.subset_fonts = 1; break;
 		case 'Z': opts.write.do_use_objstms = 1; break;
+		case 'v': opts.vectorize++; break;
 		case 0:
 		{
 			switch((int)(intptr_t)fz_optlong->opaque)
@@ -231,6 +240,9 @@ int pdfclean_main(int argc, char **argv)
 				if (fz_optarg)
 					return usage();
 				break;
+			case 22: /* structure */
+				opts.structure = structure; /* Allow for int/enum size mismatch. */
+				break;
 			}
 			break;
 		}
@@ -238,8 +250,14 @@ int pdfclean_main(int argc, char **argv)
 		}
 	}
 
-	if ((opts.write.do_ascii || opts.write.do_decompress) && !opts.write.do_compress)
-		opts.write.do_pretty = !tighten;
+	if (pretty < 0)
+	{
+		if ((opts.write.do_ascii || opts.write.do_decompress) && !opts.write.do_compress)
+			pretty = 1;
+		else
+			pretty = 0;
+	}
+	opts.write.do_pretty = pretty;
 
 	if (argc - fz_optind < 1)
 		return usage();
@@ -249,7 +267,7 @@ int pdfclean_main(int argc, char **argv)
 	if (argc - fz_optind > 0 &&
 		(strstr(argv[fz_optind], ".pdf") || strstr(argv[fz_optind], ".PDF")))
 	{
-		outfile = argv[fz_optind++];
+		outfile = fz_optpath(argv[fz_optind++]);
 	}
 
 	ctx = fz_new_context(NULL, NULL, FZ_STORE_UNLIMITED);
@@ -259,6 +277,9 @@ int pdfclean_main(int argc, char **argv)
 		exit(1);
 	}
 
+	if (opts.write.do_compress > 1)
+		fz_warn(ctx, "Brotli compression is currently non-standard and experimental. Files may not be readable in other software.");
+
 	fz_try(ctx)
 	{
 		pdf_clean_file(ctx, infile, outfile, password, &opts, argc - fz_optind, &argv[fz_optind]);
