@@ -40,7 +40,8 @@ enum
 	SHADINGS = 0x08,
 	PATTERNS = 0x10,
 	XOBJS = 0x20,
-	ALL = DIMENSIONS | FONTS | IMAGES | SHADINGS | PATTERNS | XOBJS
+	ZUGFERD = 0x40,
+	ALL = DIMENSIONS | FONTS | IMAGES | SHADINGS | PATTERNS | XOBJS | ZUGFERD
 };
 
 struct info
@@ -193,6 +194,7 @@ infousage(void)
 		"\t-P\tlist patterns\n"
 		"\t-S\tlist shadings\n"
 		"\t-X\tlist form and postscript xobjects\n"
+		"\t-Z\tlist ZUGFeRD info\n"
 		"\tpages\tcomma separated list of page numbers and ranges\n"
 		);
 }
@@ -229,6 +231,7 @@ gatherdimensions(fz_context *ctx, globals *glo, int page, pdf_obj *pageref)
 {
 	fz_rect bbox;
 	pdf_obj *obj;
+	float unit;
 	int j;
 
 	obj = pdf_dict_get(ctx, pageref, PDF_NAME(MediaBox));
@@ -237,15 +240,11 @@ gatherdimensions(fz_context *ctx, globals *glo, int page, pdf_obj *pageref)
 
 	bbox = pdf_to_rect(ctx, obj);
 
-	obj = pdf_dict_get(ctx, pageref, PDF_NAME(UserUnit));
-	if (pdf_is_number(ctx, obj))
-	{
-		float unit = pdf_to_real(ctx, obj);
-		bbox.x0 *= unit;
-		bbox.y0 *= unit;
-		bbox.x1 *= unit;
-		bbox.y1 *= unit;
-	}
+	unit = pdf_dict_get_real_default(ctx, pageref, PDF_NAME(UserUnit), 1);
+	bbox.x0 *= unit;
+	bbox.y0 *= unit;
+	bbox.x1 *= unit;
+	bbox.y1 *= unit;
 
 	for (j = 0; j < glo->dims; j++)
 		if (!memcmp(glo->dim[j].u.dim.bbox, &bbox, sizeof (fz_rect)))
@@ -962,6 +961,30 @@ showinfo(fz_context *ctx, globals *glo, char *filename, int show, const char *pa
 		printinfo(ctx, glo, filename, show, -1);
 }
 
+static void
+showzugferd(fz_context *ctx, globals *glo)
+{
+	float version;
+	fz_output *out = glo->out;
+	enum pdf_zugferd_profile profile = pdf_zugferd_profile(ctx, glo->doc, &version);
+	fz_buffer *buf;
+
+	if (profile == PDF_NOT_ZUGFERD)
+	{
+		fz_write_printf(ctx, out, "Not a ZUGFeRD file.\n");
+		return;
+	}
+
+	fz_write_printf(ctx, out, "ZUGFeRD version %g\n", version);
+	fz_write_printf(ctx, out, "%s profile\n", pdf_zugferd_profile_to_string(ctx, profile));
+
+	fz_write_printf(ctx, out, "Embedded XML:\n");
+	buf = pdf_zugferd_xml(ctx, glo->doc);
+	fz_write_buffer(ctx, out, buf);
+	fz_drop_buffer(ctx, buf);
+	fz_write_printf(ctx, out, "\n\n");
+}
+
 static void
 pdfinfo_info(fz_context *ctx, fz_output *out, char *filename, char *password, int show, char *argv[], int argc)
 {
@@ -992,11 +1015,14 @@ pdfinfo_info(fz_context *ctx, fz_output *out, char *filename, char *password, in
 				glo.doc = pdf_open_document(glo.ctx, filename);
 				if (pdf_needs_password(ctx, glo.doc))
 					if (!pdf_authenticate_password(ctx, glo.doc, password))
-						fz_throw(glo.ctx, FZ_ERROR_GENERIC, "cannot authenticate password: %s", filename);
+						fz_throw(glo.ctx, FZ_ERROR_ARGUMENT, "cannot authenticate password: %s", filename);
 				glo.pagecount = pdf_count_pages(ctx, glo.doc);
 
 				showglobalinfo(ctx, &glo);
 				state = NO_INFO_GATHERED;
+
+				if (show & ZUGFERD)
+					showzugferd(ctx, &glo);
 			}
 			else
 			{
@@ -1025,7 +1051,7 @@ int pdfinfo_main(int argc, char **argv)
 	int ret;
 	fz_context *ctx;
 
-	while ((c = fz_getopt(argc, argv, "FISPXMp:")) != -1)
+	while ((c = fz_getopt(argc, argv, "FISPXMZp:")) != -1)
 	{
 		switch (c)
 		{
@@ -1035,6 +1061,7 @@ int pdfinfo_main(int argc, char **argv)
 		case 'P': if (show == ALL) show = PATTERNS; else show |= PATTERNS; break;
 		case 'X': if (show == ALL) show = XOBJS; else show |= XOBJS; break;
 		case 'M': if (show == ALL) show = DIMENSIONS; else show |= DIMENSIONS; break;
+		case 'Z': if (show == ALL) show = ZUGFERD; else show |= ZUGFERD; break;
 		case 'p': password = fz_optarg; break;
 		default:
 			infousage();
@@ -1060,7 +1087,7 @@ int pdfinfo_main(int argc, char **argv)
 		pdfinfo_info(ctx, fz_stdout(ctx), filename, password, show, &argv[fz_optind], argc-fz_optind);
 	fz_catch(ctx)
 	{
-		fz_log_error(ctx, fz_caught_message(ctx));
+		fz_report_error(ctx);
 		ret = 1;
 	}
 	fz_drop_context(ctx);
