@@ -1,4 +1,4 @@
-// Copyright (C) 2004-2024 Artifex Software, Inc.
+// Copyright (C) 2004-2025 Artifex Software, Inc.
 //
 // This file is part of MuPDF.
 //
@@ -33,6 +33,8 @@
 
 static int x_factor = 0;
 static int y_factor = 0;
+static float margin = 0;
+static int margin_is_percent = 0;
 static int x_dir = 1;
 
 static int usage(void)
@@ -40,6 +42,7 @@ static int usage(void)
 	fprintf(stderr,
 		"usage: mutool poster [options] input.pdf [output.pdf]\n"
 		"\t-p -\tpassword\n"
+		"\t-m -\tmargin (overlap) between pages (pts, or %%)\n"
 		"\t-x\tx decimation factor\n"
 		"\t-y\ty decimation factor\n"
 		"\t-r\tsplit right-to-left\n"
@@ -113,6 +116,7 @@ static void decimatepages(fz_context *ctx, pdf_document *doc)
 		float w, h;
 		int x, y;
 		int xd, yd, y0, y1;
+		float xmargin, ymargin;
 
 		rotate = pdf_to_int(ctx, pdf_dict_get_inheritable(ctx, page_obj, PDF_NAME(Rotate)));
 		mediabox = pdf_to_rect(ctx, pdf_dict_get_inheritable(ctx, page_obj, PDF_NAME(MediaBox)));
@@ -125,19 +129,33 @@ static void decimatepages(fz_context *ctx, pdf_document *doc)
 		w = mediabox.x1 - mediabox.x0;
 		h = mediabox.y1 - mediabox.y0;
 
-		if (rotate == 90 || rotate == 270)
-		{
-			xf = y_factor;
-			yf = x_factor;
-			yd = -x_dir;
-			xd = 1;
-		}
-		else
+		switch (rotate)
 		{
+		default:
+		case 0:
 			xf = x_factor;
 			yf = y_factor;
 			xd = x_dir;
-			yd = -1;
+			yd = -x_dir;
+			break;
+		case 90:
+			xf = y_factor;
+			yf = x_factor;
+			yd = x_dir;
+			xd = x_dir;
+			break;
+		case 180:
+			xf = x_factor;
+			yf = y_factor;
+			xd = -x_dir;
+			yd = x_dir;
+			break;
+		case 270:
+			xf = y_factor;
+			yf = x_factor;
+			yd = -x_dir;
+			xd = -x_dir;
+			break;
 		}
 
 		if (xf == 0 && yf == 0)
@@ -153,6 +171,9 @@ static void decimatepages(fz_context *ctx, pdf_document *doc)
 		else if (yf == 0)
 			yf = 1;
 
+		xmargin = xf == 1 ? 0 : margin * (margin_is_percent ? .001 * w/xf : 1);
+		ymargin = yf == 1 ? 0 : margin * (margin_is_percent ? .001 * h/yf : 1);
+
 		y0 = (yd > 0) ? 0 : yf-1;
 		y1 = (yd > 0) ? yf : -1;
 		for (y = y0; y != y1; y += yd)
@@ -174,12 +195,12 @@ static void decimatepages(fz_context *ctx, pdf_document *doc)
 				if (x == xf-1)
 					mb.x1 = mediabox.x1;
 				else
-					mb.x1 = mediabox.x0 + (w/xf)*(x+1);
+					mb.x1 = mb.x0 + (w/xf) + xmargin;
 				mb.y0 = mediabox.y0 + (h/yf)*y;
 				if (y == yf-1)
 					mb.y1 = mediabox.y1;
 				else
-					mb.y1 = mediabox.y0 + (h/yf)*(y+1);
+					mb.y1 = mb.y0 + (h/yf) + ymargin;
 
 				pdf_array_push_real(ctx, newmediabox, mb.x0);
 				pdf_array_push_real(ctx, newmediabox, mb.y0);
@@ -215,15 +236,16 @@ int pdfposter_main(int argc, char **argv)
 	char *password = "";
 	int c;
 	pdf_write_options opts = pdf_default_write_options;
-	pdf_document *doc;
+	pdf_document *doc = NULL;
 	fz_context *ctx;
 	int ret = 0;
 
-	while ((c = fz_getopt(argc, argv, "x:y:p:r")) != -1)
+	while ((c = fz_getopt(argc, argv, "m:x:y:p:r")) != -1)
 	{
 		switch (c)
 		{
 		case 'p': password = fz_optarg; break;
+		case 'm': margin = atof(fz_optarg); margin_is_percent = (strchr(fz_optarg, '%') != NULL); break;
 		case 'x': x_factor = atoi(fz_optarg); break;
 		case 'y': y_factor = atoi(fz_optarg); break;
 		case 'r': x_dir = -1; break;
@@ -239,7 +261,7 @@ int pdfposter_main(int argc, char **argv)
 	if (argc - fz_optind > 0 &&
 		(strstr(argv[fz_optind], ".pdf") || strstr(argv[fz_optind], ".PDF")))
 	{
-		outfile = argv[fz_optind++];
+		outfile = fz_optpath(argv[fz_optind++]);
 	}
 
 	ctx = fz_new_context(NULL, NULL, FZ_STORE_UNLIMITED);
