@@ -1,4 +1,4 @@
-// Copyright (C) 2004-2022 Artifex Software, Inc.
+// Copyright (C) 2004-2025 Artifex Software, Inc.
 //
 // This file is part of MuPDF.
 //
@@ -32,11 +32,13 @@
 #include <string.h>
 
 static pdf_document *doc = NULL;
-static fz_context *ctx = NULL;
 static fz_output *out = NULL;
+static pdf_object_labels *labels = NULL;
 static int showbinary = 0;
 static int showdecode = 1;
-static int tight = 0;
+static int do_tight = 0;
+static int do_repair = 0;
+static int do_label = 0;
 static int showcolumn;
 
 static int usage(void)
@@ -48,6 +50,8 @@ static int usage(void)
 		"\t-e\tleave stream contents in their original form\n"
 		"\t-b\tprint only stream contents, as raw binary data\n"
 		"\t-g\tprint only object, one line per object, suitable for grep\n"
+		"\t-r\tforce repair before showing any objects\n"
+		"\t-L\tshow object labels\n"
 		"\tpath: path to an object, starting with either an object number,\n"
 		"\t\t'pages', 'trailer', or a property in the trailer;\n"
 		"\t\tpath elements separated by '.' or '/'. Path elements must be\n"
@@ -56,17 +60,17 @@ static int usage(void)
 	return 1;
 }
 
-static void showtrailer(void)
+static void showtrailer(fz_context *ctx)
 {
-	if (tight)
+	if (do_tight)
 		fz_write_printf(ctx, out, "trailer ");
 	else
 		fz_write_printf(ctx, out, "trailer\n");
-	pdf_print_obj(ctx, out, pdf_trailer(ctx, doc), tight, 1);
+	pdf_print_obj(ctx, out, pdf_trailer(ctx, doc), do_tight, 1);
 	fz_write_printf(ctx, out, "\n");
 }
 
-static void showxref(void)
+static void showxref(fz_context *ctx)
 {
 	int i;
 	int xref_len = pdf_xref_len(ctx, doc);
@@ -82,7 +86,7 @@ static void showxref(void)
 	}
 }
 
-static void showpages(void)
+static void showpages(fz_context *ctx)
 {
 	pdf_obj *ref;
 	int i, n = pdf_count_pages(ctx, doc);
@@ -116,7 +120,7 @@ static void showsafe(unsigned char *buf, size_t n)
 	}
 }
 
-static void showstream(int num)
+static void showstream(fz_context *ctx, int num)
 {
 	fz_stream *stm;
 	unsigned char buf[2048];
@@ -143,7 +147,12 @@ static void showstream(int num)
 	fz_drop_stream(ctx, stm);
 }
 
-static void showobject(pdf_obj *ref)
+static void showlabel(fz_context *ctx, void *arg, const char *label)
+{
+	fz_write_printf(ctx, arg, "%% %s\n", label);
+}
+
+static void showobject(fz_context *ctx, pdf_obj *ref)
 {
 	pdf_obj *obj = pdf_resolve_indirect(ctx, ref);
 	int num = pdf_to_num(ctx, ref);
@@ -151,11 +160,13 @@ static void showobject(pdf_obj *ref)
 	{
 		if (showbinary)
 		{
-			showstream(num);
+			showstream(ctx, num);
 		}
 		else
 		{
-			if (tight)
+			if (do_label)
+				pdf_label_object(ctx, labels, num, showlabel, out);
+			if (do_tight)
 			{
 				fz_write_printf(ctx, out, "%d 0 obj ", num);
 				pdf_print_obj(ctx, out, obj, 1, 1);
@@ -166,7 +177,7 @@ static void showobject(pdf_obj *ref)
 				fz_write_printf(ctx, out, "%d 0 obj\n", num);
 				pdf_print_obj(ctx, out, obj, 0, 1);
 				fz_write_printf(ctx, out, "\nstream\n");
-				showstream(num);
+				showstream(ctx, num);
 				fz_write_printf(ctx, out, "endstream\n");
 				fz_write_printf(ctx, out, "endobj\n");
 			}
@@ -174,7 +185,9 @@ static void showobject(pdf_obj *ref)
 	}
 	else
 	{
-		if (tight)
+		if (do_label)
+			pdf_label_object(ctx, labels, num, showlabel, out);
+		if (do_tight)
 		{
 			fz_write_printf(ctx, out, "%d 0 obj ", num);
 			pdf_print_obj(ctx, out, obj, 1, 1);
@@ -189,7 +202,7 @@ static void showobject(pdf_obj *ref)
 	}
 }
 
-static void showgrep(void)
+static void showgrep(fz_context *ctx)
 {
 	pdf_obj *ref, *obj;
 	int i, len;
@@ -200,6 +213,9 @@ static void showgrep(void)
 		pdf_xref_entry *entry = pdf_get_xref_entry_no_null(ctx, doc, i);
 		if (entry->type == 'n' || entry->type == 'o')
 		{
+			ref = NULL;
+
+			fz_var(ref);
 			fz_try(ctx)
 			{
 				ref = pdf_new_indirect(ctx, doc, i, 0);
@@ -230,7 +246,7 @@ static void showgrep(void)
 }
 
 static void
-print_outline(fz_outline *outline, int level)
+print_outline(fz_context *ctx, fz_outline *outline, int level)
 {
 	int i;
 	while (outline)
@@ -244,23 +260,23 @@ print_outline(fz_outline *outline, int level)
 			fz_write_byte(ctx, out, '\t');
 		fz_write_printf(ctx, out, "%Q\t%s\n", outline->title, outline->uri);
 		if (outline->down)
-			print_outline(outline->down, level + 1);
+			print_outline(ctx, outline->down, level + 1);
 		outline = outline->next;
 	}
 }
 
-static void showoutline(void)
+static void showoutline(fz_context *ctx)
 {
 	fz_outline *outline = fz_load_outline(ctx, (fz_document*)doc);
 	fz_try(ctx)
-		print_outline(outline, 1);
+		print_outline(ctx, outline, 1);
 	fz_always(ctx)
 		fz_drop_outline(ctx, outline);
 	fz_catch(ctx)
 		fz_rethrow(ctx);
 }
 
-static void showtext(char *buf, int indent)
+static void showtext(fz_context *ctx, char *buf, int indent)
 {
 	int bol = 1;
 	int c = *buf;
@@ -282,7 +298,7 @@ static void showtext(char *buf, int indent)
 		fz_write_byte(ctx, out, '\n');
 }
 
-static void showjs(void)
+static void showjs(fz_context *ctx)
 {
 	pdf_obj *tree;
 	int i;
@@ -295,12 +311,12 @@ static void showjs(void)
 		pdf_obj *js = pdf_dict_get(ctx, action, PDF_NAME(JS));
 		char *src = pdf_load_stream_or_string_as_utf8(ctx, js);
 		fz_write_printf(ctx, out, "// %s\n", pdf_to_name(ctx, name));
-		showtext(src, 0);
+		showtext(ctx, src, 0);
 		fz_free(ctx, src);
 	}
 }
 
-static void showaction(pdf_obj *action, const char *name)
+static void showaction(fz_context *ctx, pdf_obj *action, const char *name)
 {
 	if (action)
 	{
@@ -309,7 +325,7 @@ static void showaction(pdf_obj *action, const char *name)
 		{
 			char *src = pdf_load_stream_or_string_as_utf8(ctx, js);
 			fz_write_printf(ctx, out, "    %s: {\n", name);
-			showtext(src, 1);
+			showtext(ctx, src, 1);
 			fz_write_printf(ctx, out, "    }\n", name);
 			fz_free(ctx, src);
 		}
@@ -324,10 +340,10 @@ static void showaction(pdf_obj *action, const char *name)
 	}
 }
 
-static void showfield(pdf_obj *field)
+static void showfield(fz_context *ctx, pdf_obj *field, int recurse)
 {
 	pdf_obj *kids, *ft, *parent;
-	const char *tu;
+	const char *tu, *value;
 	char *t;
 	int ff;
 	int i, n;
@@ -381,37 +397,64 @@ static void showfield(pdf_obj *field)
 	if (parent)
 		fz_write_printf(ctx, out, "    Parent: %d\n", pdf_to_num(ctx, parent));
 
-	showaction(pdf_dict_getp(ctx, field, "A"), "Action");
+	showaction(ctx, pdf_dict_getp(ctx, field, "A"), "Action");
 
-	showaction(pdf_dict_getp_inheritable(ctx, field, "AA/K"), "Keystroke");
-	showaction(pdf_dict_getp_inheritable(ctx, field, "AA/V"), "Validate");
-	showaction(pdf_dict_getp_inheritable(ctx, field, "AA/F"), "Format");
-	showaction(pdf_dict_getp_inheritable(ctx, field, "AA/C"), "Calculate");
+	showaction(ctx, pdf_dict_getp_inheritable(ctx, field, "AA/K"), "Keystroke");
+	showaction(ctx, pdf_dict_getp_inheritable(ctx, field, "AA/V"), "Validate");
+	showaction(ctx, pdf_dict_getp_inheritable(ctx, field, "AA/F"), "Format");
+	showaction(ctx, pdf_dict_getp_inheritable(ctx, field, "AA/C"), "Calculate");
 
-	showaction(pdf_dict_getp_inheritable(ctx, field, "AA/E"), "Enter");
-	showaction(pdf_dict_getp_inheritable(ctx, field, "AA/X"), "Exit");
-	showaction(pdf_dict_getp_inheritable(ctx, field, "AA/D"), "Down");
-	showaction(pdf_dict_getp_inheritable(ctx, field, "AA/U"), "Up");
-	showaction(pdf_dict_getp_inheritable(ctx, field, "AA/Fo"), "Focus");
-	showaction(pdf_dict_getp_inheritable(ctx, field, "AA/Bl"), "Blur");
+	showaction(ctx, pdf_dict_getp_inheritable(ctx, field, "AA/E"), "Enter");
+	showaction(ctx, pdf_dict_getp_inheritable(ctx, field, "AA/X"), "Exit");
+	showaction(ctx, pdf_dict_getp_inheritable(ctx, field, "AA/D"), "Down");
+	showaction(ctx, pdf_dict_getp_inheritable(ctx, field, "AA/U"), "Up");
+	showaction(ctx, pdf_dict_getp_inheritable(ctx, field, "AA/Fo"), "Focus");
+	showaction(ctx, pdf_dict_getp_inheritable(ctx, field, "AA/Bl"), "Blur");
+
+	value = pdf_field_value(ctx, field);
+	if (value && value[0])
+		fz_write_printf(ctx, out, "    Value: %q\n", value);
 
 	fz_write_string(ctx, out, "\n");
 
-	kids = pdf_dict_get(ctx, field, PDF_NAME(Kids));
-	n = pdf_array_len(ctx, kids);
-	for (i = 0; i < n; ++i)
-		showfield(pdf_array_get(ctx, kids, i));
+	if (recurse)
+	{
+		kids = pdf_dict_get(ctx, field, PDF_NAME(Kids));
+		n = pdf_array_len(ctx, kids);
+		for (i = 0; i < n; ++i)
+			showfield(ctx, pdf_array_get(ctx, kids, i), 1);
+	}
 }
 
-static void showform(void)
+static void showform(fz_context *ctx)
 {
-	pdf_obj *fields;
-	int i, n;
+	pdf_obj *fields, *page, *obj;
+	int i, k, n, m;
 
 	fields = pdf_dict_getp(ctx, pdf_trailer(ctx, doc), "Root/AcroForm/Fields");
-	n = pdf_array_len(ctx, fields);
-	for (i = 0; i < n; ++i)
-		showfield(pdf_array_get(ctx, fields, i));
+	if (fields)
+	{
+		n = pdf_array_len(ctx, fields);
+		for (i = 0; i < n; ++i)
+			showfield(ctx, pdf_array_get(ctx, fields, i), 1);
+	}
+	else
+	{
+		fz_write_string(ctx, out, "% Root/AcroForm/Fields is missing!\n\n");
+		m = pdf_count_pages(ctx, doc);
+		for (k = 0; k < m; ++k)
+		{
+			page = pdf_lookup_page_obj(ctx, doc, k);
+			fields = pdf_dict_get(ctx, page, PDF_NAME(Annots));
+			n = pdf_array_len(ctx, fields);
+			for (i = 0; i < n; ++i)
+			{
+				obj = pdf_array_get(ctx, fields, i);
+				if (pdf_dict_get(ctx, obj, PDF_NAME(Subtype)) == PDF_NAME(Widget))
+					showfield(ctx, obj, 0);
+			}
+		}
+	}
 }
 
 #define SEP ".[]/"
@@ -429,7 +472,7 @@ static int isnumber(char *s)
 	return 1;
 }
 
-static void showpath(char *path, pdf_obj *obj)
+static void showpath(fz_context *ctx, char *path, pdf_obj *obj)
 {
 	if (path && path[0])
 	{
@@ -448,10 +491,10 @@ static void showpath(char *path, pdf_obj *obj)
 						if (path)
 						{
 							fz_strlcpy(buf, path, sizeof buf);
-							showpath(buf, pdf_array_get(ctx, obj, i));
+							showpath(ctx, buf, pdf_array_get(ctx, obj, i));
 						}
 						else
-							showpath(NULL, pdf_array_get(ctx, obj, i));
+							showpath(ctx, NULL, pdf_array_get(ctx, obj, i));
 					}
 				}
 				else if (pdf_is_dict(ctx, obj))
@@ -462,10 +505,10 @@ static void showpath(char *path, pdf_obj *obj)
 						if (path)
 						{
 							fz_strlcpy(buf, path, sizeof buf);
-							showpath(buf, pdf_dict_get_val(ctx, obj, i));
+							showpath(ctx, buf, pdf_dict_get_val(ctx, obj, i));
 						}
 						else
-							showpath(NULL, pdf_dict_get_val(ctx, obj, i));
+							showpath(ctx, NULL, pdf_dict_get_val(ctx, obj, i));
 					}
 				}
 				else
@@ -477,10 +520,10 @@ static void showpath(char *path, pdf_obj *obj)
 			{
 				int num = atoi(part);
 				num = num < 0 ? pdf_array_len(ctx, obj) + num : num - 1;
-				showpath(path, pdf_array_get(ctx, obj, num));
+				showpath(ctx, path, pdf_array_get(ctx, obj, num));
 			}
 			else
-				showpath(path, pdf_dict_gets(ctx, obj, part));
+				showpath(ctx, path, pdf_dict_gets(ctx, obj, part));
 		}
 		else
 			fz_write_string(ctx, out, "null\n");
@@ -488,16 +531,16 @@ static void showpath(char *path, pdf_obj *obj)
 	else
 	{
 		if (pdf_is_indirect(ctx, obj))
-			showobject(obj);
+			showobject(ctx, obj);
 		else
 		{
-			pdf_print_obj(ctx, out, obj, tight, 0);
+			pdf_print_obj(ctx, out, obj, do_tight, 0);
 			fz_write_string(ctx, out, "\n");
 		}
 	}
 }
 
-static void showpathpage(char *path)
+static void showpathpage(fz_context *ctx, char *path)
 {
 	if (path)
 	{
@@ -514,17 +557,17 @@ static void showpathpage(char *path)
 					if (path)
 					{
 						fz_strlcpy(buf, path, sizeof buf);
-						showpath(buf, pdf_lookup_page_obj(ctx, doc, i));
+						showpath(ctx, buf, pdf_lookup_page_obj(ctx, doc, i));
 					}
 					else
-						showpath(NULL, pdf_lookup_page_obj(ctx, doc, i));
+						showpath(ctx, NULL, pdf_lookup_page_obj(ctx, doc, i));
 				}
 			}
 			else if (isnumber(part))
 			{
 				int num = atoi(part);
 				num = num < 0 ? pdf_count_pages(ctx, doc) + num : num - 1;
-				showpath(path, pdf_lookup_page_obj(ctx, doc, num));
+				showpath(ctx, path, pdf_lookup_page_obj(ctx, doc, num));
 			}
 			else
 				fz_write_string(ctx, out, "null\n");
@@ -534,11 +577,11 @@ static void showpathpage(char *path)
 	}
 	else
 	{
-		showpages();
+		showpages(ctx);
 	}
 }
 
-static void showpathroot(char *path)
+static void showpathroot(fz_context *ctx, char *path)
 {
 	char buf[2000], *list = buf, *part;
 	fz_strlcpy(buf, path, sizeof buf);
@@ -546,9 +589,9 @@ static void showpathroot(char *path)
 	if (part && part[0])
 	{
 		if (!strcmp(part, "trailer"))
-			showpath(list, pdf_trailer(ctx, doc));
+			showpath(ctx, list, pdf_trailer(ctx, doc));
 		else if (!strcmp(part, "pages"))
-			showpathpage(list);
+			showpathpage(ctx, list);
 		else if (isnumber(part))
 		{
 			pdf_obj *obj;
@@ -556,37 +599,37 @@ static void showpathroot(char *path)
 			num = num < 0 ? pdf_xref_len(ctx, doc) + num : num;
 			obj = pdf_new_indirect(ctx, doc, num, 0);
 			fz_try(ctx)
-				showpath(list, obj);
+				showpath(ctx, list, obj);
 			fz_always(ctx)
 				pdf_drop_obj(ctx, obj);
 			fz_catch(ctx)
 				;
 		}
 		else
-			showpath(list, pdf_dict_gets(ctx, pdf_trailer(ctx, doc), part));
+			showpath(ctx, list, pdf_dict_gets(ctx, pdf_trailer(ctx, doc), part));
 	}
 	else
 		fz_write_string(ctx, out, "null\n");
 }
 
-static void show(char *sel)
+static void show(fz_context *ctx, char *sel)
 {
 	if (!strcmp(sel, "trailer"))
-		showtrailer();
+		showtrailer(ctx);
 	else if (!strcmp(sel, "xref"))
-		showxref();
+		showxref(ctx);
 	else if (!strcmp(sel, "pages"))
-		showpages();
+		showpages(ctx);
 	else if (!strcmp(sel, "grep"))
-		showgrep();
+		showgrep(ctx);
 	else if (!strcmp(sel, "outline"))
-		showoutline();
+		showoutline(ctx);
 	else if (!strcmp(sel, "js"))
-		showjs();
+		showjs(ctx);
 	else if (!strcmp(sel, "form"))
-		showform();
+		showform(ctx);
 	else
-		showpathroot(sel);
+		showpathroot(ctx, sel);
 }
 
 int pdfshow_main(int argc, char **argv)
@@ -597,22 +640,24 @@ int pdfshow_main(int argc, char **argv)
 	int c;
 	int errored = 0;
 
-	ctx = fz_new_context(NULL, NULL, FZ_STORE_UNLIMITED);
+	fz_context *ctx = fz_new_context(NULL, NULL, FZ_STORE_UNLIMITED);
 	if (!ctx)
 	{
 		fprintf(stderr, "cannot initialise context\n");
 		exit(1);
 	}
 
-	while ((c = fz_getopt(argc, argv, "p:o:beg")) != -1)
+	while ((c = fz_getopt(argc, argv, "p:o:begrL")) != -1)
 	{
 		switch (c)
 		{
 		case 'p': password = fz_optarg; break;
-		case 'o': output = fz_optarg; break;
+		case 'o': output = fz_optpath(fz_optarg); break;
 		case 'b': showbinary = 1; break;
 		case 'e': showdecode = 0; break;
-		case 'g': tight = 1; break;
+		case 'g': do_tight = 1; break;
+		case 'r': do_repair = 1; break;
+		case 'L': do_label = 1; break;
 		default: return usage();
 		}
 	}
@@ -628,6 +673,7 @@ int pdfshow_main(int argc, char **argv)
 		out = fz_stdout(ctx);
 
 	fz_var(doc);
+	fz_var(labels);
 	fz_try(ctx)
 	{
 		doc = pdf_open_document(ctx, filename);
@@ -635,17 +681,35 @@ int pdfshow_main(int argc, char **argv)
 			if (!pdf_authenticate_password(ctx, doc, password))
 				fz_warn(ctx, "cannot authenticate password: %s", filename);
 
+		if (do_repair)
+		{
+			fz_try(ctx)
+			{
+				pdf_repair_xref(ctx, doc);
+			}
+			fz_catch(ctx)
+			{
+				fz_rethrow_if(ctx, FZ_ERROR_TRYLATER);
+				fz_rethrow_if(ctx, FZ_ERROR_SYSTEM);
+				fz_rethrow_if(ctx, FZ_ERROR_REPAIRED);
+				fz_report_error(ctx);
+			}
+		}
+
+		if (do_label)
+			labels = pdf_load_object_labels(ctx, doc);
+
 		if (fz_optind == argc)
-			showtrailer();
+			showtrailer(ctx);
 
 		while (fz_optind < argc)
-			show(argv[fz_optind++]);
-
-		fz_close_output(ctx, out);
+			show(ctx, argv[fz_optind++]);
 	}
 	fz_always(ctx)
 	{
+		fz_close_output(ctx, out);
 		fz_drop_output(ctx, out);
+		pdf_drop_object_labels(ctx, labels);
 		pdf_drop_document(ctx, doc);
 	}
 	fz_catch(ctx)
